/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;
using System.Reflection;
using System.Diagnostics;

using Microsoft.Scripting.Ast;
using Microsoft.Scripting.Hosting;
using Microsoft.Scripting.Actions;
using Microsoft.Scripting.Generation;



namespace Microsoft.Scripting {
    /// <summary>
    /// ScriptCode is an instance of compiled code that is bound to a specific LanguageContext
    /// but not a specific ScriptModule.  The code can be re-executed multiple times in different
    /// contexts. Hosting API counterpart for this class is <see cref="CompiledCode"/>.
    /// </summary>
    public class ScriptCode {
        public static readonly ScriptCode[] EmptyArray = new ScriptCode[0];

        private readonly CodeBlock _code;
        private readonly LanguageContext _languageContext;
        private readonly CompilerContext _compilerContext;

        private CallTargetWithContext0 _simpleTarget;

        private CallTargetWithContext0 _optimizedTarget;
        private Scope _optimizedScope;

        public ScriptCode(CodeBlock code, LanguageContext languageContext, CompilerContext compilerContext) {
            Debug.Assert(code != null && languageContext != null && compilerContext != null);
            
            this._code = code;
            this._languageContext = languageContext;
            this._compilerContext = compilerContext;
        }

        public LanguageContext LanguageContext {
            get { return _languageContext; }
        }

        public CompilerContext CompilerContext {
            get { return _compilerContext; }
        }

        public SourceUnit SourceUnit {
            get { return _compilerContext.SourceUnit; }
        }

        internal CodeBlock CodeBlock {
            get {
                return _code;
            }
        }

        internal Scope OptimizedScope {
            set {
                Debug.Assert(_optimizedScope == null);
                _optimizedScope = value;
            }
        }

        internal CallTargetWithContext0 OptimizedTarget {
            set {
                _optimizedTarget = value;
            }
        }
        
        public void EnsureCompiled() {            
            if (_simpleTarget == null) {
                lock (this) { // TODO: mutex object
                    if (_simpleTarget == null) {
                        _simpleTarget = _code.CreateDelegate<CallTargetWithContext0>(_compilerContext);
                    }
                }
            }
        }

        private object Run(CodeContext codeContext) {
            _languageContext.ModuleContextEntering(codeContext.ModuleContext);

            if (_languageContext.Engine.Options.FastEvaluation) {
                return CodeBlock.Execute(codeContext);
            }
            
            if (codeContext.Scope == _optimizedScope) { // flag on scope - "IsOptimized"?
                return _optimizedTarget(new CodeContext(_optimizedScope, _languageContext, codeContext.ModuleContext)); // TODO: why do we create a code context here?
            }

            EnsureCompiled();
            return _simpleTarget(codeContext);
        }

        public object Run(ScriptModule module) {
            Utils.Assert.NotNull(module);
            
            ModuleContext moduleContext = _languageContext.EnsureModuleContext(module);
            return Run(new CodeContext(module.Scope, _languageContext, moduleContext));
        }

        public object Run(Scope scope, ModuleContext moduleContext) {
            Utils.Assert.NotNull(scope, moduleContext);

            return Run(new CodeContext(scope, _languageContext, moduleContext));
        }

        public override string ToString() {
            return string.Format("ScriptCode {0} from {1}", 
                SourceUnit.Name, 
                _languageContext.Engine.LanguageProvider.LanguageDisplayName);
        }

        public static ScriptCode FromCompiledCode(CompiledCode compiledCode) {
            if (compiledCode == null) throw new ArgumentNullException("compiledCode");
            return compiledCode.ScriptCode;
        }
    }
}
