/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System;
using System.IO;
using System.Text;
using System.Collections;
using System.Collections.Generic;
using System.Reflection;
using System.Threading;
using System.Diagnostics;
using System.Runtime.Serialization;

using Microsoft.Scripting.Ast;
using Microsoft.Scripting.Hosting;
using Microsoft.Scripting.Generation;

namespace Microsoft.Scripting {

    public delegate void CommandDispatcher(Delegate command);

    [Serializable]
    public class InvalidImplementationException : Exception {
        public InvalidImplementationException()
            : base() {
        }

        public InvalidImplementationException(string message)
            : base(message) {
        }

        public InvalidImplementationException(string message, Exception e)
            : base(message, e) {
        }

#if !SILVERLIGHT // SerializationInfo
        protected InvalidImplementationException(SerializationInfo info, StreamingContext context) : base(info, context) { }
#endif
    }

    [Serializable]
    public class MissingTypeException : Exception {
        public MissingTypeException() {
        }

        public MissingTypeException(string name, Exception e) : 
            base(String.Format(Resources.MissingType, name), e) {
        }

#if !SILVERLIGHT // SerializationInfo
        protected MissingTypeException(SerializationInfo info, StreamingContext context) : base(info, context) { }
#endif
    }

    public sealed class ScriptDomainManager {

        #region Fields and Initialization

        private static readonly object _singletonLock = new object();
        private static ScriptDomainManager _singleton;

        private readonly PlatformAdaptationLayer _pal;
        private readonly IScriptHost _host;
        private readonly Snippets _snippets;
        private readonly ScriptEnvironment _environment;
        private Dictionary<string, WeakReference> _modules;
        private CommandDispatcher _commandDispatcher; // can be null
        
        // singletons:
        public PlatformAdaptationLayer PAL { get { return _pal; } }
        public Snippets Snippets { get { return _snippets; } }
        public ScriptEnvironment Environment { get { return _environment; } }

        /// <summary>
        /// Gets the <see cref="ScriptDomainManager"/> associated with the current AppDomain. 
        /// If there is none, creates and initializes a new environment using setup information associated with the AppDomain 
        /// or stored in a configuration file.
        /// </summary>
        public static ScriptDomainManager CurrentManager {
            get {
                ScriptDomainManager result;
                TryCreateLocal(null, out result);
                return result;
            }
        }

        public IScriptHost Host {
            get { return _host; }
        }

        /// <summary>
        /// Creates a new local <see cref="ScriptDomainManager"/> unless it already exists. 
        /// Returns either <c>true</c> and the newly created environment initialized according to the provided setup information
        /// or <c>false</c> and the existing one ignoring the specified setup information.
        /// </summary>
        internal static bool TryCreateLocal(ScriptEnvironmentSetup setup, out ScriptDomainManager manager) {

            bool new_created = false;

            if (_singleton == null) {

                if (setup == null) {
                    setup = GetSetupInformation();
                }

                lock (_singletonLock) {
                    if (_singleton == null) {
                        ScriptDomainManager singleton = new ScriptDomainManager(setup);
                        Utils.MemoryBarrier();
                        _singleton = singleton;
                        new_created = true;
                    }
                }

            }

            manager = _singleton;
            return new_created;
        }

        private static ScriptEnvironmentSetup GetSetupInformation() {
#if !SILVERLIGHT
            ScriptEnvironmentSetup result;

            // setup provided by app-domain creator:
            result = ScriptEnvironmentSetup.GetAppDomainAssociated(AppDomain.CurrentDomain);
            if (result != null) {
                return result;
            }

            // setup provided in a configuration file:
            // This will load System.Configuration.dll which costs ~350 KB of memory. However, this does not normally 
            // need to be loaded in simple scenarios (like running the console hosts). Hence, the working set cost
            // is only paid in hosted scenarios.
            ScriptConfiguration config = System.Configuration.ConfigurationManager.GetSection(ScriptConfiguration.Section) as ScriptConfiguration;
            if (config != null) {
                // TODO:
                //return config;
            }
#endif

            // default setup:
            return new ScriptEnvironmentSetup(true);
        }

        /// <summary>
        /// Initializes environment according to the setup information.
        /// </summary>
        private ScriptDomainManager(ScriptEnvironmentSetup setup) {
            Debug.Assert(setup != null);

            // create local environment for the host:
            _environment = new ScriptEnvironment(this);
            
            // create PAL (default always available):
            _pal = setup.CreatePAL();

            // let setup register providers listed on it:
            setup.RegisterProviders(this);

            // initialize snippets:
            _snippets = new Snippets();

            // create a local host unless a remote one has already been created:
            _host = setup.CreateScriptHost(_environment);
        }

        #endregion
       
        #region Language Providers

        /// <summary>
        /// Singleton for each language.
        /// </summary>
        private sealed class LanguageProviderDesc {

            private string _assemblyName;
            private string _typeName;
            private LanguageProvider _provider;
            private Type _type;

            public string AssemblyName {
                get { return _assemblyName; }
            }

            public string TypeName {
                get { return _typeName; }
            }

            public LanguageProvider Provider {
                get { return _provider; }
            }

            public LanguageProviderDesc(Type type) {
                Debug.Assert(type != null);

                _type = type;
                _assemblyName = null;
                _typeName = null;
                _provider = null;
            }

            public LanguageProviderDesc(string typeName, string assemblyName) {
                Debug.Assert(typeName != null && assemblyName != null);

                _assemblyName = assemblyName;
                _typeName = typeName;
                _provider = null;
            }

            /// <summary>
            /// Must not be called under a lock as it can potentially call a user code.
            /// </summary>
            /// <exception cref="MissingTypeException"><paramref name="languageId"/></exception>
            /// <exception cref="InvalidImplementationException">The language provider's implementation failed to instantiate.</exception>
            public LanguageProvider LoadProvider(ScriptDomainManager manager) {
                if (_provider == null) {
                    
                    if (_type == null) {
                        try {
                            _type = ScriptDomainManager.CurrentManager.PAL.LoadAssembly(_assemblyName).GetType(_typeName, true);
                        } catch (Exception e) {
                            throw new MissingTypeException(MakeAssemblyQualifiedName(_assemblyName, _typeName), e);
                        }
                    }

                    lock (manager._languageProvidersLock) {
                        manager._languageTypes[_type.AssemblyQualifiedName] = this;
                    }

                    // needn't to be locked, we can create multiple LPs:
                    LanguageProvider provider = Utils.Reflection.CreateInstance<LanguageProvider>(_type, manager);
                    Utils.MemoryBarrier();
                    _provider = provider;
                }
                return _provider;
            }
        }

        // TODO: ReaderWriterLock (Silverlight?)
        private readonly object _languageProvidersLock = new object();
        private readonly Dictionary<string, LanguageProviderDesc> _languageIds = new Dictionary<string, LanguageProviderDesc>(StringComparer.OrdinalIgnoreCase);
        private readonly Dictionary<string, LanguageProviderDesc> _languageTypes = new Dictionary<string, LanguageProviderDesc>();

        public void RegisterLanguageProvider(string assemblyName, string typeName, params string[] identifiers) {
            RegisterLanguageProvider(assemblyName, typeName, false, identifiers);
        }

        public void RegisterLanguageProvider(string assemblyName, string typeName, bool overrideExistingIds, params string[] identifiers) {
            if (identifiers == null) throw new ArgumentNullException("identifiers");

            LanguageProviderDesc singleton_desc;
            bool add_singleton_desc = false;
            string aq_name = MakeAssemblyQualifiedName(typeName, assemblyName);

            lock (_languageProvidersLock) {
                if (!_languageTypes.TryGetValue(aq_name, out singleton_desc)) {
                    add_singleton_desc = true;
                    singleton_desc = new LanguageProviderDesc(typeName, assemblyName);
                }

                // check for conflicts:
                if (!overrideExistingIds) {
                    for (int i = 0; i < identifiers.Length; i++) {
                        LanguageProviderDesc desc;
                        if (_languageIds.TryGetValue(identifiers[i], out desc) && !ReferenceEquals(desc, singleton_desc)) {
                            throw new InvalidOperationException("Conflicting Ids");
                        }
                    }
                }

                // add singleton LP-desc:
                if (add_singleton_desc)
                    _languageTypes.Add(aq_name, singleton_desc);

                // add id mapping to the singleton LP-desc:
                for (int i = 0; i < identifiers.Length; i++) {
                    _languageIds[identifiers[i]] = singleton_desc;
                }
            }
        }

        public bool RemoveLanguageMapping(string identifier) {
            if (identifier == null) throw new ArgumentNullException("identifier");
            
            lock (_languageProvidersLock) {
                return _languageIds.Remove(identifier);
            }
        }

        /// <summary>
        /// Throws an exception on failure.
        /// </summary>
        /// <exception cref="ArgumentNullException"><paramref name="type"/></exception>
        /// <exception cref="ArgumentException"><paramref name="type"/></exception>
        /// <exception cref="MissingTypeException"><paramref name="languageId"/></exception>
        /// <exception cref="InvalidImplementationException">The language provider's implementation failed to instantiate.</exception>
        public LanguageProvider GetLanguageProvider(Type type) {
            if (type == null) throw new ArgumentNullException("type");
            if (!type.IsSubclassOf(typeof(LanguageProvider))) throw new ArgumentException("Invalid type - should be subclass of LanguageProvider"); // TODO

            LanguageProviderDesc desc = null;
            
            lock (_languageProvidersLock) {
                if (!_languageTypes.TryGetValue(type.AssemblyQualifiedName, out desc)) {
                    desc = new LanguageProviderDesc(type);
                    _languageTypes[type.AssemblyQualifiedName] = desc;
                }
            }

            if (desc != null) {
                return desc.LoadProvider(this);
            }

            // not found, not registered:
            throw new ArgumentException(Resources.UnknownLanguageProviderType);
        }

        internal string[] GetLanguageIdentifiers(Type type, bool extensionsOnly) {
            if (type != null && !type.IsSubclassOf(typeof(LanguageProvider))) {
                throw new ArgumentException("Invalid type - should be subclass of LanguageProvider"); // TODO
            }

            bool get_all = type == null;
            List<string> result = new List<string>();

            lock (_languageTypes) {
                LanguageProviderDesc singleton_desc = null;
                if (!get_all && !_languageTypes.TryGetValue(type.AssemblyQualifiedName, out singleton_desc)) {
                    return Utils.Array.EmptyStrings;
                }

                foreach (KeyValuePair<string, LanguageProviderDesc> entry in _languageIds) {
                    if (get_all || ReferenceEquals(entry.Value, singleton_desc)) {
                        if (!extensionsOnly || IsExtensionId(entry.Key)) {
                            result.Add(entry.Key);
                        }
                    }
                }
            }

            return result.ToArray();
        }

        /// <exception cref="ArgumentNullException"><paramref name="languageId"/></exception>
        /// <exception cref="MissingTypeException"><paramref name="languageId"/></exception>
        /// <exception cref="InvalidImplementationException">The language provider's implementation failed to instantiate.</exception>
        public bool TryGetLanguageProvider(string languageId, out LanguageProvider provider) {
            if (languageId == null) throw new ArgumentNullException("languageId");

            bool result;
            LanguageProviderDesc desc;

            lock (_languageProvidersLock) {
                result = _languageIds.TryGetValue(languageId, out desc);
            }

            provider = result ? desc.LoadProvider(this) : null;

            return result;
        }

        /// <exception cref="ArgumentNullException"><paramref name="languageId"/></exception>
        /// <exception cref="ArgumentException">no language registered under languageId</exception>
        /// <exception cref="MissingTypeException"><paramref name="languageId"/></exception>
        /// <exception cref="InvalidImplementationException">The language provider's implementation failed to instantiate.</exception>
        public LanguageProvider GetLanguageProvider(string languageId) {
            if (languageId == null) throw new ArgumentNullException("languageId");

            LanguageProvider result;

            if (!TryGetLanguageProvider(languageId, out result)) {
                throw new ArgumentException(Resources.UnknownLanguageId);
            }

            return result;
        }

        /// <summary>
        /// Gets language provider associated with a specified extension.
        /// </summary>
        /// <exception cref="ArgumentException"><paramref name="extension"/></exception>
        /// <exception cref="MissingTypeException"><paramref name="extension"/></exception>
        /// <exception cref="InvalidImplementationException">The language provider's implementation failed to instantiate.</exception>
        public LanguageProvider GetLanguageProviderByFileExtension(string extension) {
            if (String.IsNullOrEmpty(extension)) throw new ArgumentException("null or empty", "extension"); // TODO

            // TODO: separate hashtable for extensions (see CodeDOM config)
            if (extension[0] != '.') extension = '.' + extension;
            return GetLanguageProvider(extension);
        }

        public bool TryGetLanguageProviderByFileExtension(string extension, out LanguageProvider provider) {
            if (String.IsNullOrEmpty(extension)) {
                provider = null;
                return false;
            }

            // TODO: separate hashtable for extensions (see CodeDOM config)
            if (extension[0] != '.') extension = '.' + extension;
            return TryGetLanguageProvider(extension, out provider);
        }

        public string[] GetRegisteredFileExtensions() {
            return GetLanguageIdentifiers(null, true);
        }

        public string[] GetRegisteredLanguageIdentifiers() {
            return GetLanguageIdentifiers(null, false);
        }

        // TODO: separate hashtable for extensions (see CodeDOM config)
        private bool IsExtensionId(string id) {
            return id.StartsWith(".");
        }

        /// <exception cref="MissingTypeException"><paramref name="languageId"/></exception>
        /// <exception cref="InvalidImplementationException">The language provider's implementation failed to instantiate.</exception>
        public LanguageProvider[] GetLanguageProviders(bool usedOnly) {
            List<LanguageProvider> results = new List<LanguageProvider>(_languageIds.Count);

            List<LanguageProviderDesc> to_be_loaded = usedOnly ? null : new List<LanguageProviderDesc>();
            
            lock (_languageProvidersLock) {
                foreach (LanguageProviderDesc desc in _languageIds.Values) {
                    if (desc.Provider != null) {
                        results.Add(desc.Provider);
                    } else if (!usedOnly) {
                        to_be_loaded.Add(desc);
                    }
                }
            }

            if (!usedOnly) {
                foreach (LanguageProviderDesc desc in to_be_loaded) {
                    results.Add(desc.LoadProvider(this));
                }
            }

            return results.ToArray();
        }

        private static string MakeAssemblyQualifiedName(string typeName, string assemblyName) {
            return String.Concat(typeName, ", ", assemblyName);
        }

        #endregion

        #region Variables

        private IAttributesCollection _variables;

        /// <summary>
        /// A collection of environment variables or <c>null</c> for calling back to the host on each variable access.
        /// It's up to the host to set the property via <see cref="ScriptEnvironment"/> and to ensure its correct behavior and thread safety.
        /// </summary>
        internal IAttributesCollection Variables { get { return _variables; } set { _variables = value; } }

        public void SetVariable(CodeContext context, SymbolId name, object value) {
            IAttributesCollection variables = _variables;
            
            if (variables != null) {
                variables[name] = value;
            } else {
                if (!_host.TrySetVariable(context.LanguageContext.Engine, name, value)) {
                    // TODO:
                    throw context.LanguageContext.MissingName(name);
                }
            }
        }

        public object GetVariable(CodeContext context, SymbolId name) {
            IAttributesCollection variables = _variables;

            if (variables != null) {
                return variables[name];
            } else {
                object result;
                
                if (!_host.TryGetVariable(context.LanguageContext.Engine, name, out result)) {
                    // TODO:
                    throw context.LanguageContext.MissingName(name);
                }

                return result;
            }
        }

        #endregion

        #region Modules

        /// <summary>
        /// Uses the hosts search path and semantics to resolve the provided name to a SourceUnit.
        /// 
        /// If the host provides a SourceUnit which is equal to an already loaded SourceUnit the
        /// previously loaded module is returned.
        /// 
        /// Returns null if a module could not be found.
        /// </summary>
        /// <param name="name">an opaque parameter which has meaning to the host.  Typically a filename without an extension.</param>
        public ScriptModule UseModule(string name) {
            if (name == null) throw new ArgumentNullException("name");
            SourceFileUnit su = _host.ResolveSourceFileUnit(name);

            return CompileAndPublishModule(su);
        }

        /// <summary>
        /// Requests a SourceUnit from the provided path and compiles it to a ScriptModule.
        /// 
        /// If the host provides a SourceUnit which is equal to an already loaded SourceUnit the
        /// previously loaded module is returned.
        /// 
        /// Returns null if a module could not be found.
        /// </summary>
        /// <exception cref="ArgumentNullException"><paramref name="path"/></exception>
        /// <exception cref="ArgumentException">no language registered</exception>
        /// <exception cref="MissingTypeException"><paramref name="languageId"/></exception>
        /// <exception cref="InvalidImplementationException">The language provider's implementation failed to instantiate.</exception>
        public ScriptModule UseModule(string path, string languageId) {
            if (path == null) throw new ArgumentNullException("path");
            ScriptEngine engine = GetLanguageProvider(languageId).GetEngine();

            SourceFileUnit su = _host.GetSourceFileUnit(engine, path, Path.GetFileNameWithoutExtension(path));

            return CompileAndPublishModule(su);
        }

        /// <summary>
        /// Gets the ScriptModule that has been published under the given SourceUnit.
        /// </summary>
        public bool TryGetScriptModule(string publicName, out ScriptModule module) {
            if (_modules == null) {
                module = null;
                return false;
            }

            lock(_modules) {
                module = GetCachedModuleNoLock(publicName);
            }

            return module != null;
        }

        public void PublishModule(ScriptModule module) {
            if (module == null) throw new ArgumentNullException("module");
            PublishModule(module, GetPublicPath(module.FileName));
        }
        
        /// <summary>
        /// Sets the ScriptModule that is registered for the given SourceUnit.
        /// </summary>
        public void PublishModule(ScriptModule module, string publicName) {
            if (module == null) throw new ArgumentNullException("module");
            if (publicName == null) throw new ArgumentNullException("publicName");

            EnsureModules();

            lock (_modules) {
                _modules[publicName] = new WeakReference(module);
            }
        }

        /// <summary>
        /// Gets a list of all ScriptModule's and the associated SourceUnit which generated them.
        /// </summary>
        public IDictionary<string, ScriptModule> GetPublishedModules() {
            IDictionary<string, ScriptModule> res = new Dictionary<string, ScriptModule>();
            if (_modules != null) {
                lock (_modules) {
                    foreach (KeyValuePair<string, WeakReference> kvp in _modules) {
                        if (kvp.Value.IsAlive) {
                            res.Add(kvp.Key, (ScriptModule)kvp.Value.Target);
                        } else {
                            _modules.Remove(kvp.Key);
                        }
                    }
                }
            }
            return res;
        }

        private Dictionary<SourceFileUnit, LoadInfo> _loading = new Dictionary<SourceFileUnit,LoadInfo>();
        class LoadInfo {
            public ScriptModule Module;
            public Thread Thread;
            public Exception Exception;
            public bool Done;
            public ManualResetEvent Mre;
        }

        private ScriptModule CompileAndPublishModule(SourceFileUnit su) {
            if (su == null) return null;

            EnsureModules();

            string public_path = GetPublicPath(su.Path);
            Debug.Assert(public_path != null);

            // check if we've already published this SourceUnit
            lock (_modules) {
                ScriptModule tmp = GetCachedModuleNoLock(public_path);
                if (tmp != null) return tmp;
            }

            // compile and initialize the module...
            ScriptModule mod = su.CompileToModule();
            lock (_modules) {
                // check if someone else compiled it first...
                ScriptModule tmp = GetCachedModuleNoLock(public_path);
                if (tmp != null) return tmp;

                LoadInfo load;
                if (_loading.TryGetValue(su, out load)) {
                    if (load.Thread == Thread.CurrentThread) {
                        return load.Module;
                    }

                    Monitor.Exit(_modules);
                    try {
                        lock (load) {
                            if (!load.Done) {
                                if (load.Mre == null) load.Mre = new ManualResetEvent(false);

                                Monitor.Exit(load);
                                try {
                                    load.Mre.WaitOne();
                                } finally {
                                    Monitor.Enter(load);
                                }
                            }
                        }
                        if(load.Module != null) return load.Module;

                        throw load.Exception;
                    } finally {
                        Monitor.Enter(_modules);
                    }
                }
                load = new LoadInfo();
                load.Module = mod;
                load.Thread = Thread.CurrentThread;
                _loading[su] = load;

                bool success = false;

                Monitor.Exit(_modules);
                try {
                    mod.Execute();
                    success = true;
                    lock (load) {
                        load.Done = true;
                        if (load.Mre != null) load.Mre.Set();
                    }
                    return mod;
                } catch(Exception e) {
                    lock (load) {
                        load.Exception = e;
                        load.Done = true;
                        if (load.Mre != null) load.Mre.Set();
                    }
                    throw e;
                } finally {
                    Monitor.Enter(_modules);
                    _loading.Remove(su);
                    if(success) _modules[public_path] = new WeakReference(mod);
                }
            }
        }

        /// <summary>
        /// Must be called with _modules lock held
        /// </summary>
        private ScriptModule GetCachedModuleNoLock(string publicName) {
            WeakReference wr;
            if (_modules.TryGetValue(publicName, out wr)) {
                if (wr.IsAlive) {
                    return (ScriptModule)wr.Target;
                }

                _modules.Remove(publicName);
            }
            return null;
        }

        private void EnsureModules() {
            if (_modules == null) {
                Interlocked.CompareExchange<Dictionary<string, WeakReference>>(ref _modules,
                    new Dictionary<string, WeakReference>(),
                    null);
            }
        }

        private string GetPublicPath(string path) {
            if (path == null) {
                throw new ArgumentException("Cannot publish anonymous module"); // TODO: resource
            }

            try {
                return _host.NormalizePath(path);
            } catch (ArgumentException) {
                throw new ArgumentException("Invalid name for publication"); // TODO: resource
            }
        }

        /// <summary>
        /// Compiles a list of source units into a single module.
        /// <c>scope</c> can be <c>null</c>.
        /// <c>options</c> can be <c>null</c>.
        /// <c>errorSink</c> can be <c>null</c>.
        /// </summary>
        public ScriptModule CompileModule(string name, ScriptModuleKind kind, Scope scope, CompilerOptions options, ErrorSink errorSink, 
            params SourceUnit[] sourceUnits) {

            if (name == null) throw new ArgumentNullException("name");
            Utils.Array.CheckNonNullElements(sourceUnits, "sourceUnits");

            // TODO: Two phases: parse/compile?
            
            // compiles all source units:
            ScriptCode[] script_codes = new ScriptCode[sourceUnits.Length];
            for (int i = 0; i < sourceUnits.Length; i++) {
                script_codes[i] = ScriptCode.FromCompiledCode(sourceUnits[i].Compile(options, errorSink));
            }

            return CreateModule(name, kind, scope, script_codes);
        }

        /// <summary>
        /// Module creation factory. The only way how to create a module.
        /// </summary>
        public ScriptModule CreateModule(string name, params ScriptCode[] scriptCodes) {
            return CreateModule(name, null, scriptCodes);
        }

        /// <summary>
        /// Module creation factory. The only way how to create a module.
        /// </summary>
        public ScriptModule CreateModule(string name) {
            return CreateModule(name, null, ScriptCode.EmptyArray);
        }
        
        /// <summary>
        /// Module creation factory. The only way how to create a module.
        /// <c>scope</c> can be <c>null</c>.
        /// </summary>
        public ScriptModule CreateModule(string name, Scope scope) {
            return CreateModule(name, scope, ScriptCode.EmptyArray);
        }

        public ScriptModule CreateModule(string name, Scope scope, params ScriptCode[] scriptCodes) {
            return CreateModule(name, ScriptModuleKind.Default, scope, scriptCodes);
        }
        
        /// <summary>
        /// Module creation factory. The only way how to create a module.
        /// Modules compiled from a single source file unit get <see cref="ScriptModule.FileName"/> property set to a host 
        /// normalized full path of that source unit. The property is set to a <c>null</c> reference for other modules.
        /// <c>scope</c> can be <c>null</c>.
        /// 
        /// Ensures creation of module contexts for all languages whose code is assembled into the module.
        /// </summary>
        public ScriptModule CreateModule(string name, ScriptModuleKind kind, Scope scope, params ScriptCode[] scriptCodes) {
            if (name == null) throw new ArgumentNullException("name");
            Utils.Array.CheckNonNullElements(scriptCodes, "scriptCodes");

            OptimizedModuleGenerator generator = null;

            if (scope == null) {
                if (scriptCodes.Length == 1) {
                    if (scriptCodes[0].LanguageContext.Engine.Options.FastEvaluation) {
                        scope = new Scope();
                    } else {
                        generator = OptimizedModuleGenerator.Create(scriptCodes);
                        scope = generator.GenerateScope();
                    }
                } else {
                    scope = new Scope();
                }
            }
            
            ScriptModule result = new ScriptModule(name, kind, scope, scriptCodes);

            // single source file unit modules have unique full path:
            SourceFileUnit sfu;
            if (scriptCodes.Length == 1 && (sfu = scriptCodes[0].SourceUnit as SourceFileUnit) != null) {
                // TODO: remove normalization (test NessieFilenamesExecute shouldn't depend on it)
                try {
                    result.FileName = _host.NormalizePath(sfu.Path);
                } catch (ArgumentException) {
                    result.FileName = null;
                }
            } else {
                result.FileName = null;
            }

            // Initializes module contexts for all contained optimized script codes;
            // Module contexts stored in optimized module's code contexts cannot be changed from now on.
            // Such change would require the CodeContexts to be overwritten.
            if (generator != null) {
                generator.BindGeneratedCodeToModule(result);
            } else {
                foreach (ScriptCode code in scriptCodes) {
                    code.LanguageContext.EnsureModuleContext(result);
                }
            }
            
            _host.ModuleCreated(result);
            return result;
        }

        #endregion

        #region Command Dispatching

        // This can be set to a method like System.Windows.Forms.Control.Invoke for Winforms scenario 
        // to cause code to be executed on a separate thread.
        // It will be called with a null argument to indicate that the console session should be terminated.
        // Can be null.

        public CommandDispatcher GetCommandDispatcher() {
            return _commandDispatcher;
        }

        public CommandDispatcher SetCommandDispatcher(CommandDispatcher dispatcher) {
            return Interlocked.Exchange(ref _commandDispatcher, dispatcher);
        }

        public void DispatchCommand(Delegate command) {
            CommandDispatcher dispatcher = _commandDispatcher;
            if (dispatcher != null) {
                dispatcher(command);
            }
        }

        #endregion

        #region TODO

        // TODO: remove or reduce
        public ScriptDomainOptions GlobalOptions {
            get {
                return _options;
            }
            set {
                if (value == null) throw new ArgumentNullException("value");
                _options = value;
            }
        }

        // TODO: remove or reduce     
        private static ScriptDomainOptions _options = new ScriptDomainOptions();

        // TODO: remove or reduce
        public static ScriptDomainOptions Options {
            get { return _options; }
        }

        #endregion
    }
}
