/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;
using Ruby.Runtime;
using Microsoft.Scripting;
using Ruby.Builtins;

namespace Ruby {

    [AttributeUsage(AttributeTargets.Parameter, AllowMultiple = false)]
    public sealed class BlockAttribute : Attribute {
        public BlockAttribute() {
        }
    }

    /// <summary>
    /// Applied to assemblies containing Ruby library methods.
    /// Specifies an initializer for the library, which is a type that publishes RubyModules and RubyClasses defined in the assembly.
    /// </summary>
    [AttributeUsage(AttributeTargets.Assembly, AllowMultiple = true)]
    public sealed class RubyLibraryAttribute : Attribute {
        private Type/*!*/ _initializer;

        public Type/*!*/ Initializer {
            get { return _initializer; }
        }

        public RubyLibraryAttribute(Type/*!*/ initializer) {
            if (initializer == null) throw new ArgumentNullException("initializer");
            _initializer = initializer;
        }
    }

    [AttributeUsage(AttributeTargets.Field, AllowMultiple = true)]
    public sealed class RubyConstantAttribute : Attribute {
        private readonly string _name;

        public string Name { get { return _name; } }

        public RubyConstantAttribute() {
        }

        public RubyConstantAttribute(string/*!*/ name) {
            if (name == null) throw new ArgumentNullException("name");
            _name = name;
        }
    }

    [AttributeUsage(AttributeTargets.Method, AllowMultiple = true)]
    public sealed class RubyMethodAttribute : Attribute {
        private readonly string/*!*/ _name;
        private readonly RubyMethodAttributes _methodAttributes;

        public string/*!*/ Name {
            get { return _name; }
        }

        public RubyMethodAttributes MethodAttributes {
            get { return _methodAttributes; }
        }
        
        public RubyMethodAttribute(string/*!*/ name)
            : this(name, RubyMethodAttributes.Default) {
        }

        public RubyMethodAttribute(string/*!*/ name, RubyMethodAttributes methodAttributes)
            : base() {
            if (name == null) throw new ArgumentNullException("name");

            _name = name;
            _methodAttributes = methodAttributes;
        }
    }

    [AttributeUsage(AttributeTargets.Method, AllowMultiple = true)]
    public sealed class RubyConstructorAttribute : Attribute {
        public RubyConstructorAttribute() {
        }
    }

    [AttributeUsage(AttributeTargets.Class, AllowMultiple = true)]
    public class RubyModuleAttribute : Attribute {
        private readonly string _name;

        public string Name {
            get { return _name; }
        }

        public RubyModuleAttribute() {
        }

        public RubyModuleAttribute(string/*!*/ name) {
            if (name == null) throw new ArgumentNullException("name");
            _name = name;
        }
    }

    [AttributeUsage(AttributeTargets.Class, AllowMultiple = false)]
    public class RubyClassAttribute : RubyModuleAttribute {
        public RubyClassAttribute() {
        }

        public RubyClassAttribute(string/*!*/ name) 
            : base(name) {
        }
    }

    [AttributeUsage(AttributeTargets.Class, AllowMultiple = false)]
    public sealed class RubyExtensionClassAttribute : RubyClassAttribute {
        private Type _inherits;
        private Type _extends;

        public Type Inherits {
            get { return _inherits; }
            set { _inherits = value; }
        }

        public Type Extends {
            get { return _extends; }
        }

        public RubyExtensionClassAttribute(string/*!*/ name, Type/*!*/ extends) 
            : base(name) {

            if (extends == null) throw new ArgumentNullException("extends");
            _extends = extends;
        }
    }

    [AttributeUsage(AttributeTargets.Class, AllowMultiple = true)]
    public sealed class IncludesAttribute : Attribute {
        private readonly Type/*!*/ _type;

        public Type/*!*/ Type {
            get { return _type; }
        }

        public IncludesAttribute(Type/*!*/ type) {
            if (type == null) throw new ArgumentNullException("type");
            _type = type;
        }
    }
}
