/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using Ruby.Runtime;
using System.Collections;
using Microsoft.Scripting;
using Microsoft.Scripting.Actions;
using Ruby.Compiler;
using Ruby.Hosting;
using MSA = Microsoft.Scripting.Ast;

namespace Ruby.Builtins {

    /// <summary>
    /// Array inherits from Object, mixes in Enumerable.
    /// Ruby array is basically List{object}.
    /// </summary>
    [RubyExtensionClass("Array", typeof(List<object>), Inherits = typeof(object)), Includes(typeof(Enumerable))]
    public static class ArrayOps {

        private static int NormalizeIndex(List<object> list, int index) {
            return index < 0 ? index + list.Count : index;
        }

        [RubyConstructor]
        public static List<object> Create() {
            return new List<object>();
        }

        [RubyMethod("[]", RubyMethodAttributes.PublicSingleton)]
        public static List<object> MakeArray(params object[] args) {
            List<object> result = new List<object>(args.Length - 1);
            for (int i = 1; i < args.Length; ++i)
                result.Add(args[i]);
            return result;
        }

        [RubyConstructor]
        public static List<object>/*!*/ CreateArray() {
            return new List<object>();
        }

        [RubyConstructor]
        public static List<object>/*!*/ CreateArray(int size, object value) {
            return new List<object>(size);
        }

        [RubyConstructor]
        public static List<object>/*!*/ CreateArray(int size) {
            return new List<object>(size);
        }

        [RubyConstructor]
        public static List<object>/*!*/ CreateArray(object size) {
            return new List<object>();
        }

        [RubyConstructor]
        public static List<object>/*!*/ CreateArray(object size, object value) {
            return new List<object>();
        }

        [RubyConstructor]
        public static List<object>/*!*/ CreateArray(List<object>/*!*/ other) {
            return new List<object>(other.Count);
        }

        [RubyConstructor]
        public static List<object>/*!*/ CreateArray(Proc block, int size) {
            return new List<object>(size);
        }

        [RubyMethodAttribute("initialize", RubyMethodAttributes.PublicInstance)]
        public static void Initialize(List<object> list) {
        }

        [RubyMethodAttribute("initialize", RubyMethodAttributes.PublicInstance)]
        public static void InitializeValue(List<object> list, int size, object value) {
            for (int i = 0; i < size; ++i)
                list.Add(value);
        }

        [RubyMethodAttribute("initialize", RubyMethodAttributes.PublicInstance)]
        public static void Initialize(List<object> list, int size) {
            InitializeValue(list, size, null);
        }

        [RubyMethodAttribute("initialize", RubyMethodAttributes.PublicInstance)]
        public static void Initialize(List<object> list, object size) {
            InitializeValue(list, RubySites.ToInt(size), null);
        }

        [RubyMethodAttribute("initialize", RubyMethodAttributes.PublicInstance)]
        public static void Initialize(List<object> list, object size, object value) {
            InitializeValue(list, RubySites.ToInt(size), value);
        }

        [RubyMethodAttribute("initialize", RubyMethodAttributes.PublicInstance)]
        public static void Initialize(List<object> list, List<object> other) {
            list.AddRange(other);
        }

        private static DynamicSite<Proc, BlockReturnReasonHolder, int, object> InitializeBlockSharedSite =
            DynamicSite<Proc, BlockReturnReasonHolder, int, object>.Create(CallAction.Simple);

        [RubyMethodAttribute("initialize", RubyMethodAttributes.PublicInstance)]
        public static void Initialize(CodeContext context, List<object> self, Proc block, int size) {
            BlockReturnReasonHolder returnReason = new BlockReturnReasonHolder();
            int i = 0;
            while (i < size) {
                try {
                    self.Add(InitializeBlockSharedSite.Invoke(context, block, returnReason, i));
                } catch (StackUnwinder.BlockReturn u) {
                    returnReason.Value = u.Reason;
                }

                switch (returnReason.Value) {
                    case BlockReturnReason.Undefined:
                        break;

                    case BlockReturnReason.Next:
                        returnReason.Value = BlockReturnReason.Undefined;
                        break;

                    case BlockReturnReason.BreakOrRetry:
                        return;

                    case BlockReturnReason.Redo:
                        returnReason.Value = BlockReturnReason.Undefined;
                        continue;
                }
                i++;
            }
        }

        [RubyMethodAttribute("*", RubyMethodAttributes.PublicInstance)]
        public static List<object> Repetition(List<object> self, int repeat) {
            if (repeat < 0)
                throw RubyExceptions.CreateArgumentError("negative argument");

            List<object> result = new List<object>(self.Count * repeat);
            for (int i = 0; i < repeat; ++i)
                result.AddRange(self);
            return result;
        }

        [RubyMethodAttribute("*", RubyMethodAttributes.PublicInstance)]
        public static MutableString Repetition(List<object> self, string separator) {
            return Join(self, separator);
        }

        [RubyMethodAttribute("*", RubyMethodAttributes.PublicInstance)]
        public static MutableString Repetition(List<object> self, MutableString separator) {
            return Join(self, separator);
        }

        [RubyMethodAttribute("*", RubyMethodAttributes.PublicInstance)]
        public static object Repetition(List<object> self, object repeat) {
            if (RubySites.RespondTo(repeat, "to_str"))
                return Repetition(self, RubySites.ToStr(repeat));
            else
                return Repetition(self, RubySites.ToInt(repeat));
        }

        [RubyMethodAttribute("+", RubyMethodAttributes.PublicInstance)]
        public static List<object> Concatenate(List<object> self, List<object> other) {
            if (other == null) 
                throw RubyExceptions.CreateTypeError("cannot convert nil into Array");

            List<object> result = new List<object>(self.Count + other.Count);
            result.AddRange(self);
            result.AddRange(other);
            return result;
        }

        [RubyMethodAttribute("+", RubyMethodAttributes.PublicInstance)]
        public static List<object> Concatenate(List<object> self, object other) {
            return Concatenate(self, RubySites.ToAry(other));
        }

        [RubyMethodAttribute("<<", RubyMethodAttributes.PublicInstance)]
        public static List<object> Append(List<object> self, object value) {
            self.Add(value);
            return self;
        }

        [RubyMethodAttribute("<=>", RubyMethodAttributes.PublicInstance)]
        public static int Compare(List<object> self, List<object> other) {
            int limit = System.Math.Min(self.Count, other.Count);

            for (int i = 0; i < limit; ++i) {
                int result = RubySites.Compare(self[i], other[i]);
                if (result != 0)
                    return result;
            }

            if (self.Count < other.Count)
                return -1;
            else if (self.Count == other.Count)
                return 0;
            else
                return 1;
        }

        [RubyMethodAttribute("<=>", RubyMethodAttributes.PublicInstance)]
        public static int Compare(List<object>/*!*/ self, object other) {
            return Compare(self, RubySites.ToAry(other));
        }

        [RubyMethodAttribute("==", RubyMethodAttributes.PublicInstance)]
        public static bool Equals(List<object>/*!*/ self, object other) {
            return false;
        }
        
        [RubyMethodAttribute("==", RubyMethodAttributes.PublicInstance)]
        public static bool Equals(List<object>/*!*/ self, [NotNull]List<object>/*!*/ other) {
            Utils.Assert.NotNull(self, other);

            if (self.Count != other.Count)
                return false;

            for (int i = 0; i < self.Count; ++i) {
                bool result = RubySites.Equal(self[i], other[i]);
                if (!result)
                    return false;
            }
            return true;
        }

        #region []

        [RubyMethodAttribute("[]", RubyMethodAttributes.PublicInstance)]
        [RubyMethodAttribute("slice", RubyMethodAttributes.PublicInstance)]
        public static object GetElement(List<object> list, int index) {
            index = index < 0 ? index + list.Count : index;
            return index >= 0 && index < list.Count ? list[index] : null;
        }

        [RubyMethodAttribute("[]", RubyMethodAttributes.PublicInstance)]
        [RubyMethodAttribute("slice", RubyMethodAttributes.PublicInstance)]
        public static object GetElement(List<object> list, object index) {
            return GetElement(list, RubySites.ToInt(index));
        }

        [RubyMethodAttribute("[]", RubyMethodAttributes.PublicInstance)]
        [RubyMethodAttribute("slice", RubyMethodAttributes.PublicInstance)]
        public static List<object> GetElements(List<object> list, int index, int count) {
            index = index < 0 ? index + list.Count : index;
            if (index < 0 || count < 0 || index > list.Count)
                return null;

            if (count == 0)
                return new List<object>();

            count = index + count > list.Count ? list.Count - index : count;
            return list.GetRange(index, count);
        }

        [RubyMethodAttribute("[]", RubyMethodAttributes.PublicInstance)]
        [RubyMethodAttribute("slice", RubyMethodAttributes.PublicInstance)]
        public static List<object> GetElement(List<object> list, int index, object count) {
            return GetElements(list, index, RubySites.ToInt(count));
        }
        
        [RubyMethodAttribute("[]", RubyMethodAttributes.PublicInstance)]
        [RubyMethodAttribute("slice", RubyMethodAttributes.PublicInstance)]
        public static List<object> GetElement(List<object> list, object index, int count) {
            return GetElements(list, RubySites.ToInt(index), count);
        }

        [RubyMethodAttribute("[]", RubyMethodAttributes.PublicInstance)]
        [RubyMethodAttribute("slice", RubyMethodAttributes.PublicInstance)]
        public static List<object> GetElement(List<object> list, object index, object count) {
            return GetElements(list, RubySites.ToInt(index), RubySites.ToInt(count));
        }

        [RubyMethodAttribute("[]", RubyMethodAttributes.PublicInstance)]
        [RubyMethodAttribute("slice", RubyMethodAttributes.PublicInstance)]
        public static List<object> GetElement(List<object> array, Range range) {
            int begin = RubySites.ToInt(RubySites.RangeBegin(range));
            int end = RubySites.ToInt(RubySites.RangeEnd(range));
            
            begin = begin < 0 ? begin + array.Count : begin;

            if (begin < 0 || begin > array.Count)
                return null;

            end = end < 0 ? end + array.Count : end;

            int count = RubySites.RangeExcludeEnd(range) ? end - begin : end - begin + 1;
            return count < 0 ? new List<object>() : GetElements(array, begin, count);
        }

        #endregion

        #region []=

        public static void ExpandList(List<object> list, int index) {
            int diff = index - list.Count;
            for (int i = 0; i < diff; ++i)
                list.Add(null);
        }

        public static void OverwriteOrAdd(List<object> list, int index, object value) {
            if (index < list.Count)
                list[index] = value;
            else
                list.Add(value);
        }

        public static void DeleteItems(List<object> list, int index, int length) {
            if (index >= list.Count) {
                ExpandList(list, index);
            } else {
                // normalize for max length
                if (index + length > list.Count)
                    length = list.Count - index;
                list.RemoveRange(index, length);
            }
        }

        [RubyMethodAttribute("[]=", RubyMethodAttributes.PublicInstance)]
        public static object SetElement(List<object> list, int index, object value) {
            index = index < 0 ? index + list.Count : index;

            if (index < 0)
                throw RubyExceptions.CreateIndexError(String.Format("index {0} out of array", index));

            if (index < list.Count)
                list[index] = value;
            else {
                ExpandList(list, index);
                list.Add(value);
            }
            return value;
        }

        [RubyMethodAttribute("[]=", RubyMethodAttributes.PublicInstance)]
        public static object SetElement(List<object> list, object index, object value) {
            return SetElement(list, RubySites.ToInt(index), value);
        }

        [RubyMethodAttribute("[]=", RubyMethodAttributes.PublicInstance)]
        public static object SetElement(List<object> list, int index, int length, object value) {
            if (length < 0)
                throw RubyExceptions.CreateIndexError(String.Format("negative length ({0})", length));

            index = index < 0 ? index + list.Count : index;
            if (index < 0)
                throw RubyExceptions.CreateIndexError(String.Format("index {0} out of array", index));

            List<object> valueAsList = value as List<object>;

            if (value == null || (valueAsList != null && valueAsList.Count == 0)) {
                DeleteItems(list, index, length);
            } else {
                if (valueAsList == null) {
                    SetElement(list, index, value);
                } else {
                    ExpandList(list, index);

                    int limit = length > valueAsList.Count ? valueAsList.Count : length;

                    for (int i = 0; i < limit; ++i)
                        OverwriteOrAdd(list, index + i, valueAsList[i]);

                    if (length < valueAsList.Count) {
                        list.InsertRange(index + limit, valueAsList.GetRange(limit, valueAsList.Count - limit));
                    } else {
                        list.RemoveRange(index + limit, System.Math.Min(length - valueAsList.Count, list.Count - (index + limit)));
                    }
                }
            }

            return value;
        }

        [RubyMethodAttribute("[]=", RubyMethodAttributes.PublicInstance)]
        public static object SetElement(List<object> list, int index, object length, object value) {
            return SetElement(list, index, RubySites.ToInt(length), value);
        }

        [RubyMethodAttribute("[]=", RubyMethodAttributes.PublicInstance)]
        public static object SetElement(List<object> list, object index, int length, object value) {
            return SetElement(list, RubySites.ToInt(index), length, value);
        }

        [RubyMethodAttribute("[]=", RubyMethodAttributes.PublicInstance)]
        public static object SetElement(List<object> list, object index, object length, object value) {
            return SetElement(list, RubySites.ToInt(index), RubySites.ToInt(length), value);
        }

        [RubyMethodAttribute("[]=", RubyMethodAttributes.PublicInstance)]
        public static object SetElement(List<object> array, Range range, object value) {
            int begin = RubySites.ToInt(RubySites.RangeBegin(range));
            int end = RubySites.ToInt(RubySites.RangeEnd(range));

            begin = begin < 0 ? begin + array.Count : begin;
            if (begin < 0)
                throw RubyExceptions.CreateRangeError(String.Format("{0}..{1} out of range", begin, end));

            end = end < 0 ? end + array.Count : end;

            int count = RubySites.RangeExcludeEnd(range) ? end - begin : end - begin + 1;
            return SetElement(array, begin, System.Math.Max(count, 0), value);
        }

        #endregion

        [RubyMethodAttribute("|", RubyMethodAttributes.PublicInstance)]
        public static List<object> Union(List<object> self, List<object> other) {
            // Unoptimized version of this since it is likely that these will be small sets ... need to do perf
            // investigation of this method to see how the MRI implementation performs

            List<object> result = new List<object>();

            foreach (object item in self)
                if (!result.Contains(item))
                    result.Add(item);

            foreach (object item in other)
                if (!result.Contains(item))
                    result.Add(item);

            return result;
        }

        public static List<object> GetContainerOf(List<object> list, int index, object item) {
            foreach (object current in list) {
                List<object> subArray = current as List<object>;
                if (subArray != null && subArray.Count > index) {
                    if (RubySites.Equal(subArray[index], item))
                        return subArray;
                }
            }
            return null;
        }

        [RubyMethodAttribute("assoc", RubyMethodAttributes.PublicInstance)]
        public static List<object> GetContainerOfFirstItem(List<object> self, object item) {
            return GetContainerOf(self, 0, item);
        }

        [RubyMethodAttribute("at", RubyMethodAttributes.PublicInstance)]
        public static object At(List<object> self, int index) {
            return GetElement(self, index);
        }

        [RubyMethodAttribute("at", RubyMethodAttributes.PublicInstance)]
        public static object At(List<object> self, object index) {
            return GetElement(self, RubySites.ToInt(index));
        }

        [RubyMethodAttribute("clear", RubyMethodAttributes.PublicInstance)]
        public static List<object> Clear(List<object> self) {
            self.Clear();
            return self;
        }

        private static DynamicSite<Proc, BlockReturnReasonHolder, object, object> CollectBlockSharedSite =
            DynamicSite<Proc, BlockReturnReasonHolder, object, object>.Create(CallAction.Simple);

        [RubyMethodAttribute("collect!", RubyMethodAttributes.PublicInstance)]
        [RubyMethodAttribute("map!", RubyMethodAttributes.PublicInstance)]
        public static object CollectInPlace(CodeContext/*!*/ context, List<object>/*!*/ self, Proc/*!*/ block) {
            Utils.Assert.NotNull(context, self, block);

            BlockReturnReasonHolder returnReason = new BlockReturnReasonHolder();
            int i = 0;
            while (i < self.Count) {
                try {
                    self[i] = CollectBlockSharedSite.Invoke(context, block, returnReason, self[i]);
                } catch (StackUnwinder.BlockReturn u) {
                    returnReason.Value = u.Reason;
                }

                switch (returnReason.Value) {
                    case BlockReturnReason.Undefined:
                        break;

                    case BlockReturnReason.Next:
                        returnReason.Value = BlockReturnReason.Undefined;
                        break;

                    case BlockReturnReason.BreakOrRetry:
                        return self;

                    case BlockReturnReason.Redo:
                        returnReason.Value = BlockReturnReason.Undefined;
                        continue;
                }
                i++;
            }

            return self;
        }

        [RubyMethodAttribute("compact", RubyMethodAttributes.PublicInstance)]
        public static List<object> Compact(List<object> self) {
            List<object> result = new List<object>();

            foreach (object item in self)
                if (item != null)
                    result.Add(item);

            return result;
        }

        [RubyMethodAttribute("compact!", RubyMethodAttributes.PublicInstance)]
        public static List<object> CompactInPlace(List<object> self) {
            bool changed = false;
            int i = 0;
            while (i < self.Count) {
                if (self[i] == null) {
                    changed = true;
                    self.RemoveAt(i);
                } else
                    ++i;
            }
            return changed ? self : null;
        }

        [RubyMethodAttribute("concat", RubyMethodAttributes.PublicInstance)]
        public static List<object> Concat(List<object> self, List<object> other) {
            self.AddRange(other);
            return self;
        }

        [RubyMethodAttribute("concat", RubyMethodAttributes.PublicInstance)]
        public static List<object> Concat(List<object> self, object other) {
            if (other == null)
                throw RubyExceptions.CreateTypeError("cannot convert nil into Array");

            if (RubySites.RespondTo(other, "to_ary"))
                return Concat(self, RubySites.ToAry(other));
            else
                throw RubyExceptions.CannotConvertTypeToTargetType(other, "Array");
        }

        public static bool Remove(List<object> self, object item) {
            int i = 0;
            bool removed = false;
            while (i < self.Count) {
                if (RubySites.Equal(self[i], item)) {
                    self.RemoveAt(i);
                    removed = true;
                } else
                    ++i;
            }
            return removed;
        }

        [RubyMethodAttribute("delete", RubyMethodAttributes.PublicInstance)]
        public static object Delete(List<object> self, object item) {
            return Remove(self, item) ? item : null;
        }

        private static DynamicSite<Proc, BlockReturnReasonHolder, object> DeleteBlockSharedSite =
            DynamicSite<Proc, BlockReturnReasonHolder, object>.Create(CallAction.Simple);

        [RubyMethodAttribute("delete", RubyMethodAttributes.PublicInstance)]
        public static object Delete(CodeContext context, List<object> self, Proc block, object item) {
            bool removed = Remove(self, item);

            if (block != null) {
                BlockReturnReasonHolder returnReason = new BlockReturnReasonHolder();
                while (true) {
                    try {
                        return DeleteBlockSharedSite.Invoke(context, block, returnReason);
                    } catch (StackUnwinder.BlockReturn u) {
                        returnReason.Value = u.Reason;
                    }

                    switch (returnReason.Value) {
                        case BlockReturnReason.Undefined:
                            break;

                        case BlockReturnReason.Next:
                            returnReason.Value = BlockReturnReason.Undefined;
                            break;

                        case BlockReturnReason.BreakOrRetry:
                            return self;

                        case BlockReturnReason.Redo:
                            returnReason.Value = BlockReturnReason.Undefined;
                            continue;
                    }
                }
            }
            return removed ? item : null;
        }

        [RubyMethodAttribute("delete_at", RubyMethodAttributes.PublicInstance)]
        public static object DeleteAt(List<object> self, object indexValue) {
            int index = RubySites.ToInt(indexValue);
            index = index < 0 ? index + self.Count : index;
            if (index < 0 || index > self.Count)
                return null;

            object result = GetElement(self, index);
            self.RemoveAt(index);
            return result;
        }

        private static DynamicSite<Proc, BlockReturnReasonHolder, object, object> EachSharedSite =
            DynamicSite<Proc, BlockReturnReasonHolder, object, object>.Create(CallAction.Simple);

        [RubyMethodAttribute("each", RubyMethodAttributes.PublicInstance)]
        public static List<object>/*!*/ Each(CodeContext/*!*/ context, List<object>/*!*/ self, Proc/*!*/ block) {
            Utils.Assert.NotNull(context, self, block);

            BlockReturnReasonHolder returnReason = new BlockReturnReasonHolder();
            int i = 0;
            while (i < self.Count) {
                try {
                    EachSharedSite.Invoke(context, block, returnReason, self[i]);
                } catch (StackUnwinder.BlockReturn u) {
                    returnReason.Value = u.Reason;
                }

                switch (returnReason.Value) {
                    case BlockReturnReason.Undefined:
                        break;

                    case BlockReturnReason.Next:
                        returnReason.Value = BlockReturnReason.Undefined;
                        break;

                    case BlockReturnReason.BreakOrRetry:
                        return self;

                    case BlockReturnReason.Redo:
                        returnReason.Value = BlockReturnReason.Undefined;
                        continue;
                }
                i++;
            }
            return self;
        }

        private static DynamicSite<Proc, BlockReturnReasonHolder, int, object> EachIndexSharedSite =
            DynamicSite<Proc, BlockReturnReasonHolder, int, object>.Create(CallAction.Simple);

        [RubyMethodAttribute("each_index", RubyMethodAttributes.PublicInstance)]
        public static object EachIndex(CodeContext/*!*/ context, List<object>/*!*/ self, Proc/*!*/ block) {
            Utils.Assert.NotNull(context, self, block);

            BlockReturnReasonHolder returnReason = new BlockReturnReasonHolder();
            int i = 0;
            while (i < self.Count) {
                try {
                    EachIndexSharedSite.Invoke(context, block, returnReason, i);
                } catch (StackUnwinder.BlockReturn u) {
                    returnReason.Value = u.Reason;
                }

                switch (returnReason.Value) {
                    case BlockReturnReason.Undefined:
                        break;

                    case BlockReturnReason.Next:
                        returnReason.Value = BlockReturnReason.Undefined;
                        break;

                    case BlockReturnReason.BreakOrRetry:
                        return self;

                    case BlockReturnReason.Redo:
                        returnReason.Value = BlockReturnReason.Undefined;
                        continue;
                }
                i++;
            }

            return self;
        }

        [RubyMethodAttribute("empty?", RubyMethodAttributes.PublicInstance)]
        public static bool Empty(List<object> self) {
            return self.Count == 0;
        }

        [RubyMethodAttribute("fetch", RubyMethodAttributes.PublicInstance)]
        public static object Fetch(List<object> list, int index) {
            index = index < 0 ? index + list.Count : index;
            if (index < 0 || index >= list.Count)
                throw RubyExceptions.CreateIndexError(String.Format("index {0} out of array", index));
            return list[index];
        }

        [RubyMethodAttribute("fetch", RubyMethodAttributes.PublicInstance)]
        public static object Fetch(List<object> list, object index) {
            return Fetch(list, RubySites.ToInt(index));
        }

        [RubyMethodAttribute("fetch", RubyMethodAttributes.PublicInstance)]
        public static object Fetch(List<object> list, int index, object defaultValue) {
            index = index < 0 ? index + list.Count : index;
            return (index < 0 || index >= list.Count) ? defaultValue : list[index];
        }

        [RubyMethodAttribute("first", RubyMethodAttributes.PublicInstance)]
        public static object First(List<object> self) {
            return self.Count == 0 ? null : self[0];
        }

        [RubyMethodAttribute("first", RubyMethodAttributes.PublicInstance)]
        public static List<object> First(List<object> self, int count) {
            if (count < 0)
                throw RubyExceptions.CreateArgumentError("negative array size (or size too big)");

            count = count > self.Count ? self.Count : count;
            return self.GetRange(0, count);
        }

        [RubyMethodAttribute("first", RubyMethodAttributes.PublicInstance)]
        public static object First(List<object> self, object count) {
            return First(self, RubySites.ToInt(count));
        }

        public static bool TryRecursiveFlatten(List<object> list, List<object> result, Dictionary<object, bool> seen) {
            bool flattened = false;
            for (int i = 0; i < list.Count; ++i) {
                object item = list[i];

                if (item is List<object>) {
                    flattened = true;
                    if (!seen.ContainsKey(item)) {
                        seen.Add(item, true);
                        bool f = TryRecursiveFlatten((List<object>)item, result, seen);
                        if (f)
                            flattened = true;
                    } else
                        throw RubyExceptions.CreateArgumentError("tried to flatten recursive array");
                } else
                    result.Add(item);
            }
            return flattened;
        }

        public static bool TryFlattenArray(List<object> list, out List<object> result) {
            Dictionary<object, bool> seen = new Dictionary<object, bool>();
            seen.Add(list, true);
            result = new List<object>();
            return TryRecursiveFlatten(list, result, seen);
        }

        [RubyMethodAttribute("flatten", RubyMethodAttributes.PublicInstance)]
        public static List<object> Flatten(List<object> self) {
            List<object> result;
            TryFlattenArray(self, out result);
            return result;
        }

        [RubyMethodAttribute("flatten!", RubyMethodAttributes.PublicInstance)]
        public static List<object> FlattenInPlace(List<object> self) {
            List<object> result;
            if (!TryFlattenArray(self, out result))
                return null;

            self.Clear();
            self.AddRange(result);
            return self;
        }


        [RubyMethodAttribute("include?", RubyMethodAttributes.PublicInstance)]
        public static bool Include(List<object> self, object item) {
            return Index(self, item) != null;
        }

        [RubyMethodAttribute("index", RubyMethodAttributes.PublicInstance)]
        public static object Index(List<object> self, object item) {
            for (int i = 0; i < self.Count; ++i) {
                if (RubySites.Equal(self[i], item))
                    return i;
            }
            return null;
        }

        public static object GetIndicesAsNestedArrays(List<object> list, object[] parameters) {
            List<object> result = new List<object>();

            for (int i = 0; i < parameters.Length; ++i) {
                Range range = parameters[i] as Range;
                if (range != null) {
                    List<object> fragment = GetElement(list, range);
                    if (fragment != null)
                        result.Add(fragment);
                } else
                    result.Add(GetElement(list, RubySites.ToInt(parameters[i])));
            }

            return result;
        }

        // BUG? parameters are not splatted into object[], but instead are being passed in as List<object>
        [RubyMethodAttribute("indexes", RubyMethodAttributes.PublicInstance)]
        public static object Indexes(List<object> self, [NotNull] params object[] args) {
            if (args == null) args = RubyOps.NullArray;
            RubyConsole.Warning("Array#indexes is deprecated; use Array#values_at");
            return GetIndicesAsNestedArrays(self, args);
        }

        [RubyMethodAttribute("indices", RubyMethodAttributes.PublicInstance)]
        public static object Indices(List<object> self, [NotNull] params object[] args) {
            if (args == null) args = RubyOps.NullArray;
            RubyConsole.Warning("Array#indices is deprecated; use Array#values_at");
            return GetIndicesAsNestedArrays(self, args);
        }

        public static List<object> Append(List<object> self, params object[] args) {
            self.AddRange(args);
            return self;
        }

        [RubyMethodAttribute("insert", RubyMethodAttributes.PublicInstance)]
        public static List<object> Insert(List<object> self, int index, [NotNull] params object[] args) {
            if (args == null) args = RubyOps.NullArray;

            if (args.Length == 0)
                return self;

            if (index == -1)
                return Append(self, args);

            index = index < 0 ? index + self.Count + 1 : index;
            if (index < 0)
                throw RubyExceptions.CreateIndexError(String.Format("index {0} out of array", index));

            if (index >= self.Count) {
                ExpandList(self, index);
                return Append(self, args);
            }

            List<object> newList = new List<object>();
            for (int i = 0; i < args.Length; ++i)
                newList.Add(args[i]);

            self.InsertRange(index, newList);
            return self;

        }

        [RubyMethodAttribute("inspect", RubyMethodAttributes.PublicInstance)]
        public static MutableString Inspect(List<object> self) {
            MutableString str = Join(self, ", ");
            str.Insert("[", 0);
            str.Append("]");
            return str;
        }

        public static void RecursiveJoin(List<object>/*!*/ list, string separator, MutableString/*!*/ result, Dictionary<object, bool>/*!*/ seen) {
            Utils.Assert.NotNull(list, result, seen);

            for (int i = 0; i < list.Count; ++i) {
                object item = list[i];

                if (item is ValueType) {
                    result.Append(RubySites.ToS(item));
                } else if (item == null) {
                    result.Append("nil");
                } else {
                    if (!seen.ContainsKey(item)) {
                        seen.Add(item, true);
                        List<object> listItem = item as List<object>;
                        if (listItem != null) {
                            result.Append("[");
                            RecursiveJoin(listItem, separator, result, seen);
                            result.Append("]");
                        } else
                            result.Append(RubySites.ToS(item));
                    } else {
                        if (!(item is List<object>)) {
                            result.Append(RubySites.ToS(item));
                        } else {
                            result.Append("[...]");
                        }
                    }
                }

                if (i < list.Count - 1)
                    result.Append(separator);
            }
        }

        [RubyMethodAttribute("join", RubyMethodAttributes.PublicInstance)]
        [RubyMethodAttribute("to_s", RubyMethodAttributes.PublicInstance)]
        public static MutableString/*!*/ Join(CodeContext/*!*/ context, [NotNull]List<object>/*!*/ self) {
            Utils.Assert.NotNull(context, self);

            object separator = RubyOps.GetGlobalVariable(context, Symbols.Comma);
            return Join(self, (separator != null ? separator.ToString() : ""));
        }

        [RubyMethodAttribute("join", RubyMethodAttributes.PublicInstance)]
        public static MutableString Join(List<object>/*!*/ self, string separator) {
            MutableString result = new MutableString();
            Dictionary<object, bool> seen = new Dictionary<object, bool>();
            seen.Add(self, true);
            RecursiveJoin(self, separator, result, seen);
            return result;
        }

        [RubyMethodAttribute("join", RubyMethodAttributes.PublicInstance)]
        public static MutableString Join(List<object>/*!*/ self, [NotNull]MutableString/*!*/ separator) {
            return Join(self, separator.ToString());
        }

        [RubyMethodAttribute("last", RubyMethodAttributes.PublicInstance)]
        public static object Last(List<object>/*!*/ self) {
            return self.Count == 0 ? null : self[self.Count - 1];
        }

        [RubyMethodAttribute("last", RubyMethodAttributes.PublicInstance)]
        public static object Last(List<object>/*!*/ self, int count) {
            if (count < 0)
                throw RubyExceptions.CreateArgumentError("negative array size (or size too big)");

            count = count > self.Count ? self.Count : count;
            return self.GetRange(self.Count - count, count);
        }

        [RubyMethodAttribute("last", RubyMethodAttributes.PublicInstance)]
        public static object Last(List<object> self, object count) {
            return Last(self, RubySites.ToInt(count));
        }

        [RubyMethodAttribute("length", RubyMethodAttributes.PublicInstance)]
        [RubyMethodAttribute("size", RubyMethodAttributes.PublicInstance)]
        public static int Length(List<object> self) {
            return self.Count;
        }

        [RubyMethodAttribute("nitems", RubyMethodAttributes.PublicInstance)]
        public static int NumberOfNonNilItems(List<object> self) {
            int count = 0;
            foreach (object obj in self)
                if (obj != null)
                    count++;
            return count;
        }

        [RubyMethodAttribute("pop", RubyMethodAttributes.PublicInstance)]
        public static object Pop(List<object> self) {
            if (self.Count == 0)
                return null;

            object result = self[self.Count - 1];
            self.RemoveAt(self.Count - 1);
            return result;
        }

        [RubyMethodAttribute("push", RubyMethodAttributes.PublicInstance)]
        public static object Push(List<object> self, [NotNull] params object[] values) {
            if (values == null) values = RubyOps.NullArray;
            self.AddRange(values);
            return self;
        }

        [RubyMethodAttribute("reverse", RubyMethodAttributes.PublicInstance)]
        public static List<object> Reverse(List<object> self) {
            List<object> reversedList = new List<object>(self.Count);
            for (int i = 0; i < self.Count; ++i)
                reversedList.Add(self[self.Count - i - 1]);
            return reversedList;
        }

        [RubyMethodAttribute("reverse!", RubyMethodAttributes.PublicInstance)]
        public static List<object> InPlaceReverse(List<object> self) {
            self.Reverse();
            return self;
        }

        [RubyMethodAttribute("rassoc", RubyMethodAttributes.PublicInstance)]
        public static List<object> GetContainerOfSecondItem(List<object> self, object item) {
            return GetContainerOf(self, 1, item);
        }

        [RubyMethodAttribute("replace", RubyMethodAttributes.PublicInstance)]
        public static List<object> Replace(List<object> self, List<object> other) {
            self.Clear();
            self.AddRange(other);
            return self;
        }

        [RubyMethodAttribute("replace", RubyMethodAttributes.PublicInstance)]
        public static List<object> Replace(List<object> self, object other) {
            if (RubySites.RespondTo(other, "to_ary"))
                return Replace(self, RubySites.ToAry(other));
            else
                throw RubyExceptions.CannotConvertTypeToTargetType(other, "Array");
        }

        [RubyMethodAttribute("rindex", RubyMethodAttributes.PublicInstance)]
        public static object ReverseIndex(List<object> self, object item) {
            for (int i = self.Count - 1; i >= 0; --i) {
                if (RubySites.Equal(self[i], item))
                    return i;
            }
            return null;
        }

        [RubyMethodAttribute("shift", RubyMethodAttributes.PublicInstance)]
        public static object Shift(List<object> self) {
            if (self.Count == 0)
                return null;

            object result = self[0];
            self.RemoveAt(0);
            return result;
        }

        [RubyMethodAttribute("slice!", RubyMethodAttributes.PublicInstance)]
        public static object SliceInPlace(List<object> self, int index) {
            index = index < 0 ? index + self.Count : index;
            if (index >= 0 && index < self.Count) {
                object result = self[index];
                SetElement(self, index, 1, null);
                return result;
            } else
                return null;
        }

        [RubyMethodAttribute("slice!", RubyMethodAttributes.PublicInstance)]
        public static object SliceInPlace(List<object> self, Range range) {
            object result = GetElement(self, range);
            SetElement(self, range, null);
            return result;
        }

        [RubyMethodAttribute("slice!", RubyMethodAttributes.PublicInstance)]
        public static object SliceInPlace(List<object> self, object indexOrRange) {
            if (RubySites.RespondTo(indexOrRange, "to_int")) {
                return SliceInPlace(self, RubySites.ToInt(indexOrRange));
            } else if (RubySites.RespondTo(indexOrRange, "to_range")) {
                return SliceInPlace(self, RubySites.ToRange(indexOrRange));
            } else
                throw RubyExceptions.CannotConvertTypeToTargetType(indexOrRange, "Integer or Range");
        }

        [RubyMethodAttribute("slice!", RubyMethodAttributes.PublicInstance)]
        public static List<object> SliceInPlace(List<object> self, int start, int length) {
            List<object> result = GetElements(self, start, length);
            SetElement(self, start, length, null);
            return result;
        }

        // TODO: implement version of method that takes a block parameter
        [RubyMethodAttribute("sort", RubyMethodAttributes.PublicInstance)]
        public static List<object> Sort(List<object> self) {
            List<object> result = new List<object>(self);
            // TODO: compare using <=> operator
            result.Sort();
            return result;
        }

        [RubyMethodAttribute("sort!", RubyMethodAttributes.PublicInstance)]
        public static List<object> SortInPlace(List<object> self) {
            // TODO: compare using <=> operator
            self.Sort();
            return self;
        }

        [RubyMethodAttribute("to_a", RubyMethodAttributes.PublicInstance)]
        [RubyMethodAttribute("to_ary", RubyMethodAttributes.PublicInstance)]
        public static List<object> ToArray(List<object> self) {
            return self;
        }

        [RubyMethodAttribute("uniq", RubyMethodAttributes.PublicInstance)]
        public static List<object> Unique(List<object> self) {
            List<object> result = new List<object>();

            Dictionary<object, bool> seen = new Dictionary<object, bool>();
            foreach (object item in self) {
                object key = item is MutableString ? item.ToString() : item;
                if (!seen.ContainsKey(key)) {
                    result.Add(item);
                    seen.Add(key, true);
                }
            }

            return result;
        }

        [RubyMethodAttribute("uniq!", RubyMethodAttributes.PublicInstance)]
        public static List<object> UniqueSelf(List<object> self) {
            Dictionary<object, bool> seen = new Dictionary<object, bool>();
            bool modified = false;
            int i = 0;
            while (i < self.Count) {
                object key = self[i] is MutableString ? self[i].ToString() : self[i];
                if (!seen.ContainsKey(key)) {
                    seen.Add(key, true);
                    i++;
                } else {
                    self.RemoveAt(i);
                    modified = true;
                }
            }

            return modified ? self : null;
        }

        [RubyMethodAttribute("unshift", RubyMethodAttributes.PublicInstance)]
        public static List<object> UnShift(List<object> self, [NotNull] params object[] args) {
            if (args == null) args = RubyOps.NullArray;
            self.InsertRange(0, args);
            return self;
        }

        [RubyMethodAttribute("values_at", RubyMethodAttributes.PublicInstance)]
        public static List<object> ValuesAt(List<object> list, [NotNull] params object[] values) {
            if (values == null) values = RubyOps.NullArray;
            List<object> result = new List<object>();

            for (int i = 0; i < values.Length; ++i) {
                Range range = values[i] as Range;
                if (range != null) {
                    List<object> fragment = GetElement(list, range);
                    if (fragment != null)
                        result.AddRange(fragment);
                } else
                    result.Add(GetElement(list, RubySites.ToInt(values[i])));
            }

            return result;
        }
    }
}