/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;
using Ruby.Runtime;
using System.Diagnostics;
using Microsoft.Scripting;
using Microsoft.Scripting.Actions;
using MSA = Microsoft.Scripting.Ast;
using Ast = Microsoft.Scripting.Ast.Ast;

using Ruby.Hosting;
using Ruby.Compiler;
using System.Reflection;
using System.Threading;

namespace Ruby.Builtins {

    [RubyClassAttribute("Class")]
    public sealed class RubyClass : RubyModule, IDynamicObject {

        #region Implementation

        private readonly RubyClass _superClass;

        private readonly Dictionary<SymbolId, object>/*!*/ _instanceVariables;
        private readonly Dictionary<SymbolId, object>/*!*/ _classVariables;
        
        private readonly bool _isProxy;
        private readonly bool _underlyingTypeMembersVisible;
        private readonly Type/*!*/ _underlyingSystemType;
        private Delegate[] _factories;

        public RubyClass SuperClass {
            get { return _superClass; } 
        }

        public bool IsProxy {
            get { return _isProxy; }
        }

        public override bool IsClass {
            get { return true; }
        }

        public Type/*!*/ UnderlyingSystemType {
            get { return _underlyingSystemType; }
        }

        private PropertyInfo/*!*/ UnderlyingSystemTypeProperty {
            get { return typeof(RubyClass).GetProperty("UnderlyingSystemType"); }
        }

        public Delegate[] Factories {
            get { return _factories; }
            set { _factories = value; }
        }

        // TODO:
        public bool ShowUnderlyingTypeMembers {
            get { return _underlyingTypeMembersVisible; }
        }

        internal RubyClass(RubyExecutionContext/*!*/ context, SymbolId name, Type/*!*/ type, Action<RubyModule> initializer, 
            RubyClass superClass, bool isProxy, bool underlyingTypeMembersVisible)
            : base(context, name, initializer) {

            Utils.Assert.NotNull(name);
            Debug.Assert((superClass == null) == (type == typeof(object)));

            _underlyingSystemType = type;
            _superClass = superClass;

            if (_superClass != null) {
                _superClass.AddDependentModule(this);
            }
            
            _instanceVariables = new Dictionary<SymbolId, object>();
            _classVariables = new Dictionary<SymbolId, object>();
            _isProxy = isProxy;
            _underlyingTypeMembersVisible = underlyingTypeMembersVisible;
        }

        public bool IsSubclassOf(RubyClass/*!*/ super) {
            Utils.Assert.NotNull(super);

            RubyClass c = this;
            do {
                if (c == super) return true;
                c = c.SuperClass;
            } while (c != null);

            return false;
        }

        public bool HasAncestor(RubyModule/*!*/ module) {
            RubyClass c = this;
            do {
                if (c == module) return true;
                if (Array.IndexOf(c.Mixins, module) != -1) return true;
                
                c = c.SuperClass;
            } while (c != null);

            return false;
        }

        public bool IsException() {
            return IsSubclassOf(ExecutionContext.ExceptionClass);
        }

        public sealed override RubyMethodBase ResolveMethod(SymbolId name, bool isInstance) {
            RubyClass c = this;
            do {
                RubyMethodBase method = c.ResolveDeclaredMethod(name, isInstance);
                if (method != null) {
                    return method;
                }

                c = c._superClass;
            } while (c != null);

            return null;
        }

        public sealed override bool TryResolveConstant(SymbolId name, out object value) {
            RubyClass c = this;
            do {
                if (c.TryGetDeclaredConstant(name, out value)) {
                    return true;
                }

                c = c._superClass;
            } while (c != null);

            return false;
        }

        public sealed override void ForEachMethod(Action<RubyMethodBase>/*!*/ action, bool instance) {
            Utils.Assert.NotNull(action);
            for (RubyClass c = this; c != null; c = c._superClass) {
                c.ForEachDeclaredMethod(action, instance);
            }
        }

        public sealed override void ForEachConstant(Action<SymbolId, object>/*!*/ action) {
            Utils.Assert.NotNull(action);
            for (RubyClass c = this; c != null; c = c._superClass) {
                c.ForEachDeclaredConstant(action);
            }
        }

        #endregion

        #region Private Instance Methods

        //  inherited
        //  initialize (TODO: [RubyConstructor]
        //  initialize_copy

        #endregion

        #region Public Instance Methods

        [RubyMethodAttribute("allocate", RubyMethodAttributes.PublicInstance | RubyMethodAttributes.IsInstanceAllocator)]
        public static object Allocate(RubyClass/*!*/ self) {
            if (self == null) throw new ArgumentNullException("self");
            // TODO: dynamic allocation
            //return self.CreateDefaultInstance();
            throw Utils.Assert.Unreachable;
        }

        [RubyMethodAttribute("new", RubyMethodAttributes.PublicInstance | RubyMethodAttributes.IsInstanceConstructor)]
        public static object New(RubyClass/*!*/ self, [NotNull]params object[] args) {
            if (self == null) throw new ArgumentNullException("self");
            if (args == null) args = RubyOps.NullArray;
            // TODO: dynamic construction
            //return self.CreateInstance(args);
            throw Utils.Assert.Unreachable;
        }

        [RubyMethodAttribute("inspect", RubyMethodAttributes.PublicInstance)]
        [RubyMethodAttribute("to_s", RubyMethodAttributes.PublicInstance)]
        [RubyMethodAttribute("name", RubyMethodAttributes.PublicInstance)]
        public static MutableString/*!*/ GetName(RubyClass/*!*/ self) {
            if (self == null) throw new ArgumentNullException("self");
            return new MutableString(SymbolTable.IdToString(self.Name));
        }

        [RubyMethodAttribute("inspect", RubyMethodAttributes.PublicInstance)]
        [RubyMethodAttribute("to_s", RubyMethodAttributes.PublicInstance)]
        [RubyMethodAttribute("name", RubyMethodAttributes.PublicInstance)]
        public static MutableString/*!*/ GetName(DynamicType/*!*/ self) { // TODO
            if (self == null) throw new ArgumentNullException("self");
            return new MutableString(self.UnderlyingSystemType.Name);
        }

        [RubyMethodAttribute("superclass", RubyMethodAttributes.PublicInstance)]
        public static RubyClass GetSuperclass(RubyClass/*!*/ self) {
            return self.SuperClass;
        }

        [RubyMethodAttribute("superclass", RubyMethodAttributes.PublicInstance)]
        public static DynamicType GetSuperclass(DynamicType/*!*/ self) { // TODO:
            return self.BaseTypes.Count > 0 ? self.BaseTypes[0] : null;
        }

        #endregion

        #region IDynamicObject Members

        public override StandardRule<T>/*!*/ GetRule<T>(Action/*!*/ action, CodeContext/*!*/ callerContext, object[]/*!*/ args) {

            switch (action.Kind) {
                case ActionKind.CreateInstance:
                    StandardRule<T> result = new StandardRule<T>();
                    this.SetRuleForCreateInstance<T>(result, callerContext.LanguageContext.Binder, ((CallAction)action).ArgumentInfos, args, true, true);
                    return result;

                default:
                    return base.GetRule<T>(action, callerContext, args);
            }
        }

        internal void SetRuleForCreateInstance<T>(StandardRule<T>/*!*/ result, ActionBinder/*!*/ binder, ArgumentInfo[] argInfos,
            object[]/*!*/ args, bool invokeInitialize, bool testTarget) {
            Utils.Assert.NotNull(result, binder, args);

            if (testTarget) {
                // check for reference-equality to this RubyClass instance and its version:
                result.AddTest(Ast.Equal(result.Parameters[0], Ast.RuntimeConstant(this)));
                AddVersionTest(result);

            }

            // TODO: could we inline these sites?
            MSA.Expression targetExpr;

            if (_isProxy) {
                // TODO: ShowUnderlying...
                MethodBase[] overloads = (_factories != null) ? (MethodBase[])Utils.Reflection.GetMethodInfos(_factories) : _underlyingSystemType.GetConstructors();
                RubyMethodGroupInfo.SetInvocationRuleInternal<T>("ctor", overloads, binder, result,
                    Utils.Array.RemoveFirst(argInfos), Utils.Array.RemoveFirst(args), Utils.Array.RemoveFirst(result.Parameters), null);

                // TODO:
                if (!(result.Target is MSA.ReturnStatement)) {
                    Debug.Assert(result.Target != null && result.Test != null);
                    return;
                }
                
                targetExpr = ((MSA.ReturnStatement)result.Target).Expression;

            } else {
                
                // Ruby user class (always has a default ctor):
                targetExpr = Ast.New(_underlyingSystemType.GetConstructor(Type.EmptyTypes));
            }

            if (invokeInitialize) {
                RubyMethodBase initialize = this.ResolveMethod(Symbols.Initialize, true);
                if (initialize != null) {
                    MSA.Variable resultVariable = result.GetTemporary(targetExpr.ExpressionType, "#result");

                    MSA.Expression[] ps = Utils.Array.MakeArray(result.Parameters, 0, 0);
                    ps[0] = Ast.Read(resultVariable);

                    targetExpr = Ast.Comma(2,
                        Ast.Assign(resultVariable, targetExpr),
                        Ast.Action.InvokeMember(Symbols.Initialize, typeof(object), InvokeMemberActionFlags.IsCallWithThis, argInfos, ps),
                        Ast.Read(resultVariable)
                    );
                }
            }

            result.SetTarget(result.MakeReturn(binder, targetExpr));

            Debug.Assert(result.Target != null && result.Test != null);
        }

        #endregion
    }
}
