/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;
using Ruby.Runtime;
using Microsoft.Scripting;
using System.Diagnostics;
using System.Threading;
using Microsoft.Scripting.Actions;
using MSA = Microsoft.Scripting.Ast;
using System.Collections;
using System.Runtime.InteropServices;

namespace Ruby.Builtins {

    // Exception
    // -- fatal
    // -- NoMemoryError                                 
    // -- ScriptError
    // ---- LoadError
    // ---- NotImplementedError
    // ---- SyntaxError
    // -- SignalException
    // ---- Interrupt
    // -- StandardError
    // ---- ArgumentError
    // ---- IOError
    // ------ EOFError
    // ---- IndexError
    // ---- LocalJumpError
    // ---- NameError
    // ------ NoMethodError
    // ---- RangeError
    // ------ FloatDomainError
    // ---- RegexpError
    // ---- RuntimeError
    // ---- SecurityError
    // ---- SystemCallError
    // ------ system-dependent-exceptions Errno::XXX
    // ---- ThreadError
    // ---- TypeError
    // ---- ZeroDivisionError
    // -- SystemExit
    // -- SystemStackError
    [RubyExtensionClass("Exception", typeof(Exception))]
    public static class ExceptionOps {

        [Serializable]
        private sealed class ExceptionData {
            public MutableString Message;
            public List<object> Backtrace;
            public bool HasBacktrace;
        }

        private static readonly object/*!*/ _DataKey = new object();

        private static ExceptionData/*!*/ GetOrCreateExceptionData(Exception/*!*/ e) {
            Utils.Assert.NotNull(e);

            IDictionary dict = Utils.GetDataDictionary(e);
            ExceptionData result = dict[_DataKey] as ExceptionData;
            if (result == null) {
                dict[_DataKey] = result = new ExceptionData();
                result.Message = new MutableString(e.Message);
            }
            return result;
        }

        /// <summary>
        /// Creates an instance of a specified exception class.
        /// </summary>
        internal static Exception/*!*/ MakeException(CodeContext/*!*/ context, RubyClass/*!*/ exceptionClass, MutableString message, List<object> backtrace) {
            if (exceptionClass == null) {
                throw RubyExceptions.CreateTypeError("exception object expected");
            }

            if (!exceptionClass.IsException()) {
                throw RubyExceptions.CreateTypeError("exception class/object expected");
            }

            Exception result = CreateException(context, exceptionClass, message);
            SetBacktrace(result, backtrace);
            return result;
        }

        public static void SetMessage(Exception/*!*/ self, MutableString message) {
            GetOrCreateExceptionData(self).Message = message;
        }

        #region Private Instance Methods

        [RubyConstructor]
        public static Exception/*!*/ CreateException([Optional]MutableString message) {
            Exception result = new Exception((message != null) ? message.ToString() : null);
            SetMessage(result, message);
            return result;
        }

        [RubyMethod("initialize", RubyMethodAttributes.PrivateInstance)]
        public static Exception/*!*/ Initialize(Exception/*!*/ self) {
            Utils.Assert.NotNull(self);

            IDictionary dict = Utils.GetDataDictionary(self);
            ExceptionData data = dict[_DataKey] as ExceptionData;
            if (self.Message != null) {
                dict[_DataKey] = new ExceptionData();
            } else if (data != null) {
                dict.Remove(_DataKey);
            }
            
            return self;
        }

        [RubyMethod("initialize", RubyMethodAttributes.PrivateInstance)]
        public static Exception/*!*/ Initialize(Exception/*!*/ self, MutableString message) {
            Utils.Assert.NotNull(self);

            ExceptionData data = GetOrCreateExceptionData(self);

            data.Message = message;
            data.HasBacktrace = false;
            data.Backtrace = null;

            return self;
        }

        #endregion

        #region Public Instance Methods

        [RubyMethod("backtrace", RubyMethodAttributes.PublicInstance)]
        public static List<object>/*!*/ GetBacktrace(Exception/*!*/ self) {
            Utils.Assert.NotNull(self);

            ExceptionData data = GetOrCreateExceptionData(self);
            if (data.HasBacktrace) {
                return data.Backtrace;
            }

            StackTrace trace = new StackTrace(self);
            List<object> result = new List<object>(trace.FrameCount);

            // TODO:
            for (int i = 0; i < trace.FrameCount; i++) {
                StackFrame frame = trace.GetFrame(i);
                result.Add(new MutableString(String.Format("{0}:{1}:in `{2}'", frame.GetFileName(), frame.GetFileLineNumber(), frame.GetMethod().Name)));
            }

            data.HasBacktrace = true;
            data.Backtrace = result;
            return result;
        }

        [RubyMethod("set_backtrace", RubyMethodAttributes.PublicInstance)]
        public static List<object> SetBacktrace(Exception/*!*/ self, [NotNull]MutableString/*!*/ backtrace) {
            Utils.Assert.NotNull(backtrace);
            return SetBacktrace(self, Utils.Collections.MakeList<object>(backtrace));
        }
        
        [RubyMethod("set_backtrace", RubyMethodAttributes.PublicInstance)]
        public static List<object> SetBacktrace(Exception/*!*/ self, List<object> backtrace) {
            Utils.Assert.NotNull(self);

            if (backtrace != null && !Utils.Collections.TrueForAll(backtrace, delegate(object item) { return item is MutableString; })) {
                throw RubyExceptions.CreateTypeError("backtrace must be Array of String");
            }

            ExceptionData data = GetOrCreateExceptionData(self);
            data.HasBacktrace = true;
            data.Backtrace = backtrace;
            return backtrace;
        }

        [RubyMethod("exception", RubyMethodAttributes.PublicInstance)]
        public static Exception/*!*/ GetException(Exception/*!*/ self) {
            Utils.Assert.NotNull(self);
            return self;
        }

        [RubyMethod("exception", RubyMethodAttributes.PublicInstance)]
        public static Exception/*!*/ GetException(CodeContext/*!*/ context, Exception/*!*/ self, MutableString message) {
            Utils.Assert.NotNull(context, self);
            return MakeException(context, RubyOps.GetExecutionContext(context).GetClassOf(self), message, null);
        }

        [RubyMethod("to_s", RubyMethodAttributes.PublicInstance)]
        [RubyMethod("message", RubyMethodAttributes.PublicInstance)]
        public static MutableString GetMessage(Exception/*!*/ self) {
            Utils.Assert.NotNull(self);

            IDictionary dict = Utils.GetDataDictionary(self);
            ExceptionData data = dict[_DataKey] as ExceptionData;
            if (data != null) {
                return data.Message;
            }

            dict[_DataKey] = data = new ExceptionData();
            return data.Message = new MutableString(self.Message);
        }

        #endregion

        #region Singleton methods

        private static DynamicSite<RubyClass, Exception> NewCallSite1 = DynamicSite<RubyClass, Exception>.Create(
            new InvokeMemberAction(SymbolTable.StringToId("new"), InvokeMemberActionFlags.IsCallWithThis,
                new ArgumentInfo[] { new ArgumentInfo(MSA.ArgumentKind.Instance) }));
        
        [RubyMethod("exception", RubyMethodAttributes.PublicSingleton)]
        public static Exception/*!*/ CreateException(CodeContext/*!*/ context, RubyClass/*!*/ self) {
            return NewCallSite1.Invoke(context, self);
        }

        private static DynamicSite<RubyClass, MutableString, Exception> NewCallSite2 = DynamicSite<RubyClass, MutableString, Exception>.Create(
            new InvokeMemberAction(SymbolTable.StringToId("new"), InvokeMemberActionFlags.IsCallWithThis,
                new ArgumentInfo[] { new ArgumentInfo(MSA.ArgumentKind.Instance), new ArgumentInfo(MSA.ArgumentKind.Simple) }));

        [RubyMethod("exception", RubyMethodAttributes.PublicSingleton)]
        public static Exception/*!*/ CreateException(CodeContext/*!*/ context, RubyClass/*!*/ self, MutableString message) {
            return NewCallSite2.Invoke(context, self, message);
        }

        #endregion
    }
}
