/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;

using Microsoft.Scripting;
using Microsoft.Scripting.Actions;
using MSA = Microsoft.Scripting.Ast;
using Ast = Microsoft.Scripting.Ast.Ast;

using Ruby.Hosting;
using Ruby.Compiler;
using Ruby.Runtime;
using System.Threading;
using System.Reflection;

namespace Ruby.Builtins {

    [RubyClass("Module")]
    public class RubyModule : Object, IDynamicObject {

        public static readonly RubyModule[]/*!*/ EmptyArray = new RubyModule[0];

        private enum State {
            Uninitialized,
            Initializing,
            Initialized
        }

        #region Implementation

        private static int _globalVersion = 0;

        // TODO: more versions: constants, methods, etc.
        private int _version;

        private readonly RubyExecutionContext/*!*/ _executionContext;

        // null means lookup the execution context's global namespace 
        private Dictionary<SymbolId, object> _constants;

        private Dictionary<SymbolId, RubyMethodBase> _instanceMethods;
        private Dictionary<SymbolId, RubyMethodBase> _classMethods;

        private RubyModule[]/*!*/ _mixins;

        private SymbolId _name;
        private State _state;
        private Action<RubyModule> _initializer;

        /// <summary>
        /// A list of dependent modules. Forms a DAG.
        /// Class dependencies are associated with uncollectible Types so we don't need to make them weak referenced.
        /// TODO: Module dependencies could be potentially made weak.
        /// </summary>
        private List<RubyModule>/*!*/ _dependentModules = new List<RubyModule>();

        public int Version {
            get { return _version; }
        }

        internal static PropertyInfo/*!*/ VersionProperty {
            get { return typeof(RubyModule).GetProperty("Version"); }
        }

        public SymbolId Name {
            get { return _name; }
        }

        internal RubyExecutionContext/*!*/ ExecutionContext {
            get { return _executionContext; }
        }

        internal RubyModule[] Mixins {
            get { return _mixins; }
        }

        public virtual bool IsClass {
            get { return false; }
        }

        internal RubyModule(RubyExecutionContext/*!*/ context, SymbolId name, Action<RubyModule> initializer) {
            Utils.Assert.NotNull(context);

            _executionContext = context;
            _name = name;

            _state = State.Uninitialized;
            _mixins = EmptyArray;
            _initializer = initializer;
            _version = Interlocked.Increment(ref _globalVersion);
        }

        private void InitializeMembers() {
            Debug.Assert(_state == State.Uninitialized && _constants == null && _instanceMethods == null && _classMethods == null);
            Debug.Assert(_executionContext.ObjectClass != null, "ObjectClass should already be initialized");

            if (ReferenceEquals(this, _executionContext.ObjectClass)) {
                _constants = null;                
            } else {
                _constants = new Dictionary<SymbolId, object>();
            }

            _instanceMethods = new Dictionary<SymbolId, RubyMethodBase>();
            _classMethods = new Dictionary<SymbolId, RubyMethodBase>();

            _state = State.Initializing;

            if (_initializer != null) {
                _initializer(this);
            }

            _state = State.Initialized;
        }

        internal void AddDependentModule(RubyModule/*!*/ dependentModule) {
            Utils.Assert.NotNull(dependentModule);
            _dependentModules.Add(dependentModule);
        }

        internal void SetDependency(IList<RubyModule>/*!*/ modules) {
            for (int i = 0; i < modules.Count; i++) {
                modules[i].AddDependentModule(this);
            }
        }

        // TODO: optimize
        internal void Updated() {
            _version = Interlocked.Increment(ref _globalVersion);
            
            for (int i = 0; i < _dependentModules.Count; i++) {
                _dependentModules[i].Updated();
            }
        }

        public override string/*!*/ ToString() {
            return SymbolTable.IdToString(_name);
        }

        #region Constants

        private bool TryGetConstant(SymbolId name, out object value) {
            if (_state != State.Initialized) {
                InitializeMembers();
            }

            if (_constants == null) {
                return _executionContext.GlobalNamespaceDictionary.TryGetCustomMember(_executionContext.Engine.DefaultContext, name, out value);   
            }

            return _constants.TryGetValue(name, out value);
        }
        
        public void SetConstant(SymbolId name, object value) {
            if (_state != State.Initialized) {
                InitializeMembers();
            }

            if (_constants == null) {
                _executionContext.GlobalNamespaceDictionary.SetCustomMember(_executionContext.Engine.DefaultContext, name, value);
            } else {
                _constants[name] = value;
            }
            Updated();
        }

        public void SetConstant(string/*!*/ name, object value) {
            SetConstant(SymbolTable.StringToId(name), value);
        }

        protected bool TryGetDeclaredConstant(SymbolId name, out object value) {
            if (TryGetConstant(name, out value)) {
                return true;
            }

            // all modules are included in the list (transitively):
            for (int i = 0; i < _mixins.Length; i++) {
                if (_mixins[i].TryGetConstant(name, out value)) {
                    return true;
                }
            }
            
            return false;
        }

        public virtual bool TryResolveConstant(SymbolId name, out object value) {
            return TryGetDeclaredConstant(name, out value);
        }

        private void ForEachConstantInternal(Action<SymbolId, object>/*!*/ action) {
            if (_state != State.Initialized) {
                InitializeMembers();
            }

            if (_constants != null) {
                foreach (KeyValuePair<SymbolId, object> constant in _constants) {
                    action(constant.Key, constant.Value);
                }
            } else {
                foreach (KeyValuePair<object, object> constant in _executionContext.GlobalNamespaceDictionary.GetCustomMemberDictionary(_executionContext.Engine.DefaultContext)) {
                    if (constant.Key is SymbolId) {
                        action((SymbolId)constant.Key, constant.Value);
                    }
                }
            }
        }

        protected void ForEachDeclaredConstant(Action<SymbolId, object>/*!*/ action) {
            Utils.Assert.NotNull(action);

            ForEachConstantInternal(action);

            for (int i = 0; i < _mixins.Length; i++) {
                _mixins[i].ForEachConstantInternal(action);
            }
        }

        public virtual void ForEachConstant(Action<SymbolId, object>/*!*/ action) {
            ForEachDeclaredConstant(action);
        }

        #endregion

        #region Methods

        internal void SetClassMethod(RubyMethodBase/*!*/ method) {
            if (_state == State.Uninitialized) {
                InitializeMembers();
            }

            _classMethods[method.Name] = method;
            Updated();
        }

        internal void SetInstanceMethod(RubyMethodBase/*!*/ method) {
            if (_state == State.Uninitialized) {
                InitializeMembers();
            }

            _instanceMethods[method.Name] = method;
            Updated();
        }

        public RubyMethodBase ResolveMethod(SymbolId name) {
            RubyMethodBase method = this.ResolveMethod(name, false);
            if (method != null) {
                return method;
            }
            return _executionContext.ClassClass.ResolveMethod(name, true);
        }
        
        // TODO: should resolve CLR methods as well
        public virtual RubyMethodBase ResolveMethod(SymbolId name, bool instance) {
            return ResolveDeclaredMethod(name, instance);
        }

        /// <summary>
        /// Resolves declared or mixed-in methods on this class.
        /// </summary>
        protected RubyMethodBase ResolveDeclaredMethod(SymbolId name, bool instance) {
            RubyMethodBase method;

            if (GetMethodTable(instance).TryGetValue(name, out method)) {
                return method;
            }

            // all modules are included in the list (transitively):
            for (int i = 0; i < _mixins.Length; i++) {
                if (_mixins[i].GetMethodTable(instance).TryGetValue(name, out method)) {
                    return method;
                }
            }

            return null;
        }

        protected void ForEachDeclaredMethod(Action<RubyMethodBase>/*!*/ action, bool instance) {
            Utils.Assert.NotNull(action);
            
            foreach (RubyMethodBase method in GetMethodTable(instance).Values) {
                action(method);
            }

            for (int i = 0; i < _mixins.Length; i++) {
                foreach (RubyMethodBase method in _mixins[i].GetMethodTable(instance).Values) {
                    action(method);
                }
            }
        }

        public virtual void ForEachMethod(Action<RubyMethodBase>/*!*/ action, bool instance) {
            ForEachDeclaredMethod(action, instance);
        }

        private Dictionary<SymbolId, RubyMethodBase>/*!*/ GetMethodTable(bool instance) {
            if (_state != State.Initialized) {
                InitializeMembers();
            }
            return (instance ? _instanceMethods : _classMethods);
        }

        public void DefineInstanceMethod(string/*!*/ name, RubyMethodVisibility visibility, params Delegate[]/*!*/ overloads) {
            DefineInstanceMethod(name, visibility, MethodKind.Regular, overloads);
        }

        public void DefineInstanceMethod(string/*!*/ name, RubyMethodVisibility visibility, MethodKind kind, params Delegate[]/*!*/ overloads) {
            Utils.Assert.NotNull(name);
            Utils.Assert.NotNullItems(overloads);

            SetInstanceMethod(new RubyMethodGroupInfo(SymbolTable.StringToId(name), overloads, visibility, kind));
        }
        
        public void DefineClassMethod(string/*!*/ name, RubyMethodVisibility visibility, params Delegate[]/*!*/ overloads) {
            DefineClassMethod(name, visibility, MethodKind.Regular, overloads);
        }
        
        public void DefineClassMethod(string/*!*/ name, RubyMethodVisibility visibility, MethodKind kind, params Delegate[]/*!*/ overloads) {
            Utils.Assert.NotNull(name);
            Utils.Assert.NotNullItems(overloads);

            SetClassMethod(new RubyMethodGroupInfo(SymbolTable.StringToId(name), overloads, visibility, kind));
        }

        #endregion

        internal void SetMixins(RubyModule[]/*!*/ modules) {
            Utils.Assert.NotNull(modules);
            Debug.Assert(_mixins.Length == 0);
            Debug.Assert(Utils.Collections.TrueForAll(modules, delegate(RubyModule m) { return !m.IsClass; }));

            List<RubyModule> allModules = new List<RubyModule>();
            GetMixinsRecursive(allModules, modules);
            SetDependency(modules);

            _mixins = allModules.ToArray();
            Updated();
        }
        
        internal void IncludeModule(RubyModule/*!*/ module) {
            Utils.Assert.NotNull(module);
            Debug.Assert(!module.IsClass);
            
            // transitively include modules:
            List<RubyModule> modules = new List<RubyModule>();
            modules.Add(module);
            GetMixinsRecursive(modules, module._mixins);
            SetDependency(modules);
            _mixins = Utils.Array.AppendRange(_mixins, modules, 0);

            Updated();
        }

        internal static void GetMixinsRecursive(List<RubyModule>/*!*/ result, IList<RubyModule>/*!*/ modules) {
            for (int i = 0; i < modules.Count; i++) {
                result.Add(modules[i]);
                GetMixinsRecursive(result, modules[i]._mixins);
            }
        }

        #endregion

        #region IDynamicObject Members

        public virtual StandardRule<T>/*!*/ GetRule<T>(Action/*!*/ action, CodeContext/*!*/ callerContext, object[]/*!*/ args) {

            switch (action.Kind) {
                case ActionKind.InvokeMember:
                    return RubyActionBinder.MakeRuleForInvokeMember<T>((InvokeMemberAction)action, callerContext, args);

                default:
                    return null;
            }
        }

        public LanguageContext/*!*/ LanguageContext {
            get { return RubyEngine.CurrentEngine.DefaultContext.LanguageContext; }
        }

        internal void AddVersionTest<T>(StandardRule<T>/*!*/ rule) {
            Utils.Assert.NotNull(rule);
            rule.AddTest(Ast.Equal(Ast.ReadProperty(Ast.RuntimeConstant(this), RubyModule.VersionProperty), Ast.Constant(_version)));
        }
        
        #endregion

        #region Private Instance Methods

        //  alias_method

        [RubyMethod("append_features", RubyMethodAttributes.PrivateInstance)]
        public static RubyModule/*!*/ AppendFeatures(RubyModule/*!*/ self, RubyModule/*!*/ includedModule) {
            if (includedModule == null) throw new ArgumentNullException("includedModule");
            includedModule.IncludeModule(self);
            return self;
        }
          //  attr
          //  attr_accessor
          //  attr_reader
          //  attr_writer
          //  class_variable_get
          //  class_variable_set
          //  define_method
          //  extend_object
          //  extended

        [RubyMethod("include", RubyMethodAttributes.PrivateInstance)]
        public static RubyModule/*!*/ Include(RubyModule/*!*/ self, [NotNull]params RubyModule[]/*!*/ modules) {
            Utils.Assert.NotNull(self, modules);
            for (int i = 0; i < modules.Length; i++) {
                if (modules[i].IsClass) throw RubyExceptions.CreateTypeError("wrong argument type Class (expected Module)");
            }

            for (int i = modules.Length - 1; i >= 0; i--) {
                RubySites.ModuleAppendFeatures(modules[i], self);
                RubySites.ModuleIncluded(modules[i], self);
            }

            return self;
        }

        [RubyMethod("included", RubyMethodAttributes.PrivateInstance)]
        public static RubyModule/*!*/ Included(RubyModule/*!*/ self, RubyModule/*!*/ includedModule) {
            // nop
            return self;
        }

        //  initialize
          //  initialize_copy
          //  method_added
          //  method_removed
          //  method_undefined
          //  module_function

        [RubyMethod("private", RubyMethodAttributes.PrivateInstance)]
        public static RubyModule/*!*/ SetPrivateVisibility(CodeContext/*!*/ context, RubyModule/*!*/ self, [NotNull]params SymbolId[]/*!*/ methodNames) {
            return SetVisibility(context, self, methodNames, RubyMethodVisibility.Private);
        }

        [RubyMethod("protected", RubyMethodAttributes.PrivateInstance)]
        public static RubyModule/*!*/ SetProtectedVisibility(CodeContext/*!*/ context, RubyModule/*!*/ self, [NotNull]params SymbolId[]/*!*/ methodNames) {
            return SetVisibility(context, self, methodNames, RubyMethodVisibility.Protected);
        }

        [RubyMethod("public", RubyMethodAttributes.PrivateInstance)]
        public static RubyModule/*!*/ SetPublicVisibility(CodeContext/*!*/ context, RubyModule/*!*/ self, [NotNull]params SymbolId[]/*!*/ methodNames) {
            return SetVisibility(context, self, methodNames, RubyMethodVisibility.Public);
        }

        private static RubyModule/*!*/ SetVisibility(CodeContext/*!*/ context, RubyModule/*!*/ self, SymbolId[]/*!*/ methodNames, RubyMethodVisibility visibility) {
            if (context == null) throw new ArgumentNullException("context");
            if (self == null) throw new ArgumentNullException("self");
            if (methodNames == null) throw new ArgumentNullException("methodNames");

            if (methodNames.Length > 0) {
                // TODO:
                foreach (SymbolId methodName in methodNames) {
                    self.ResolveMethod(methodName, true).Visibility = visibility;
                    self.ResolveMethod(methodName, false).Visibility = visibility;
                }
            } else {
                ((RubyContext)context.LanguageContext).CurrentVisibility = visibility;
            }
            return self;
        }
        
          //  remove_class_variable
          //  remove_const
          //  remove_method
          //  undef_method
        
        #endregion

        #region Public Instance Methods

          //  <
          //  <=
          //  <=>
          //  ==

        [RubyMethodAttribute("===")]
        public static bool CaseEquals(RubyModule/*!*/ self, object other) {
            return self.ExecutionContext.GetClassOf(other).HasAncestor(self);
        }

          //  >
          //  >=
          //  ancestors
          //  autoload
          //  autoload?

        #region class_eval

        [RubyMethod("class_eval")]
        public static object Evaluate(object self, string code) {
            return null;
        }

        [RubyMethod("class_eval")]
        public static object Evaluate(object self, string code, string fileName) {
            return null;
        }

        [RubyMethod("class_eval")]
        public static object Evaluate(object self, string code, string fileName, int lineNumber) {
            return null;
        }

        [RubyMethod("class_eval")]
        public static object Evaluate(object self, [Block]object block) {
            return null;
        }

        #endregion

          //  class_variables
          //  const_defined?
          //  const_get
          //  const_missing
          //  const_set

        [RubyMethodAttribute("constants")]
        public static List<object>/*!*/ GetDefinedConstants(RubyModule/*!*/ self) {
            List<object> result = new List<object>();
            self.ForEachConstant(delegate(SymbolId name, object value) {
                result.Add(name);
            });
            return result;
        }

          //  freeze
          //  include?
          //  included_modules

        [RubyMethodAttribute("instance_method")]
        public static UnboundMethod/*!*/ GetInstanceMethod(RubyModule/*!*/ self, SymbolId name) {
            RubyMethodBase method = self.ResolveMethod(name, true);
            if (method == null) {
                throw RubyExceptions.CreateNameError(String.Format("undefined method `{0}' for class `{1}'", name, self.Name));
            }

            return new UnboundMethod(method);
        }

          //  method_defined?
          //  module_eval
          //  name
          //  private_class_method
          //  private_instance_methods
          //  private_method_defined?
          //  protected_instance_methods
          //  protected_method_defined?
          //  public_class_method
          //  public_instance_methods
          //  public_method_defined?
          //  to_s

        #endregion

        #region Singleton Methods
        
        //  constants
        //  nesting

        #endregion
    }
}
