/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;
using Ruby.Runtime;
using Ruby.Compiler;
using Microsoft.Scripting;
using Microsoft.Scripting.Actions;
using Ruby.Hosting;
using System.Runtime.InteropServices;

namespace Ruby.Builtins {

    [RubyClass, Serializable]
    public class MutableString {
        // StringBuilder will eventually be replaced by a byte[]
        private StringBuilder _data;

        private int CalculateNextPowerOfTwo(int v) {
            v--;
            v |= v >> 1;
            v |= v >> 2;
            v |= v >> 4;
            v |= v >> 8;
            v |= v >> 16;
            return v++;
        }

        public override int GetHashCode() {
            return _data.GetHashCode();
        }

        public static implicit operator string(MutableString/*!*/ self) {
            return self.ToString();
        }

        public MutableString(string str) {
            _data = new StringBuilder(str);
        }

        public MutableString(MutableString str) {
            _data = new StringBuilder(str != null ? str.ToString() : null);
        }

        public MutableString(string str, int start, int length) {
            _data = new StringBuilder(str, start, length, length > 0 ? CalculateNextPowerOfTwo(length) : 0);
        }

        public MutableString() {
            _data = new StringBuilder();
        }

        public char this[int index] {
            get { return _data[index]; }
            set { _data[index] = value; }
        }

        public StringBuilder Chars {
            get { return _data; }
        }

        public MutableString Append(string str) {
            _data.Append(str);
            return this;
        }

        public MutableString Append(char c) {
            _data.Append(c);
            return this;
        }

        public MutableString Append(MutableString str) {
            _data.Append(str.Chars);
            return this;
        }

        public MutableString Append(object obj) {
            _data.Append(obj.ToString());
            return this;
        }

        public MutableString Insert(string str, int position) {
            _data.Insert(position, str);
            return this;
        }

        public MutableString Insert(MutableString str, int position) {
            // TODO: optimize
            _data.Insert(position, str.ToString());
            return this;
        }

        // TODO: write a real implementation instead of converting to string first and then calling String.IndexOf()
        public int IndexOf(char c) {
            return _data.ToString().IndexOf(c);
        }

        public int IndexOf(char c, int index) {
            return _data.ToString().IndexOf(c, index);
        }

        public int IndexOf(string str) {
            return _data.ToString().IndexOf(str);
        }

        public int IndexOf(MutableString str) {
            return _data.ToString().IndexOf(str.ToString());
        }

        public int IndexOf(string str, int offset) {
            return _data.ToString().IndexOf(str, offset);
        }

        public int IndexOf(MutableString str, int offset) {
            return _data.ToString().IndexOf(str.ToString(), offset);
        }

        public MutableString Remove(int position, int count) {
            _data.Remove(position, count);
            return this;
        }

        public MutableString Substring(int start, int length) {
            return new MutableString(_data.ToString(), start, length);
        }

        public void Clear() {
            _data.Remove(0, _data.Length);
        }

        public int Length {
            get { return _data.Length; }
        }

        public override string ToString() {
            return _data.ToString();
        }

        [RubyConstructor]
        public static MutableString Create() {
            return new MutableString();
        }

        [RubyMethodAttribute("*", RubyMethodAttributes.PublicInstance)]
        public static MutableString Repeat(MutableString str, int times) {
            if (times < 0)
                throw RubyExceptions.CreateArgumentError("negative argument");

            if (times == 0) {
                str.Clear();
            } else {
                string current = str.Chars.ToString();
                for (int i = 0; i < times - 1; ++i)
                    str.Append(current);
            }
                
            return str;
        }

        [RubyMethodAttribute("*", RubyMethodAttributes.PublicInstance)]
        public static MutableString Repeat(MutableString str, object times) {
            return Repeat(str, RubySites.ToInt(times));
        }

        public static MutableString Concatenate(MutableString lhs, string rhs) {
            return lhs.Append(rhs);
        }

        [RubyMethodAttribute("+", RubyMethodAttributes.PublicInstance)]
        public static MutableString Concatenate(MutableString self, MutableString other) {
            return self.Append(other);
        }

        [RubyMethodAttribute("+", RubyMethodAttributes.PublicInstance)]
        public static MutableString Concatenate(CodeContext context, MutableString self, object other) {
            return Concatenate(self, RubySites.ToStr(other));
        }

        [RubyMethodAttribute("<<", RubyMethodAttributes.PublicInstance)]
        [RubyMethodAttribute("concat", RubyMethodAttributes.PublicInstance)]
        public static MutableString Append(MutableString self, MutableString other) {
            return self.Append(other);
        }

        [RubyMethodAttribute("<<", RubyMethodAttributes.PublicInstance)]
        [RubyMethodAttribute("concat", RubyMethodAttributes.PublicInstance)]
        public static MutableString Append(MutableString self, int c) {
            if (c < 0 || c > 255)
                throw RubyExceptions.CannotConvertTypeToTargetType(c, "String");

            return self.Append((char)c);
        }
        
        [RubyMethodAttribute("<<", RubyMethodAttributes.PublicInstance)]
        [RubyMethodAttribute("concat", RubyMethodAttributes.PublicInstance)]
        public static MutableString Append(MutableString self, object other) {
            if (RubySites.RespondTo(other, "to_str"))
                return Append(self, RubySites.ToStr(other));
            else
                throw RubyExceptions.CannotConvertTypeToTargetType(other, "String");
        }

        [RubyMethodAttribute("<=>", RubyMethodAttributes.PublicInstance)]
        public static int Compare(MutableString self, MutableString other) {
            return self.ToString().CompareTo(other.ToString());
        }

        [RubyMethodAttribute("<=>", RubyMethodAttributes.PublicInstance)]
        public static object Compare(MutableString self, object other) {
            if (!RubySites.RespondTo(other, "to_str"))
                return null;
            else
                return Compare(self, RubySites.ToStr(other));
        }

        [RubyMethodAttribute("==", RubyMethodAttributes.PublicInstance)]
        [RubyMethodAttribute("===", RubyMethodAttributes.PublicInstance)]
        public static bool StringEquals(MutableString lhs, MutableString rhs) {
            if (lhs.Length != rhs.Length)
                return false;

            for (int i = 0; i < lhs.Length; ++i)
                if (lhs.Chars[i] != rhs.Chars[i])
                    return false;

            return true;
        }

        [RubyMethodAttribute("==", RubyMethodAttributes.PublicInstance)]
        [RubyMethodAttribute("===", RubyMethodAttributes.PublicInstance)]
        public static bool Equals(MutableString self, object other) {
            if (!RubySites.RespondTo(other, "to_str"))
                return false;
            else
                return StringEquals(self, RubySites.ToStr(other));
        }

        [RubyMethodAttribute("[]", RubyMethodAttributes.PublicInstance)]
        public static object GetChar(MutableString str, int index) {
            index = index < 0 ? index + str.Length : index;
            return (index < 0 || index >= str.Length) ? null : (object)(int)str[index];
        }

        [RubyMethodAttribute("[]", RubyMethodAttributes.PublicInstance)]
        public static object GetChar(MutableString str, object index) {
            return GetChar(str, RubySites.ToInt(index));
        }

        [RubyMethodAttribute("[]", RubyMethodAttributes.PublicInstance)]
        public static MutableString GetChars(MutableString str, int start, int length) {
            start = start < 0 ? start + str.Length : start;

            if (start == str.Length)
                return new MutableString();

            if (start < 0 || start > str.Length || length < 0)
                return null;

            if (start + length > str.Length)
                length = str.Length - start;

            return str.Substring(start, length);
        }

        [RubyMethodAttribute("[]", RubyMethodAttributes.PublicInstance)]
        public static MutableString GetChars(MutableString str, int start, object length) {
            return GetChars(str, start, RubySites.ToInt(length));
        }

        [RubyMethodAttribute("[]", RubyMethodAttributes.PublicInstance)]
        public static MutableString GetChars(MutableString str, object start, int length) {
            return GetChars(str, RubySites.ToInt(start), length);
        }

        [RubyMethodAttribute("[]", RubyMethodAttributes.PublicInstance)]
        public static MutableString GetChars(MutableString str, object start, object length) {
            return GetChars(str, RubySites.ToInt(start), RubySites.ToInt(length));
        }

        [RubyMethodAttribute("[]", RubyMethodAttributes.PublicInstance)]
        public static MutableString GetChars(MutableString str, Range range) {
            int begin = RubySites.ToInt(RubySites.RangeBegin(range));
            int end = RubySites.ToInt(RubySites.RangeEnd(range));

            begin = begin < 0 ? begin + str.Length : begin;

            if (begin < 0 || begin > str.Length)
                return null;

            end = end < 0 ? end + str.Length : end;

            int count = RubySites.RangeExcludeEnd(range) ? end - begin : end - begin + 1;
            return count < 0 ? new MutableString() : GetChars(str, begin, count);
        }

        [RubyMethodAttribute("[]", RubyMethodAttributes.PublicInstance)]
        public static MutableString GetChars(MutableString str, string searchStr) {
            return str.ToString().Contains(searchStr) ? new MutableString(searchStr) : null;
        }

        [RubyMethodAttribute("[]", RubyMethodAttributes.PublicInstance)]
        public static MutableString GetChars(MutableString str, MutableString searchStr) {
            return str.ToString().Contains(searchStr.ToString()) ? searchStr : null; 
        }

        [RubyMethodAttribute("[]", RubyMethodAttributes.PublicInstance)]
        public static MutableString GetChars(MutableString str, Regex regex) {
            if (regex.ToString() == String.Empty)
                return new MutableString();

            string result = regex.Match(str.ToString()).Value;
            return result == String.Empty ? null : new MutableString(result);
        }

        [RubyMethodAttribute("[]", RubyMethodAttributes.PublicInstance)]
        public static MutableString GetChars(MutableString str, Regex regex, int occurrance) {
            if (regex.ToString() == String.Empty)
                return new MutableString();

            MatchCollection matches = regex.Matches(str.ToString());
            if (occurrance < 0 || occurrance >= matches.Count)
                return null;

            string result = matches[occurrance].Value;
            return result == String.Empty ? null : new MutableString(result);
        }

        public static MutableString RemoveChar(MutableString str, int index) {
            str.Remove(index, 1);
            return new MutableString();
        }

        [RubyMethodAttribute("[]=", RubyMethodAttributes.PublicInstance)]
        public static MutableString SetChar(MutableString str, int index, MutableString value) {
            index = index < 0 ? index + str.Length : index;
            if (index < 0 || index >= str.Length)
                throw RubyExceptions.CreateIndexError(String.Format("index {0} out of string", index));

            if (value.Length == 0)
                return RemoveChar(str, index);

            str[index] = value[value.Length - 1];
            str.Insert(value.Substring(0, value.Length - 1), index);
            return value;
        }

        [RubyMethodAttribute("[]=", RubyMethodAttributes.PublicInstance)]
        public static int SetChar(MutableString str, int index, int value) {
            index = index < 0 ? index + str.Length : index;
            if (index < 0 || index >= str.Length)
                throw RubyExceptions.CreateIndexError(String.Format("index {0} out of string", index));

            str[index] = (char)((int)value & 0xFF);
            return value;
        }

        [RubyMethodAttribute("[]=", RubyMethodAttributes.PublicInstance)]
        public static int SetChar(MutableString str, object index, int value) {
            return SetChar(str, RubySites.ToInt(index), value);
        }

        [RubyMethodAttribute("[]=", RubyMethodAttributes.PublicInstance)]
        public static object SetChar(MutableString str, int index, object value) {
            if (RubySites.RespondTo(value, "to_int"))
                return SetChar(str, index, RubySites.ToInt(value));
            else if (RubySites.RespondTo(value, "to_str"))
                return SetChar(str, index, RubySites.ToStr(value));
            else
                throw RubyExceptions.CreateTypeError("can't convert ZZZ to String");
        }

        [RubyMethodAttribute("[]=", RubyMethodAttributes.PublicInstance)]
        public static object SetChar(MutableString str, object index, object value) {
            return SetChar(str, RubySites.ToInt(index), value);
        }

        [RubyMethodAttribute("[]=", RubyMethodAttributes.PublicInstance)]
        public static MutableString SetChar(MutableString str, int start, int charsToOverwrite, MutableString value) {
            if (charsToOverwrite < 0)
                throw RubyExceptions.CreateIndexError(String.Format("negative length {0}", charsToOverwrite));

            if (System.Math.Abs(start) > str.Length)
                throw RubyExceptions.CreateIndexError(String.Format("index {0} out of string", start));

            start = start < 0 ? start + str.Length : start;

            if (charsToOverwrite <= value.Length) {
                int insertIndex = start + charsToOverwrite;
                int limit = charsToOverwrite;
                if (insertIndex > str.Length) {
                    limit -= insertIndex - str.Length;
                    insertIndex = str.Length;
                }
                
                for (int i = 0; i < limit; ++i)
                    str[start + i] = value[i];

                Insert(str, insertIndex, value.ToString().Substring(limit));
            } else {
                int deleteIndex = start + value.Length;
                int limit = value.Length;
                if (deleteIndex > str.Length) {
                }

                for (int i = 0; i < value.Length; ++i)
                    str[start + i] = value[i];
                
                int pos = start + value.Length;
                int charsToRemove = charsToOverwrite - value.Length;
                int charsLeftInString = str.Length - pos;

                str.Remove(pos, System.Math.Min(charsToRemove, charsLeftInString));
            }
            return value;
        }

        [RubyMethodAttribute("[]=", RubyMethodAttributes.PublicInstance)]
        public static MutableString SetChar(MutableString str, int start, int charsToOverwrite, object value) {
            if (RubySites.RespondTo(value, "to_str"))
                return SetChar(str, start, charsToOverwrite, RubySites.ToStr(value));
            else
                throw RubyExceptions.CannotConvertTypeToTargetType(value, "String");
        }

        [RubyMethodAttribute("[]=", RubyMethodAttributes.PublicInstance)]
        public static MutableString SetChar(MutableString str, Range range, MutableString value) {
            int begin = RubySites.ToInt(RubySites.RangeBegin(range));
            int end = RubySites.ToInt(RubySites.RangeEnd(range));

            begin = begin < 0 ? begin + str.Length : begin;

            if (begin < 0 || begin > str.Length)
                throw RubyExceptions.CreateRangeError(String.Format("{0}..{1} out of range", begin, end));

            end = end < 0 ? end + str.Length : end;

            int count = RubySites.RangeExcludeEnd(range) ? end - begin : end - begin + 1;
            return SetChar(str, begin, count, value);
        }

        public static bool UpCaseChar(MutableString str, int index) {
            char current = str.Chars[index];
            if (current >= 'a' && current <= 'z') {
                str.Chars[index] = Char.ToUpper(current);
                return true;
            }
            return false;
        }

        public static bool DownCaseChar(MutableString str, int index) {
            char current = str.Chars[index];
            if (current >= 'A' && current <= 'Z') {
                str.Chars[index] = Char.ToLower(current);
                return true;
            }
            return false;
        }

        public static bool CapitalizeMutableString(MutableString str) {
            bool changed = false;
            if (str.Length > 0) {
                if (UpCaseChar(str, 0))
                    changed = true;
                for (int i = 1; i < str.Length; ++i) {
                    if (DownCaseChar(str, i))
                        changed = true;
                }
            }
            return changed;
        }

        public static bool UpCaseMutableString(MutableString str) {
            bool changed = false;
            for (int i = 0; i < str.Length; ++i) {
                if (UpCaseChar(str, i))
                    changed = true;
            }
            return changed;
        }

        public static bool DownCaseMutableString(MutableString str) {
            bool changed = false;
            for (int i = 0; i < str.Length; ++i) {
                if (DownCaseChar(str, i))
                    changed = true;
            }
            return changed;
        }

        [RubyMethodAttribute("capitalize!", RubyMethodAttributes.PublicInstance)]
        public static MutableString CapitalizeInPlace(MutableString str) {
            return CapitalizeMutableString(str) ? str : null;
        }

        [RubyMethodAttribute("capitalize", RubyMethodAttributes.PublicInstance)]
        public static MutableString Capitalize(MutableString str) {
            MutableString result = new MutableString(str);
            CapitalizeMutableString(result);
            return result;
        }

        [RubyMethodAttribute("downcase", RubyMethodAttributes.PublicInstance)]
        public static MutableString DownCase(MutableString str) {
            MutableString result = new MutableString(str);
            DownCaseMutableString(result);
            return result;
        }

        [RubyMethodAttribute("downcase!", RubyMethodAttributes.PublicInstance)]
        public static MutableString DownCaseInPlace(MutableString str) {
            return DownCaseMutableString(str) ? str : null;
        }

        public static string GetOctalRepresentationOfChar(char c) {
            string s = System.Convert.ToString(c, 8);
            return "\\" + s.PadLeft(3, '0');
        }

        public static string GetStringRepresentationOfChar(char c) {
            switch(c) {
                case '\a':
                    return "\\a";
                case '\b':
                    return "\\b";
                case '\n':
                    return "\\n";
                case '\v':
                    return "\\v";
                case '\f':
                    return "\\f";
                case '\r':
                    return "\\r";
                case (char)27:
                    return "\\e";
                case '"':
                    return "\\\"";
                default:
                    if (c < 32 || c > 126)
                        return GetOctalRepresentationOfChar(c);
                    else
                        return c.ToString();
            }
        }

        [RubyMethodAttribute("dump", RubyMethodAttributes.PublicInstance)]
        [RubyMethodAttribute("inspect", RubyMethodAttributes.PublicInstance)]
        public static MutableString Dump(MutableString self) {
            MutableString str = new MutableString();
            str.Append("\"");
            foreach (char c in self.ToString()) 
                str.Append(GetStringRepresentationOfChar(c));
            str.Append("\"");
            return str;
        }

        private static DynamicSite<Proc, BlockReturnReasonHolder, int, object> EachByteSharedSite =
            DynamicSite<Proc, BlockReturnReasonHolder, int, object>.Create(CallAction.Simple);

        [RubyMethodAttribute("each_byte", RubyMethodAttributes.PublicInstance)]
        public static MutableString EachByte(CodeContext context, MutableString str, [Optional]Proc block) {
            // TODO: this test cannot be verified since missing blocks aren't detected correctly upstream
            if (block == null)
                throw RubyExceptions.CreateLocalJumpError("no block given");

            BlockReturnReasonHolder returnReason = new BlockReturnReasonHolder();
            int i = 0;
            while (i < str.Length) {
                try {
                    EachByteSharedSite.Invoke(context, block, returnReason, str[i]);
                } catch (StackUnwinder.BlockReturn u) {
                    returnReason.Value = u.Reason;
                }

                switch (returnReason.Value) {
                    case BlockReturnReason.Undefined:
                        break;

                    case BlockReturnReason.Next:
                        returnReason.Value = BlockReturnReason.Undefined;
                        break;

                    case BlockReturnReason.BreakOrRetry:
                        return str;

                    case BlockReturnReason.Redo:
                        returnReason.Value = BlockReturnReason.Undefined;
                        continue;
                }
                i++;
            }
            return str;
        }

        public static void EachLine(CodeContext context, string[] lines, object block) {
            //object[] args = new object[1];
            //for (int i = 0; i < lines.Length; ++i) {
            //    // TODO: naive now, collapse empty ones later

            //    args[0] = lines[i];
            //    // TODO: fill out block template
            //    BlockReturnReason returnReason;
            //    RuntimeFlowControl.ClrMethod.YieldBlock(context, block, args, out returnReason);
            //    if (returnReason == BlockReturnReason.BreakOrRetry) break;
            //}
        }

        public static MutableString EachLine(CodeContext context, MutableString str, string separator, object block) {
            string[] lines = Utils.String.Split(str.ToString(), separator, Int32.MaxValue, StringSplitOptions.None);
            //string[] lines = str.ToString().Split(new string[] { separator }, StringSplitOptions.None);
            EachLine(context, lines, block);
            return str;
        }

        // TODO: implement overload of this method that accepts a single parameter
        [RubyMethodAttribute("each", RubyMethodAttributes.PublicInstance)]
        [RubyMethodAttribute("each_line", RubyMethodAttributes.PublicInstance)]
        public static object EachLine(CodeContext context, object self, [Block] object block) {
            Utils.Assert.NotNull(context, self, block);

            string separator = (string)RubyOps.GetGlobalVariable(context, Symbols.ForwardSlash);
            return EachLine(context, (MutableString)self, separator, block);
        }

        [RubyMethodAttribute("empty?", RubyMethodAttributes.PublicInstance)]
        public static bool Empty(MutableString str) {
            return str.Length == 0;
        }

        [RubyMethodAttribute("index", RubyMethodAttributes.PublicInstance)]
        public static object Index(MutableString str, string substring) {
            int result = str.IndexOf(substring);
            return result == -1 ? null : (object)result;
        }

        [RubyMethodAttribute("index", RubyMethodAttributes.PublicInstance)]
        public static object Index(MutableString str, MutableString substring) {
            int result = str.ToString().IndexOf(substring.ToString());
            return result == -1 ? null : (object)result;
        }

        [RubyMethodAttribute("index", RubyMethodAttributes.PublicInstance)]
        public static object Index(MutableString str, object substring) {
            if (RubySites.RespondTo(substring, "to_str"))
                return Index(str, RubySites.ToStr(substring));
            else
                throw RubyExceptions.CannotConvertTypeToTargetType(substring, "String");
        }

        [RubyMethodAttribute("index", RubyMethodAttributes.PublicInstance)]
        public static object Index(MutableString str, string substring, int offset) {
            offset = offset < 0 ? offset + str.Length : offset;
            if (offset < 0 || offset > str.Length)
                return null;
            int result = str.IndexOf(substring, offset);
            return result == -1 ? null : (object)result;
        }

        [RubyMethodAttribute("index", RubyMethodAttributes.PublicInstance)]
        public static object Index(MutableString str, MutableString substring, int offset) {
            return Index(str, substring.ToString(), offset);
        }

        [RubyMethodAttribute("index", RubyMethodAttributes.PublicInstance)]
        public static object Index(MutableString str, MutableString substring, object offset) {
            return Index(str, substring.ToString(), RubySites.ToInt(offset));
        }

        [RubyMethodAttribute("index", RubyMethodAttributes.PublicInstance)]
        public static object Index(MutableString str, object substring, int offset) {
            if (RubySites.RespondTo(substring, "to_str"))
                return Index(str, RubySites.ToStr(substring), offset);
            else
                throw RubyExceptions.CannotConvertTypeToTargetType(substring, "String");
        }

        [RubyMethodAttribute("index", RubyMethodAttributes.PublicInstance)]
        public static object Index(MutableString str, object substring, object offset) {
            return Index(str, substring, RubySites.ToInt(offset));
        }

        [RubyMethodAttribute("index", RubyMethodAttributes.PublicInstance)]
        public static object Index(MutableString str, int character) {
            int result = str.IndexOf((char)character);
            return result == -1 ? null : (object)result;
        }

        [RubyMethodAttribute("index", RubyMethodAttributes.PublicInstance)]
        public static object Index(MutableString str, int character, int offset) {
            offset = offset < 0 ? offset + str.Length : offset;
            if (offset < 0 || offset > str.Length)
                return null;
            int result = str.IndexOf((char)character, offset);
            return result == -1 ? null : (object)result;
        }

        [RubyMethodAttribute("index", RubyMethodAttributes.PublicInstance)]
        public static object Index(MutableString str, int character, object offset) {
            return Index(str, character, RubySites.ToInt(offset));
        }

        [RubyMethodAttribute("index", RubyMethodAttributes.PublicInstance)]
        public static object Index(MutableString str, Regex regex) {
            throw new NotImplementedError();
        }

        [RubyMethodAttribute("index", RubyMethodAttributes.PublicInstance)]
        public static object Index(MutableString str, Regex regex, int offset) {
            throw new NotImplementedError();
        }

        [RubyMethodAttribute("include?", RubyMethodAttributes.PublicInstance)]
        public static bool Include(MutableString str, string subString) {
            return str.ToString().IndexOf(subString) != -1;
        }

        [RubyMethodAttribute("include?", RubyMethodAttributes.PublicInstance)]
        public static bool Include(MutableString str, object other) {
            if (RubySites.RespondTo(other, "to_str"))
                return Include(str, RubySites.ToStr(other));
            else
                throw RubyExceptions.CannotConvertTypeToTargetType(other, "String");
        }

        [RubyMethodAttribute("include?", RubyMethodAttributes.PublicInstance)]
        public static bool Include(MutableString str, MutableString subString) {
            return str.IndexOf(subString.ToString()) != -1;
        }

        [RubyMethodAttribute("include?", RubyMethodAttributes.PublicInstance)]
        public static bool Include(MutableString str, int c) {
            return str.ToString().IndexOf((char)c) != -1;
        }

        [RubyMethodAttribute("insert", RubyMethodAttributes.PublicInstance)]
        public static MutableString Insert(MutableString str, int start, string value) {
            int offset = start < 0 ? start + str.Length + 1 : start;
            if (offset > str.Length || offset < 0)
                throw RubyExceptions.CreateIndexError(String.Format("index {0} out of string", start));

            return str.Insert(value, offset);
        }

        [RubyMethodAttribute("insert", RubyMethodAttributes.PublicInstance)]
        public static MutableString Insert(MutableString str, int start, MutableString value) {
            return Insert(str, start, value.ToString());
        }

        [RubyMethodAttribute("insert", RubyMethodAttributes.PublicInstance)]
        public static MutableString Insert(MutableString self, int start, object value) {
            if (!RubySites.RespondTo(value, "to_str"))
                throw RubyExceptions.CannotConvertTypeToTargetType(value, "String");

            return Insert(self, start, RubySites.ToStr(value));
        }

        [RubyMethodAttribute("size", RubyMethodAttributes.PublicInstance)]
        [RubyMethodAttribute("length", RubyMethodAttributes.PublicInstance)]
        public static int MutableStringLength(MutableString str) {
            return str.Length;
        }

        public static int GetIndexOfRightmostAlphaNumericCharacter(MutableString str, int index) {
            for (int i = index; i >= 0; --i) 
                if (Char.IsLetterOrDigit(str[i]))
                    return i;

            return -1;
        }

        // TODO: remove recursion
        public static void IncrementAlphaNumericChar(MutableString str, int index) {
            char c = str[index];
            if (c == 'z' || c == 'Z' || c == '9') {
                int nextIndex = GetIndexOfRightmostAlphaNumericCharacter(str, index - 1);
                if (c == 'z') {
                    str[index] = 'a';
                    if (nextIndex == -1)
                        str.Insert("a", index);
                    else
                        IncrementAlphaNumericChar(str, nextIndex);
                } else if (c == 'Z') {
                    str[index] = 'A';
                    if (nextIndex == -1)
                        str.Insert("A", index);
                    else
                        IncrementAlphaNumericChar(str, nextIndex);
                } else {
                    str[index] = '0';
                    if (nextIndex == -1)
                        str.Insert("1", index);
                    else
                        IncrementAlphaNumericChar(str, nextIndex);
                }
            } else {
                IncrementChar(str, index);
            }
        }

        public static void IncrementChar(MutableString str, int index) {
            byte c = (byte)str[index];
            if (c == 255) {
                str[index] = (char)0;
                if (index > 0)
                    IncrementChar(str, index - 1);
                else
                    str.Insert("\u0001", 0);
            } else {
                str[index] = ((char)(c + 1));
            }
        }

        [RubyMethodAttribute("succ!", RubyMethodAttributes.PublicInstance)]
        public static MutableString SuccInPlace(MutableString str) {
            int index = GetIndexOfRightmostAlphaNumericCharacter(str, str.Length - 1);
            if (index == -1) {
                IncrementChar(str, str.Length - 1);
            } else {
                IncrementAlphaNumericChar(str, index);
            }
            return str;
        }

        [RubyMethodAttribute("succ", RubyMethodAttributes.PublicInstance)]
        public static MutableString Succ(MutableString str) {
            return SuccInPlace(new MutableString(str));
        }

        [RubyMethodAttribute("upcase", RubyMethodAttributes.PublicInstance)]
        public static MutableString UpCase(MutableString str) {
            MutableString result = new MutableString(str);
            UpCaseMutableString(result);
            return result;
        }

        [RubyMethodAttribute("upcase!", RubyMethodAttributes.PublicInstance)]
        public static MutableString UpCaseInPlace(MutableString str) {
            return UpCaseMutableString(str) ? str : null;
        }

        [RubyMethodAttribute("split", RubyMethodAttributes.PublicInstance)]
        public static List<object> Split(CodeContext/*!*/ context, MutableString/*!*/ self) {
            object separator = RubyOps.GetGlobalVariable(context, Symbols.Semicolon);
            string[] elements;
            if (separator == null) 
                elements = Utils.String.Split(self.ToString(), new char[] { ' ', '\n', '\r', '\t' }, Int32.MaxValue, StringSplitOptions.RemoveEmptyEntries);
            else
                elements = Utils.String.Split(self.ToString(), separator.ToString(), Int32.MaxValue, StringSplitOptions.RemoveEmptyEntries);

            List<object> result = new List<object>(elements.Length);
            foreach (string item in elements)
                result.Add(new MutableString(item));
            return result;
        }

        [RubyMethodAttribute("to_clr_string", RubyMethodAttributes.PublicInstance)]
        public static string ToClrString(MutableString str) {
            return str.ToString();
        }

        [RubyMethodAttribute("to_s", RubyMethodAttributes.PublicInstance)]
        [RubyMethodAttribute("to_str", RubyMethodAttributes.PublicInstance)]
        public static MutableString ToS(MutableString self) {
            return self;
        }
    }
#if false
    public class MutableStringNextGen {

        //^ invariant (_chars == null || _string == null) == (_bytes != null) 
        //^ invariant (_bytes != null) == (_encoding != null)
        //^ invariant (_bytes != null) ==> (_count <= _bytes.length)
        //^ invariant (_chars != null) ==> (_count <= _chars.length)
        //^ invariant (_count >= 0)

        private byte[] _bytes;
        private char[] _chars;
        private string _string;
        private int _byteBufferSize;
        private int _charBufferSize;

        // When we convert to a CLR string, this flag indicates that we need to enable
        // round-tripping of the encoding.
        private bool _preserveEncoding;

        // Used whenever we convert to Unicode (for scenarios like upcase and downcase)
        private Encoding _encoding;

        public MutableStringNextGen() {
            _charBufferSize = 0;
            _chars = null;
        }

        public MutableStringNextGen(byte[]/*!*/ bytes, Encoding/*!*/ encoding) : this(bytes, encoding, false) { }

        public MutableStringNextGen(byte[]/*!*/ bytes, Encoding/*!*/ encoding, bool preserveEncoding) {
            Utils.Assert.NotNull(bytes, encoding);
            _bytes = bytes;
            _byteBufferSize = bytes.Length;
            _string = null;
            _encoding = encoding;
            _preserveEncoding = preserveEncoding;
        }

        public MutableStringNextGen(string str) {
            Utils.Assert.NotNull(str);
            _bytes = null;
            _byteBufferSize = 0;
            _string = str;
            _encoding = null;
            _preserveEncoding = false;
        }

        public static void AssociateEncoding(string str, Encoding encoding) {
            // spooky hashtable magic
        }

        public static Encoding/*!*/ GetAssociatedEncoding(string str) {
            // return default encoding if we can't find it in the weak hash
            return Encoding.UTF8;
        }

        public Encoding Encoding {
            get { return _encoding == null ? Encoding.UTF8 : _encoding; }
        }

        public char[] Chars {
            get {
                if (_chars == null) {
                    _chars = _string.ToCharArray();
                    _charBufferSize = _chars.Length;
                }
                return _chars;
            }
        }

        public byte[] Bytes {
            get {
                if (_bytes == null) {
                    _bytes = _chars == null ? Encoding.GetBytes(_string) : Encoding.GetBytes(_chars);
                    _byteBufferSize = _bytes.Length;
                }
                return _bytes;
            }
        }

        public int Length {
            get {
                if (_chars != null)
                    return _charBufferSize;
                else if (_bytes != null)
                    return _byteBufferSize;
                else
                    return _string.Length;
            }
        }

        public override string ToString() {
            if (Chars != null) {
                return new string(Chars);
            } else {
                string result = _encoding.GetString(_bytes);
                if (_preserveEncoding) {
                    AssociateEncoding(result, _encoding);
                }
                return result;
            }
        }

        public byte GetByte(int index) {
            return Bytes[index];
        }

        public char GetChar(int index) {
            return Chars[index];
        }

        public void SetChar(int index, char value) {
            Chars[index] = value;
        }

        public void SetByte(int index, byte value) {
            Bytes[index] = value;
        }

        private int CalculateNextPowerOfTwo(int v) {
            v--;
            v |= v >> 1;
            v |= v >> 2;
            v |= v >> 4;
            v |= v >> 8;
            v |= v >> 16;
            return v++;
        }

        private void ExpandByteBuffer(int desiredSize) {
            int newSize = CalculateNextPowerOfTwo(desiredSize);
            byte[] current = _bytes;
            _bytes = new byte[newSize];
            Array.Copy(current, _bytes, current.Length);
        }

        private void ExpandCharBuffer(int desiredSize) {
            int newSize = CalculateNextPowerOfTwo(desiredSize);
            char[] current = _chars;
            _chars = new char[newSize];
            Array.Copy(current, _chars, current.Length);
        }

        public void Append(byte[] bytes) {
            if (_byteBufferSize + bytes.Length > _bytes.Length)
                ExpandByteBuffer(_byteBufferSize + bytes.Length);

            Array.Copy(bytes, 0, _bytes, _byteBufferSize, bytes.Length);
            _byteBufferSize += bytes.Length;
        }

        public void Append(string str) {
            char[] chars = str.ToCharArray();
            if (_charBufferSize + chars.Length > _chars.Length)
                ExpandCharBuffer(_charBufferSize + chars.Length);

            Array.Copy(chars, 0, _chars, _charBufferSize, chars.Length);
            _charBufferSize += chars.Length;
        }
    }
#endif
}
