/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Scripting;
using System.Runtime.InteropServices;

using Ruby.Runtime;

namespace Ruby.Builtins {

    [RubyExtensionClass("Object", typeof(object)), Includes(typeof(Kernel))]
    public static class ObjectOps {

        #region Public Instance Methods

        [RubyMethodAttribute("to_s", RubyMethodAttributes.PublicInstance)]
        [RubyMethodAttribute("inspect", RubyMethodAttributes.PublicInstance)]
        public static MutableString/*!*/ ToString(CodeContext/*!*/ context, object self) {
            return new MutableString(String.Format("#<{0}:{1:x8}>", RubyOps.GetExecutionContext(context).GetClassOf(self).Name, GetObjectId(self)));
        }

        [RubyMethodAttribute("class", RubyMethodAttributes.PublicInstance)]
        public static RubyClass/*!*/ GetClass(CodeContext/*!*/ context, object self) {
            return RubyOps.GetExecutionContext(context).GetClassOf(self);
        }

        [RubyMethodAttribute("id", RubyMethodAttributes.PublicInstance)]
        public static int GetId(object self) {
            RubyConsole.Warning("Object#id will be deprecated; use Object#object_id");
            return GetObjectId(self);
        }

        [RubyMethodAttribute("object_id", RubyMethodAttributes.PublicInstance)]
        public static int GetObjectId(object self) {
            if (self == null) return 4;
            if (self is TrueClass) return 2;
            if (self is FalseClass) return 0;
            if (self is bool) return (bool)self ? 2 : 0;
            if (self is Int32) return (int)self * 2 + 1;

            // TODO:
            return unchecked((int)IdDispenser.GetId(self));
        }

        [RubyMethodAttribute("equal?", RubyMethodAttributes.PublicInstance)]
        public static bool Equal(object self, object other) {
            return GetObjectId(self) == GetObjectId(other);
        }

        [RubyMethodAttribute("==", RubyMethodAttributes.PublicInstance)]
        public static new bool Equals(object self, object other) {
            if (self == null) return other == null;
            return self.Equals(other);
        }

        [RubyMethodAttribute("nil?", RubyMethodAttributes.PublicInstance)]
        public static bool IsNil(object self) {
            return self == null;
        }

        [RubyMethodAttribute("is_a?", RubyMethodAttributes.PublicInstance)]
        [RubyMethodAttribute("kind_of?", RubyMethodAttributes.PublicInstance)]
        public static bool IsKindOf(object self, RubyModule/*!*/ other) {
            if (other == null) throw new ArgumentNullException("other");
            return other.ExecutionContext.GetClassOf(self).HasAncestor(other);
        }

        [RubyMethodAttribute("instance_of?", RubyMethodAttributes.PublicInstance)]
        public static bool IsOfClass(object self, RubyModule/*!*/ other) {
            if (other == null) throw new ArgumentNullException("other");
            return other.ExecutionContext.GetClassOf(self) == other;
        }

        [RubyMethodAttribute("respond_to?", RubyMethodAttributes.PublicInstance)]
        public static bool RespondTo(CodeContext context, object self, MutableString methodName) {
            return RubyOps.GetExecutionContext(context).ResolveMethod(self, SymbolTable.StringToId(methodName)) != null;
        }

        [RubyMethodAttribute("respond_to?", RubyMethodAttributes.PublicInstance)]
        public static bool RespondTo(CodeContext context, object self, SymbolId methodName) {
            return RubyOps.GetExecutionContext(context).ResolveMethod(self, methodName) != null;
        }

        #endregion
    }
}
