/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;

using Microsoft.Scripting;
using Microsoft.Scripting.Actions;
using MSA = Microsoft.Scripting.Ast;
using Ast = Microsoft.Scripting.Ast.Ast;

using Ruby.Runtime;
using Ruby.Hosting;
using System.Reflection;
using System.Threading;

namespace Ruby.Builtins {

    internal enum ProcKind {
        Block,
        Lambda
    }

    [RubyClass]
    public class Proc : IDynamicObject {
        private static int _CurrentId = 0;

        private readonly RubyMethodInfo/*!*/ _info;
        private readonly ProcKind _kind;

        // blocks are closed over self and context:
        private readonly object/*!*/ _self;
        private readonly CodeContext/*!*/ _context;
        private readonly int _id;

        internal RubyMethodInfo Info { get { return _info; } }
        internal ProcKind Kind { get { return _kind; } }
        public object/*!*/ Self { get { return _self; } }
        public CodeContext/*!*/ Context { get { return _context; } }
        public int Id { get { return _id; } }

        internal static PropertyInfo/*!*/ SelfProperty { get { return typeof(Proc).GetProperty("Self"); } }
        internal static PropertyInfo/*!*/ IdProperty { get { return typeof(Proc).GetProperty("Id"); } }
        internal static PropertyInfo/*!*/ CodeContextProperty { get { return typeof(Proc).GetProperty("Context"); } }

        internal Proc(ProcKind kind, RubyMethodInfo/*!*/ info, CodeContext/*!*/ context, object/*!*/ self) {
            Utils.Assert.NotNull(info, context, self);
            _kind = kind;
            _info = info;
            _context = context;
            _self = self;
            _id = Interlocked.Increment(ref _CurrentId);
        }

        #region Public Instance Methods

        //    ==
        //    arity
        //    binding

        [RubyMethod("[]")]
        [RubyMethod("call")]
        // TODO:        [ActionOnCallAttribute]
        public static object Call(Proc/*!*/ self, [NotNull]params object[] args) {
            if (args == null) args = RubyOps.NullArray;

            // TODO: (less/more args, unsplatting)
            return Utils.Reflection.InvokeDelegate(self._info.Method, Utils.Array.Insert<object>(self, new BlockReturnReasonHolder(), args));
        }

        //    clone
        //    dup
        //    to_proc
        //    to_s

        #endregion

        #region Singleton Methods

        //    new

        #endregion

        #region IDynamicObject Members

        public LanguageContext/*!*/ LanguageContext {
            get { return RubyEngine.CurrentEngine.DefaultContext.LanguageContext; }
        }

        public StandardRule<T>/*!*/ GetRule<T>(Action/*!*/ action, CodeContext/*!*/ context, object[]/*!*/ args) {
            StandardRule<T> result = new StandardRule<T>();

            ActionBinder binder = context.LanguageContext.Binder;

            switch (action.Kind) {
                case ActionKind.Call:
                    SetCallRule<T>(result, binder, (CallAction)action, args);
                    break;
                
                default:
                    return null;
            }

            return result;
        }

        /// <summary>
        /// Yield.
        /// </summary>
        internal void SetCallRule<T>(StandardRule<T>/*!*/ result, ActionBinder/*!*/ binder, CallAction/*!*/ action, object[]/*!*/ args) {
            Utils.Assert.NotNull(result, binder, action, args);

            // TODO: do we need to cache the rule and share it accross yields at all?
            result.AddTest(Ast.Equal(Ast.ReadProperty(Ast.Cast(result.Parameters[0], typeof(Proc)), IdProperty), Ast.Constant(_id)));
            _info.SetInvocationRule(binder, result, action.ArgumentInfos, args, null);
        }

        #endregion

    }
}
