/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System.Collections.Generic;
using System.Collections;
using System.Text;
using System.Diagnostics;
using Microsoft.Scripting;

using MSA = Microsoft.Scripting.Ast;
using Ast = Microsoft.Scripting.Ast.Ast;
using Ruby.Runtime;
using Microsoft.Scripting.Actions;

namespace Ruby.Compiler.AST {

    public class Arguments : Node {
        private readonly List<Expression> _arguments;
        private readonly List<Maplet> _maplets;
        private readonly Expression _array;
        private Block _block;

        public Block Block { get { return _block; } internal set { _block = value; } }

        public bool IsEmpty {
            get { return _arguments == null && _maplets == null && _array == null && _block == null; }
        }
        
        public Arguments(SourceSpan location)
            : base(location) {
            _arguments = null;
            _maplets = null;
            _array = null;
            _block = null;
        }

        public Arguments(Expression/*!*/ arg)
            : base(arg.Location) {
            Utils.Assert.NotNull(arg);

            _arguments = new List<Expression>();
            _arguments.Add(arg);
            _maplets = null;
            _array = null;
            _block = null;
        }

        public Arguments(List<Expression> arguments, List<Maplet> maplets, Expression array, Block block, SourceSpan location)
            : base(location) {
            _arguments = arguments;
            _maplets = maplets;
            _array = array;
            _block = block;
        }

        #region Transform to Actual Arguments

        /// <summary>
        /// Prepares a list of expressions and argument infos for the call. 
        /// Used by both block-call and method-invocation.
        /// </summary>
        private void MakeActualArguments(AstGenerator/*!*/ gen, MSA.Expression/*!*/ target, MSA.Expression specialArg, MSA.Expression additionalArg,
            out ArgumentInfo[]/*!*/ infos, out MSA.Expression[]/*!*/ args) {

            Utils.Assert.NotNull(gen, target);

            List<MSA.Expression> simpleArgs = MakeSimpleArgumentExpressionList(gen, additionalArg);

            args = new MSA.Expression[
                1 +                                     // target
                (specialArg != null ? 1 : 0) +     // special argument
                (_block != null ? 1 : 0) +              // block
                simpleArgs.Count +                      // simple args
                (_array != null ? 1 : 0) +              // array args
                (additionalArg != null ? 1 : 0)         // additional arg
            ];

            infos = new ArgumentInfo[args.Length];

            int j = 0;

            args[j] = target;
            infos[j] = new ArgumentInfo(MSA.ArgumentKind.Instance);
            j++;

            if (specialArg != null) {
                args[j] = specialArg;
                infos[j] = new ArgumentInfo(MSA.ArgumentKind.Simple);
                j++;
            }

            if (_block != null) {
                args[j] = _block.Transform(gen);
                infos[j] = new ArgumentInfo(MSA.ArgumentKind.Block);
                j++;
            }

            for (int i = 0; i < simpleArgs.Count; i++, j++) {
                args[j] = simpleArgs[i];
                infos[j] = new ArgumentInfo(MSA.ArgumentKind.Simple);
            }

            if (_array != null) {
                args[j] = _array.TransformRead(gen);
                infos[j] = new ArgumentInfo(MSA.ArgumentKind.List);
                j++;
            }

            if (additionalArg != null) {
                args[j] = additionalArg;
                infos[j] = new ArgumentInfo(MSA.ArgumentKind.Simple);
                j++;
            }
        }

        private List<MSA.Expression>/*!*/ MakeSimpleArgumentExpressionList(AstGenerator/*!*/ gen, MSA.Expression additionalArg) {
            Utils.Assert.NotNull(gen);

            List<MSA.Expression> args = new List<MSA.Expression>(
                ((_arguments != null) ? _arguments.Count : 0) +
                ((additionalArg != null) ? 1 : 0) +
                ((_maplets != null) ? 1 : 0)
            );

            if (_arguments != null) {
                gen.TranformExpressions(_arguments, args);
            }

            if (_maplets != null) {
                args.Add(gen.TransformToHashConstructor(_maplets));
            }

            return args;
        }

        internal static MSA.Expression/*!*/ TransformToBlockCall(AstGenerator/*!*/ gen, Arguments arguments, MSA.Variable/*!*/ blockVariable, MSA.Variable/*!*/ returnReasonVariable) {
            Utils.Assert.NotNull(gen, blockVariable, returnReasonVariable);

            MSA.Expression[] args;
            CallAction action;
            
            if (arguments != null) {
                Debug.Assert(arguments._block == null);
                
                ArgumentInfo[] infos;
                arguments.MakeActualArguments(gen, Ast.Read(blockVariable), Ast.Read(returnReasonVariable), null, out infos, out args); 
                action = CallAction.Make(infos);

            } else {
                args = new MSA.Expression[] { Ast.Read(blockVariable), Ast.Read(returnReasonVariable) };
                action = CallAction.Simple;
            }

            return Ast.Action.Call(action, typeof(object), args);
        }

        internal static MSA.Expression/*!*/ TransformToMethodInvocation(AstGenerator/*!*/ gen, SymbolId methodName, Expression target, 
            Arguments arguments, MSA.Expression additionalArg) {
            Utils.Assert.NotNull(gen);

            MSA.Expression transformedTarget = (target != null) ? target.TransformRead(gen) : gen.MakeSelfRead();
            
            InvokeMemberActionFlags flags = InvokeMemberActionFlags.ReturnNonCallable;
            if (target != null) flags |= InvokeMemberActionFlags.IsCallWithThis;

            ArgumentInfo[] infos;
            MSA.Expression[] args;

            if (arguments != null) {
                arguments.MakeActualArguments(gen, transformedTarget, null, additionalArg, out infos, out args);
            } else {
                if (additionalArg != null) {
                    args = new MSA.Expression[] { transformedTarget, additionalArg };
                    infos = new ArgumentInfo[] { new ArgumentInfo(MSA.ArgumentKind.Instance), new ArgumentInfo(MSA.ArgumentKind.Simple) };
                } else {
                    args = new MSA.Expression[] { transformedTarget };
                    infos = new ArgumentInfo[] { new ArgumentInfo(MSA.ArgumentKind.Instance) };
                }
            }

            return Ast.Action.InvokeMember(methodName, typeof(object), flags, infos, args);
        }
        
        private MSA.Expression/*!*/ PackAllArguments(AstGenerator/*!*/ gen, MSA.Expression additionalArg, bool packSingleArg) {
            Utils.Assert.NotNull(gen); 
            
            return PackArguments(gen, MakeSimpleArgumentExpressionList(gen, additionalArg), additionalArg, packSingleArg);
        }

        /// <summary>
        /// The <c>additinalArg</c> is appended after regular arguments but before splatted ones.
        /// </summary>
        private MSA.Expression/*!*/ PackArguments(AstGenerator/*!*/ gen, List<MSA.Expression>/*!*/ simpleArgs, MSA.Expression additionalArg, 
            bool packSingleArg) {

            Utils.Assert.NotNull(gen, simpleArgs);

            if (_array != null) {
                if (additionalArg != null) {
                    return gen.OpCall("AddSplattedAndAdditionalArguments", new MSA.Expression[] { 
                        Ast.NewArray(typeof(object[]), simpleArgs), 
                        _array.TransformRead(gen),
                        additionalArg
                    });
                } else {
                    return gen.OpCall("AddSplattedArguments", new MSA.Expression[] { 
                        Ast.NewArray(typeof(object[]), simpleArgs),
                        _array.TransformRead(gen)
                    });
                }
            } else {
                if (additionalArg != null) {
                    simpleArgs.Add(additionalArg);
                }

                if (packSingleArg || simpleArgs.Count > 1) {
                    return Ast.NewArray(typeof(object[]), simpleArgs);
                }

                return simpleArgs[0];
            }
        }

        public static MSA.Expression/*!*/ TransformToArray(AstGenerator/*!*/ gen, Arguments arguments, MSA.Expression additionalArg) {
            Utils.Assert.NotNull(gen);
            
            if (arguments == null) {
                List<MSA.Expression> result = new List<MSA.Expression>();
                
                if (additionalArg != null) {
                    result.Add(additionalArg);
                }

                return Ast.NewArray(typeof(object[]), result);
            } else {
                return arguments.PackAllArguments(gen, additionalArg, true);
            }
        }

        #endregion

        #region Tranform to Return Value

        private MSA.Expression/*!*/ TransformToReturnValue(AstGenerator/*!*/ gen) {
            Utils.Assert.NotNull(gen);
            Debug.Assert(_block == null);

            if (_array == null) {
                if (_arguments != null && _arguments.Count == 1 && _maplets == null) {
                    return PackAllArguments(gen, null, false);
                }

                if (_maplets != null && _arguments == null) {
                    return gen.TransformToHashConstructor(_maplets);
                }
            }

            return gen.OpCall("MakeArray", new MSA.Expression[] { 
                PackAllArguments(gen, null, true)
            });
        }
        
        public static MSA.Expression/*!*/ TransformToReturnValue(AstGenerator/*!*/ gen, Arguments arguments) {
            return (arguments != null) ? arguments.TransformToReturnValue(gen) : Ast.Null();
        }

        #endregion
    }
}
