/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

// Copyright (c) Microsoft Corporation. All rights reserved.

using System;
using System.Collections.Generic;
using System.Text;
using MSA = Microsoft.Scripting.Ast;
using Microsoft.Scripting;
using System.Reflection;
using Ruby.Runtime;
using System.Diagnostics;

using Ast = Microsoft.Scripting.Ast.Ast;

namespace Ruby.Compiler.AST {
    internal static class AstFactory {

        public static readonly MSA.Statement/*!*/ EmptyStatement = Ast.Empty();
        public static readonly MSA.Expression[]/*!*/ EmptyExpressions = new MSA.Expression[0];
        public static readonly MSA.Expression/*!*/ Null = Ast.Null();
        public static readonly MSA.Expression/*!*/ True = Ast.True();
        public static readonly MSA.Expression/*!*/ False = Ast.False();
        public static readonly MSA.Statement/*!*/ Rethrow = Ast.Statement(Ast.Rethrow());

        #region If

        public static MSA.IfStatement/*!*/ If(MSA.Expression/*!*/ condition, params MSA.Statement[]/*!*/ statements) {
            Utils.Assert.NotNull(condition, statements);

            return Ast.If(new MSA.IfStatementTest[] { Ast.IfCondition(condition, Ast.Block(statements)) }, null);
        }

        public static MSA.IfStatement/*!*/ If(MSA.Expression/*!*/ condition, MSA.Statement/*!*/ trueStatement, MSA.Statement falseStatement) {
            Utils.Assert.NotNull(condition, trueStatement);

            return Ast.If(new MSA.IfStatementTest[] { Ast.IfCondition(condition, trueStatement) }, falseStatement);
        }

        #endregion

        #region Assignments

        public static MSA.Statement/*!*/ Assign(MSA.Variable/*!*/ variable, MSA.Expression/*!*/ expression) {
            Utils.Assert.NotNull(variable, expression);

            return Ast.Statement(Ast.Assign(variable, expression));
        }

        public static MSA.Statement/*!*/ Assign(MSA.Variable/*!*/ variable, FieldInfo/*!*/ field, MSA.Expression/*!*/ expression) {
            Utils.Assert.NotNull(variable, field, expression);
            return Ast.Statement(Ast.AssignField(Ast.Read(variable), field, expression));
        }

        public static MSA.Statement/*!*/ AssignValueOpt(MSA.Variable variable, object value) {
            return (variable != null) ? AssignValue(variable, value) : EmptyStatement;
        }

        public static MSA.Statement/*!*/ AssignValue(MSA.Variable/*!*/ variable, object value) {
            return Assign(variable, Ast.Constant(value));
        }

        public static MSA.Statement/*!*/ AssignValue(MSA.Variable variable, FieldInfo/*!*/ field, object value) {
            return Assign(variable, field, Ast.Constant(value));
        }

        public static MSA.Statement/*!*/ AssignValueWithCast(MSA.Variable/*!*/ variable, Type/*!*/ variableType, FieldInfo/*!*/ field, object value) {
            Utils.Assert.NotNull(variable, variableType, field);
            return Ast.Statement(
                Ast.AssignField(
                    Ast.Cast(Ast.Read(variable), variableType), field, Ast.Constant(value)
                )
            );
        }

        #endregion

        public static MSA.Expression[]/*!*/ ToArgs(MSA.Variable/*!*/ variable) {
            return new MSA.Expression[] { Ast.Read(variable) };
        }

        public static MSA.Expression/*!*/ Result(MSA.Variable/*!*/ resultVariable, MSA.Statement/*!*/ statement) {
            return Ast.Comma(
                1,
                Ast.Void(statement),
                Ast.Read(resultVariable)
            );
        }

        #region Equal

        public static MSA.Expression/*!*/ Equal(MSA.Variable leftVariable, FieldInfo leftField, object value) {
            return Equal(leftVariable, leftField, Ast.Constant(value));
        }

        public static MSA.Expression/*!*/ Equal(MSA.Variable leftVariable, FieldInfo leftField, MSA.Expression/*!*/ expression) {
            Debug.Assert(expression != null);
            Debug.Assert(leftVariable != null || leftField != null);

            MSA.Expression lhs = (leftField == null) ? Ast.Read(leftVariable) :
                (MSA.Expression)Ast.ReadField(leftVariable != null ? Ast.Read(leftVariable) : null, leftField);
            return Ast.Equal(lhs, expression);
        }

        public static MSA.Expression/*!*/ Equal(MSA.Variable leftVariable, FieldInfo leftField, MSA.Variable rightVariable, FieldInfo rightField) {
            Debug.Assert(rightVariable != null || rightField != null);

            MSA.Expression rhs = (rightField == null) ? Ast.Read(rightVariable) : 
                (MSA.Expression)Ast.ReadField(rightVariable != null ? Ast.Read(rightVariable) : null, rightField);

            return Equal(leftVariable, leftField, rhs);
        }

        #endregion
    }
}
