/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;

using MSA = Microsoft.Scripting.Ast;
using Microsoft.Scripting;
using Ruby.Runtime;
using Ast = Microsoft.Scripting.Ast.Ast;
using Ruby.Builtins;

namespace Ruby.Compiler.AST {

    public class BlockDefinition : Block {
        //	{ |args| body }

        // context, self (TODO)
        internal const int HiddenParameterCount = 2;

        private readonly CompoundLeftValue/*!*/ _parameters;
        private readonly List<Statement>/*!*/ _body;

        /// <summary>
        /// Null for for-loops.
        /// </summary>
        private readonly LexicalScope _definedScope;

        // transformation:
        public MSA.Variable ReturnReasonVariable { get { return _returnReasonVariable; } }
        private MSA.Variable _returnReasonVariable;

        public BlockDefinition(LexicalScope definedScope, CompoundLeftValue/*!*/ parameters, List<Statement>/*!*/ body, SourceSpan location)
            : base(location) {
            Utils.Assert.NotNull(parameters, body);

            _definedScope = definedScope;
            _body = body;
            _parameters = parameters;
        }

        public override MSA.Expression/*!*/ Transform(AstGenerator/*!*/ gen) {
            MSA.CodeBlock methodBlock = Ast.CodeBlock(Location, "#block");
            methodBlock.Parent = gen.CurrentCodeBlock;

            // TODO: unsplatting, nesting
            
            MSA.Variable[] parameters = new MSA.Variable[
                HiddenParameterCount + 
                _parameters.LeftValues.Count + 
                (_parameters.UnsplattedValue != null ? 1 : 0)
            ];
            
            // hidden parameters:
            parameters[0] = MSA.Variable.Parameter(methodBlock, SymbolTable.StringToId("#this-block"), typeof(Proc));
            parameters[1] = _returnReasonVariable = MSA.Variable.Parameter(methodBlock, SymbolTable.StringToId("#return-reason"), typeof(BlockReturnReasonHolder)); // TODO: out BlockReturnReason

            for (int i = HiddenParameterCount; i < parameters.Length; i++) {
                parameters[i] = MSA.Variable.Parameter(methodBlock, SymbolTable.StringToId("#" + (i - HiddenParameterCount)), typeof(object)); // TODO: last variable of type List<object>
            }

            methodBlock.Parameters.AddRange(parameters);

            MSA.Variable oldSelf = gen.CurrentSelfVariable;
            MSA.CodeBlock oldBlock = gen.CurrentCodeBlock;

            gen.CurrentCodeBlock = methodBlock;
            gen.CurrentSelfVariable = methodBlock.CreateLocalVariable(Symbols.Self, typeof(object));
            BlockDefinition enclosingBlock = gen.EnterRubyBlock(this);

            if (_definedScope != null) {
                gen.CurrentScope = _definedScope;
                _definedScope.TransformLocals(methodBlock, gen);
            }
            
            List<MSA.Statement> prologue = new List<MSA.Statement>();

            // TODO:
            // self = #this-block.self
            prologue.Add(Ast.Write(gen.CurrentSelfVariable, Ast.ReadProperty(Ast.Read(parameters[0]), Proc.SelfProperty)));

            int j = HiddenParameterCount;
            foreach (LeftValue parameter in _parameters.LeftValues) {
                prologue.Add(Ast.Statement(
                    parameter.TransformWrite(gen, gen.MakeParameterRead(parameters[j])))
                );
                j++;
            }

            if (_parameters.UnsplattedValue != null) {
                prologue.Add(Ast.Statement(
                    _parameters.UnsplattedValue.TransformWrite(gen, gen.MakeParameterRead(parameters[j])))
                );
                j++;
            }

            methodBlock.Body = gen.TransformBodyStatements(_body, prologue, null);

            if (_definedScope != null) {
                gen.CurrentScope = _definedScope.ParentScope;
            }

            gen.CurrentSelfVariable = oldSelf;
            gen.CurrentCodeBlock = oldBlock;
            gen.LeaveRubyBlock(enclosingBlock);

            return gen.OpCall("DefineBlock", new MSA.Expression[] {
                Ast.CodeContext(),
                gen.MakeSelfRead(),
                Ast.CodeBlockExpression(methodBlock, false, true),
                Ast.Constant(_parameters.LeftValues.Count),
                Ast.Constant(_parameters.UnsplattedValue != null)
            });
        }
    }
}
