/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Scripting;
using MSA = Microsoft.Scripting.Ast;
using Ast = Microsoft.Scripting.Ast.Ast;
using Ruby.Runtime;

namespace Ruby.Compiler.AST {
    public class Body : Node {
        // begin/class/def
        //   statements
        // rescue-clauses 
        // else
        //   statements
        // ensure
        //   statements
        // end

        private List<Statement>/*!*/ _statements;		     
        private List<RescueClause> _rescueClauses;	// optional
        private List<Statement> _elseStatements;	// optional
        private List<Statement> _ensureStatements;	// optional
        private MSA.Variable _retryingVariable;

        public MSA.Variable RetryingVariable { get { return _retryingVariable; } }

        private bool HasExceptionHandling {
            get {
                return (_statements.Count > 0 && _rescueClauses != null) || _elseStatements != null || _ensureStatements != null;
            }
        }

        public Body(List<Statement>/*!*/ statements, List<RescueClause> rescueClauses, List<Statement> elseStatements,
            List<Statement> ensureStatements, SourceSpan location)
            : base(location) {

            Utils.Assert.NotNull(statements);

            _statements = statements;
            _rescueClauses = rescueClauses;
            _elseStatements = elseStatements;
            _ensureStatements = ensureStatements;
        }

        public MSA.Expression/*!*/ TransformToExpression(AstGenerator/*!*/ gen) {
            Utils.Assert.NotNull(gen);

            if (HasExceptionHandling) {
                MSA.Variable resultVariable = gen.CurrentCodeBlock.CreateTemporaryVariable(SymbolTable.StringToId("#block-result"), typeof(object));

                return Ast.Comma(
                    1,
                    Ast.Void(TransformExceptionHandling(gen, resultVariable)),
                    Ast.Read(resultVariable)
                );

            } else {
                return gen.TransformStatementsToExpression(_statements);
            }
        }
        
        public MSA.Statement/*!*/ TransformToBodyStatement(AstGenerator/*!*/ gen, IList<MSA.Statement> prologue) {
            Utils.Assert.NotNull(gen);

            if (HasExceptionHandling) {
                MSA.Statement stmt = TransformExceptionHandling(gen, null);

                if (prologue != null) {
                    MSA.Statement[] result = Utils.Array.MakeArray(prologue, 0, 1);
                    result[prologue.Count] = stmt;
                    return Ast.Block(result);
                } else {
                    return stmt;
                }
            } else {
                return gen.TransformBodyStatements(_statements, prologue, null);
            }
        }

        private MSA.Statement/*!*/ TransformExceptionHandling(AstGenerator/*!*/ gen, MSA.Variable resultVariable) {
            Utils.Assert.NotNull(gen);

            Body enclosingTryRescue = gen.EnterTryRescueBlock(this);

            MSA.Statement/*!*/ transformedBody;
            MSA.Statement transformedEnsure = (_ensureStatements != null) ? gen.TransformStatements(_ensureStatements) : null;
            MSA.Statement transformedElse = (_elseStatements != null) ? gen.TransformBodyStatements(_elseStatements, null, resultVariable) : null;

            if (_statements.Count > 0) {

                // body should do return, but there is else-clause => we cannot do return from the guarded statements: 
                // (the value of the last expression in the body cannot be the last executed expression statement => we can ignore it):
                if (_elseStatements != null) {
                    transformedBody = gen.TransformStatements(_statements);
                } else {
                    transformedBody = gen.TransformBodyStatements(_statements, null, resultVariable);
                }

            } else {
                // no body => no exception can be thrown => no rescue clauses could be invoked => no retry is applicable => no loop

                if (_elseStatements == null) {
                    return transformedEnsure ?? AstFactory.EmptyStatement;
                }

                if (_ensureStatements == null) {
                    return transformedElse ?? AstFactory.EmptyStatement;
                }

                return Ast.TryFinally(transformedElse, transformedEnsure);
            }

            MSA.Variable exceptionThrownVariable = (_elseStatements != null) ? gen.CurrentCodeBlock.CreateTemporaryVariable(SymbolTable.StringToId("#exception-thrown"), typeof(bool)) : null;

            if (_rescueClauses == null) {
                // no rescue => no retry => no exception catching 

                if (_elseStatements == null) {
                    return Ast.TryFinally(transformedBody, transformedEnsure);
                }

                return Ast.Block(
                    Ast.Try(
                        transformedBody
                    ).Catch(typeof(Exception),
                        AstFactory.AssignValue(exceptionThrownVariable, true) 
                    ),
                    Ast.Unless(Ast.Read(exceptionThrownVariable), transformedEnsure)
                );
            }

            MSA.Variable/*!*/ exceptionVariable = gen.CurrentCodeBlock.CreateTemporaryVariable(SymbolTable.StringToId("#exception"), typeof(Exception));
            MSA.Variable/*!*/ oldExceptionVariable = gen.CurrentCodeBlock.CreateTemporaryVariable(SymbolTable.StringToId("#old-exception"), typeof(Exception));
            _retryingVariable = (_ensureStatements != null) ? gen.CurrentCodeBlock.CreateTemporaryVariable(SymbolTable.StringToId("#retrying"), typeof(bool)) : null;

            MSA.IfStatementTest[] handlers = new MSA.IfStatementTest[_rescueClauses.Count];
            for (int i = 0; i < handlers.Length; i++) {
                handlers[i] = _rescueClauses[i].Transform(gen, resultVariable);
            }

            // outer-most EH blocks sets and clears runtime flag RuntimeFlowControl.InTryRescue:
            MSA.Statement setInTryRescue, clearInTryRescue;
            if (enclosingTryRescue == null) {
                setInTryRescue = AstFactory.AssignValue(gen.CurrentRuntimeFlowControlVariable, gen.RfcInTryRescueField, true);
                clearInTryRescue = AstFactory.AssignValue(gen.CurrentRuntimeFlowControlVariable, gen.RfcInTryRescueField, false);
            } else {
                setInTryRescue = clearInTryRescue = AstFactory.EmptyStatement;
            }


            // TODO: optimize number of try-finaly blocks for misisng else/ensure clauses

            MSA.Statement result = Ast.Loop(
                AstFactory.AssignValueOpt(exceptionThrownVariable, false),
                AstFactory.AssignValueOpt(_retryingVariable, false),
                Ast.Try(
                    Ast.Try(
                        setInTryRescue,
                        transformedBody
                    ).Catch(typeof(StackUnwinder), 
                        Ast.Statement(Ast.Rethrow())
                    ).Catch(typeof(Exception), exceptionVariable,
                        AstFactory.AssignValueOpt(exceptionThrownVariable, true),
                        clearInTryRescue,
                        
                        // save exception:
                        AstFactory.Assign(oldExceptionVariable, gen.OpCall("SaveException", Ast.CodeContext(), Ast.Read(exceptionVariable))),

                        Ast.Try(
                            Ast.If(handlers, Ast.Statement(Ast.Rethrow()))
                        ).Catch(typeof(StackUnwinder.Retry),
                            AstFactory.AssignValueOpt(_retryingVariable, true),
                            Ast.Continue()
                        )
                    ).Finally(
                        // restore previous exception:
                        Ast.Statement(gen.OpCall("RestoreException", Ast.CodeContext(), Ast.Read(oldExceptionVariable))),
                        clearInTryRescue
                    ),

                    // unless (exception_thrown) do <else-statements> end
                    (_elseStatements != null) ? Ast.Unless(Ast.Read(exceptionThrownVariable), transformedElse) : AstFactory.EmptyStatement
                ).Finally(
                    (_ensureStatements != null) ? Ast.Unless(Ast.Read(_retryingVariable), transformedEnsure) : AstFactory.EmptyStatement
                ),
                // break
                Ast.Break()
            );

            gen.LeaveTryRescueBlock(enclosingTryRescue);

            return result;
        }
    }
}
