/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;

using Microsoft.Scripting;
using MSA = Microsoft.Scripting.Ast;
using Ast = Microsoft.Scripting.Ast.Ast;

using Ruby.Runtime;

namespace Ruby.Compiler.AST {
    public class RescueClause : Node {
        // rescue types => target
        //   statements

        private readonly List<Expression> _types;	// optional
        private readonly LeftValue _target;		    // optional
        private readonly List<Statement> _body;   // optional

        public RescueClause(List<Statement> body, SourceSpan location)
            : base(location) {
            _body = body;
        }

        public RescueClause(List<Expression> types, LeftValue target, List<Statement> body, SourceSpan location)
            : base(location) {
            Debug.Assert(_target == null || _types != null);

            _types = types;
            _target = target;
            _body = body;
        }

        //
        // rescue stmts                     ... if (StandardError === $!) { stmts; } 
        // rescue <types> stmts             ... temp1 = type1; ...; if (<temp1> === $! || ...) { stmts; }
        // rescue <types> => <lvalue> stmts ... temp1 = type1; ...; if (<temp1> === $! || ...) { <lvalue> = $!; stmts; }
        // 
        public MSA.IfStatementTest/*!*/ Transform(AstGenerator/*!*/ gen, MSA.Variable resultVariable) {
            Utils.Assert.NotNull(gen);

            MSA.Expression condition;
            if (_types != null) {

                if (_types.Count == 1) {
                    condition = MakeCompareException(gen, _types[0].TransformRead(gen));
                } else {
                    
                    // forall{i}: <temps[i]> = evaluate type[i]
                    MSA.Expression[] temps = new MSA.Expression[_types.Count];
                    MSA.Expression[] exprs = new MSA.Expression[_types.Count + 1];
                    for (int i = 0; i < temps.Length; i++) {
                        MSA.Variable tmp = gen.CurrentCodeBlock.CreateTemporaryVariable(SymbolTable.StringToId("#type_" + i), typeof(object));
                        temps[i] = Ast.Read(tmp);
                        exprs[i] = Ast.Assign(tmp, _types[i].TransformRead(gen));
                    }

                    // disjoin-all{i}: RubyOps.CompareException(<temps[i]>)
                    condition = MakeCompareException(gen, temps[0]);
                    for (int i = 1; i < _types.Count; i++) {
                        condition = Ast.Or(_types[i].Location, condition, MakeCompareException(gen, temps[i]));
                    }

                    exprs[exprs.Length - 1] = condition;
                    condition = Ast.Comma(exprs.Length - 1, exprs); 
                }
                
            } else {
                condition = gen.OpCall("CompareDefaultException", new MSA.Expression[] {
                    Ast.CodeContext(),
                    gen.MakeSelfRead()
                });
            }

            MSA.Statement[] prologue = null;
            if (_target != null) {
                // <lvalue> = e;
                prologue = new MSA.Statement[] { 
                    Ast.Statement(_target.TransformWrite(gen, gen.OpCall("GetCurrentException", Ast.CodeContext()))) 
                };
            }

            return Ast.IfCondition(condition, gen.TransformBodyStatements(_body, prologue, resultVariable));
        }

        private MSA.Expression/*!*/ MakeCompareException(AstGenerator/*!*/ gen, MSA.Expression/*!*/ expression) {
            Utils.Assert.NotNull(gen, expression);

            return gen.OpCall("CompareException", new MSA.Expression[] { 
                Ast.CodeContext(),
                expression
            });
        }
    }
}
