/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Scripting;

using MSA = Microsoft.Scripting.Ast;
using Ruby.Runtime;
using System.Diagnostics;

namespace Ruby.Compiler.AST {
    using Ast = Microsoft.Scripting.Ast.Ast;
    using Ruby.Builtins;

    public sealed class MethodDeclaration : DeclarationExpression {
        //	def target.name(params)
        //		body
        //	end

        // object DefineMethod(RubyModule/*!*/ module, string/*!*/ name, Delegate/*!*/ realMethod) {

        // self, block
        internal const int HiddenParameterCount = 2;
            
            
        private readonly Expression/*!*/ _target;
        private readonly SymbolId _name;
        private readonly Parameters/*!*/ _parameters;

        public MethodDeclaration(LexicalScope/*!*/ definedScope, Expression target, SymbolId name, Parameters parameters, Body/*!*/ body, 
            SourceSpan location)
            : base(definedScope, body, location) {

            _target = target;
            _name = name;
            _parameters = parameters ?? Parameters.Empty;
        }

        public override MSA.Expression/*!*/ TransformRead(AstGenerator/*!*/ gen) {
            MSA.Expression transformedTarget = (_target != null) ? _target.TransformRead(gen) : Ast.Null();

            MethodDeclaration enclosingMethodDeclaration = gen.EnterMethodDeclaration(this);
            MSA.Variable oldSelf = gen.CurrentSelfVariable;
            MSA.Variable oldBlockParam = gen.CurrentBlockParameter;
            MSA.CodeBlock oldBlock = gen.CurrentCodeBlock;
            MSA.Variable oldRuntimeControlFlowVariable = gen.CurrentRuntimeFlowControlVariable;

            MSA.CodeBlock methodBlock = Ast.CodeBlock(Location, _name);

            // hidden parameters:
            MSA.Variable[] parameters = DefinedScope.TransformParameters(_parameters, methodBlock, HiddenParameterCount, gen);
            parameters[0] = MSA.Variable.Parameter(methodBlock, Symbols.Self, typeof(object));

            if (_parameters.Block != null) {
                parameters[1] = MSA.Variable.Parameter(methodBlock, _parameters.Block.Name, typeof(Proc));
                DefinedScope.SetTransformedParameter(_parameters.Block.Name, parameters[1]);
            } else {
                parameters[1] = MSA.Variable.Parameter(methodBlock, SymbolTable.StringToId("#block"), typeof(Proc));
            }

            methodBlock.Parameters.AddRange(parameters);

            gen.CurrentScope = DefinedScope;
            gen.CurrentCodeBlock = methodBlock;
            gen.CurrentSelfVariable = parameters[0];
            gen.CurrentBlockParameter = parameters[1];
            gen.CurrentRuntimeFlowControlVariable = methodBlock.CreateVariable(SymbolTable.StringToId("#rfc"), MSA.Variable.VariableKind.Local, typeof(RuntimeFlowControl));

            DefinedScope.TransformLocals(methodBlock, gen);

            MSA.Variable unwinder = gen.CurrentCodeBlock.CreateTemporaryVariable(SymbolTable.StringToId("#unwinder"), typeof(StackUnwinder.Return));

            methodBlock.Body = Ast.Try(
                // rfc = new RuntimeFlowControl(true)
                Ast.Write(gen.CurrentRuntimeFlowControlVariable,
                    Ast.New(RuntimeFlowControl.Constructor, new MSA.Expression[] { AstFactory.True })
                ),
                
                _parameters.TransformOptionalsInitialization(gen),
                
                Body.TransformToBodyStatement(gen, null)

            ).Filter(typeof(StackUnwinder.Return), unwinder, AstFactory.Equal(unwinder, StackUnwinder.Return.FlowControlField, gen.CurrentRuntimeFlowControlVariable, null),

                // return unwinder.ReturnValue;
                Ast.Return(Ast.ReadField(Ast.Read(unwinder), StackUnwinder.Return.ReturnValueField))

            ).Finally(
                Ast.Write(gen.CurrentRuntimeFlowControlVariable, gen.RfcInMethodField, Ast.False())
            );

            gen.CurrentScope = DefinedScope.ParentScope;
            gen.CurrentSelfVariable = oldSelf;
            gen.CurrentBlockParameter = oldBlockParam;
            gen.CurrentCodeBlock = oldBlock;
            gen.CurrentRuntimeFlowControlVariable = oldRuntimeControlFlowVariable;
            gen.LeaveMethodDeclaration(enclosingMethodDeclaration);

            return gen.OpCall("DefineMethod", new MSA.Expression[] {
                Ast.CodeContext(),
                gen.MakeSelfRead(),
                transformedTarget,
                Ast.Constant(_name),
                Ast.CodeBlockExpression(methodBlock, false, true),
                Ast.Constant(_parameters.MandatoryCount),
                Ast.Constant(_parameters.OptionalCount), 
                Ast.Constant(_parameters.Array != null)
            });
        }
    }
}
