/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using Microsoft.Scripting;
using MSA = Microsoft.Scripting.Ast;
using Ast = Microsoft.Scripting.Ast.Ast;
using Ruby.Runtime;

namespace Ruby.Compiler.AST {
    public class WhileLoopExpression : Expression {
        //	while/until cond
        //		body
        //	end

        private readonly LexicalScope/*!*/ _scope;
        private readonly Expression _condition;
        private readonly List<Statement> _body;		// optional
        private readonly bool _isWhileLoop;

        // tranformation:
        private MSA.Variable _redoVariable;
        private MSA.Variable _resultVariable;

        // used by jump statements:
        internal MSA.Variable ResultVariable { get { Utils.Assert.NotNull(_resultVariable); return _resultVariable; } }
        internal MSA.Variable RedoVariable { get { Utils.Assert.NotNull(_redoVariable); return _redoVariable; } }

        public WhileLoopExpression(LexicalScope/*!*/ scope, Expression/*!*/ condition, bool isWhileLoop, List<Statement>/*!*/ body, SourceSpan location)
            : base(location) {

            Utils.Assert.NotNull(scope, condition);
            Utils.Assert.NotNullItems(body);

            _scope = scope;
            _condition = condition;
            _isWhileLoop = isWhileLoop;
            _body = body;
        }

        // see Ruby Language.doc/Runtime/Control Flow Implementation/While-Until
        public override MSA.Expression/*!*/ TransformRead(AstGenerator/*!*/ gen) {
            _resultVariable = gen.CurrentCodeBlock.CreateTemporaryVariable(SymbolTable.StringToId("#loop-result"), typeof(object));
            _redoVariable = gen.CurrentCodeBlock.CreateTemporaryVariable(SymbolTable.StringToId("#redo"), typeof(bool));

            WhileLoopExpression enclosingLoop = gen.EnterLoop(this);
            MSA.BlockStatement transformedBody = gen.TransformStatements(_body);
            gen.LeaveLoop(enclosingLoop);

            MSA.Statement conditionPositiveStmt, conditionNegativeStmt;
            if (_isWhileLoop) {
                conditionPositiveStmt = Ast.Empty();
                conditionNegativeStmt = Ast.Break();
            } else {
                conditionPositiveStmt = Ast.Break();
                conditionNegativeStmt = Ast.Empty();
            }

            MSA.Statement clearRedoStmt = Ast.Statement(
                Ast.Assign(_redoVariable, Ast.False())
            );
            
            MSA.Variable unwinder = gen.CurrentCodeBlock.CreateTemporaryVariable(SymbolTable.StringToId("#unwinder"), typeof(StackUnwinder.BlockReturn));
            
            MSA.Statement loop = Ast.Loop( 
                Ast.Try(
                    Ast.If(Ast.Read(_redoVariable),
                        clearRedoStmt
                    ).ElseIf(_condition.TransformRead(gen),
                        conditionPositiveStmt
                    ).Else(
                        conditionNegativeStmt
                    ),

                    transformedBody 

                ).Catch(typeof(StackUnwinder.BlockReturn), unwinder,
                    // if (unwinder.Reason == BreakOrRetry) { result = unwinder.ReturnValue; break; }
                    Ast.IfThen(AstFactory.Equal(unwinder, StackUnwinder.BlockReturn.ReasonField, BlockReturnReason.BreakOrRetry), 
                        AstFactory.Assign(_resultVariable, Ast.ReadField(Ast.Read(unwinder), StackUnwinder.BlockReturn.ReturnValueField)),
                        Ast.Break()
                    ),

                    // redo = (u.Reason == Redo)
                    AstFactory.Assign(_redoVariable, AstFactory.Equal(unwinder, StackUnwinder.BlockReturn.ReasonField, BlockReturnReason.BreakOrRetry))
                )
            );

            // TODO: currently this is not correct test - while { def foo() { while } }
            if (enclosingLoop == null) {
                loop = Ast.Try(
                    AstFactory.AssignValue(gen.CurrentRuntimeFlowControlVariable, gen.RfcInLoopField, true),
                    loop
                ).Finally(
                    AstFactory.AssignValue(gen.CurrentRuntimeFlowControlVariable, gen.RfcInLoopField, false)
                );
            }

            return AstFactory.Result(_resultVariable, loop);
        }
    }
}
