/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System;
using System.Text;
using System.Diagnostics;
using System.Collections.Generic;

using Ruby.Compiler.AST;

using Microsoft.Scripting;
using MSA = Microsoft.Scripting.Ast;
using Microsoft.Scripting.Hosting;
using Ruby.Runtime;
using System.Reflection;
using Ast = Microsoft.Scripting.Ast.Ast;
using Microsoft.Scripting.Actions;

namespace Ruby.Compiler {

    public sealed class AstGenerator {

        #region Reflection Cache

        private static readonly FieldInfo/*!*/ _rfcInTryRescueField;
        private static readonly FieldInfo/*!*/ _rfcInLoopField;
        private static readonly FieldInfo/*!*/ _rfcInMethodField;
        private static readonly FieldInfo/*!*/ _rfcInBlockField;
        private static readonly FieldInfo/*!*/ _rfcRetrySingleton;

        internal FieldInfo/*!*/ RfcInTryRescueField { get { return _rfcInTryRescueField; } }
        internal FieldInfo/*!*/ RfcInLoopField { get { return _rfcInLoopField; } }
        internal FieldInfo/*!*/ RfcInMethodField { get { return _rfcInMethodField; } }
        internal FieldInfo/*!*/ RfcInBlockField { get { return _rfcInBlockField; } }
        internal FieldInfo/*!*/ RfcRetrySingleton { get { return _rfcRetrySingleton; } }
        
        #endregion

        private LexicalScope _scope;
        private MSA.Variable _currentSelfVariable;
        private MSA.Variable _currentBlockParameter;
        private MSA.CodeBlock _currentBlock;
        private WhileLoopExpression _currentLoop;
        private BlockDefinition _currentRubyBlock;
        private MethodDeclaration _currentMethod;
        private ModuleDeclaration _currentModule;
        private MSA.Variable _currentRuntimeFlowControlVariable;
        private Body _currentTryRescueBlock;

        #region State

        public LexicalScope CurrentScope {
            get { return _scope; }
            set { _scope = value; }
        }

        public MSA.Variable CurrentSelfVariable {
            get { return _currentSelfVariable; }
            set { _currentSelfVariable = value; }
        }

        public MSA.Variable CurrentBlockParameter {
            get { return _currentBlockParameter; }
            set { _currentBlockParameter = value; }
        }

        public MSA.CodeBlock CurrentCodeBlock {
            get { return _currentBlock; }
            set { _currentBlock = value; }
        }

        public WhileLoopExpression CurrentLoop {
            get { return _currentLoop; }
            set { _currentLoop = value; }
        }

        public BlockDefinition CurrentRubyBlock {
            get { return _currentRubyBlock; }
            set { _currentRubyBlock = value; }
        }

        public MSA.Variable CurrentRuntimeFlowControlVariable {
            get { return _currentRuntimeFlowControlVariable; }
            set { _currentRuntimeFlowControlVariable = value; }
        }

        public Body CurrentTryRescueBlock {
            get { return _currentTryRescueBlock; }
            set { _currentTryRescueBlock = value; }
        }

        public MethodDeclaration CurrentMethod {
            get { return _currentMethod; }
            set { _currentMethod = value; }
        }

        public ModuleDeclaration CurrentModule {
            get { return _currentModule; }
            set { _currentModule = value; }
        }

        #endregion

        static AstGenerator() {
            _rfcInTryRescueField = typeof(RuntimeFlowControl).GetField("InTryRescue");
            _rfcInLoopField = typeof(RuntimeFlowControl).GetField("InLoop");
            _rfcInMethodField = typeof(RuntimeFlowControl).GetField("InMethod");
            _rfcInBlockField = typeof(RuntimeFlowControl).GetField("InBlock");
            _rfcRetrySingleton = typeof(RuntimeFlowControl).GetField("RetrySingleton");

            Utils.Assert.NotNull(_rfcInTryRescueField, _rfcInLoopField, _rfcInMethodField, _rfcInBlockField);
            Utils.Assert.NotNull(_rfcRetrySingleton);
        }

        public AstGenerator() {
            _scope = null;
            _currentBlock = null;
            _currentSelfVariable = null;
            _currentBlockParameter = null;
            _currentTryRescueBlock = null;
            _currentLoop = null;
            _currentRubyBlock = null;
            _currentRuntimeFlowControlVariable = null;
            _currentMethod = null;
            _currentModule = null;
        }

        #region Entering and Leaving Flow Significant Nodes

        public WhileLoopExpression EnterLoop(WhileLoopExpression/*!*/ loop) {
            Utils.Assert.NotNull(loop);
            WhileLoopExpression result = _currentLoop;
            _currentLoop = loop;
            return result;
        }

        public void LeaveLoop(WhileLoopExpression enclosingLoop) {
            _currentLoop = enclosingLoop;
        }

        public BlockDefinition EnterRubyBlock(BlockDefinition/*!*/ block) {
            Utils.Assert.NotNull(block);
            BlockDefinition result = _currentRubyBlock;
            _currentRubyBlock = block;
            return result;
        }

        public void LeaveRubyBlock(BlockDefinition enclosingBlock) {
            _currentRubyBlock = enclosingBlock;
        }

        public Body EnterTryRescueBlock(Body/*!*/ blockBody) {
            Utils.Assert.NotNull(blockBody);
            Body result = _currentTryRescueBlock;
            _currentTryRescueBlock = blockBody;
            return result;
        }

        public void LeaveTryRescueBlock(Body enclosingBlockBody) {
            _currentTryRescueBlock = enclosingBlockBody;
        }

        public MethodDeclaration EnterMethodDeclaration(MethodDeclaration/*!*/ method) {
            Utils.Assert.NotNull(method);
            MethodDeclaration result = _currentMethod;
            _currentMethod = method;
            return result;
        }

        public void LeaveMethodDeclaration(MethodDeclaration enclosingMethod) {
            _currentMethod = enclosingMethod;
        }

        #endregion

        #region Calls

        internal MSA.MethodCallExpression/*!*/ OpCall(string/*!*/ methodName, params MSA.Expression[]/*!*/ args) {
            Utils.Assert.NotNull(methodName, args);
            return Ast.Call(null, RubyOps.GetMethod(methodName), args);
        }
        
        internal MSA.MethodCallExpression/*!*/ OpCall(string/*!*/ methodName, MSA.Variable/*!*/ arg) {
            Utils.Assert.NotNull(methodName, arg);
            return Ast.Call(null, RubyOps.GetMethod(methodName), AstFactory.ToArgs(arg));
        }

        internal MSA.MethodCallExpression/*!*/ OpCall(string/*!*/ methodName) {
            Utils.Assert.NotNull(methodName);
            return Ast.Call(null, RubyOps.GetMethod(methodName), AstFactory.EmptyExpressions);
        }

        internal MSA.MethodCallExpression/*!*/ RfcCall(string/*!*/ methodName, params MSA.Expression[]/*!*/ args) {
            Utils.Assert.NotNull(methodName, args);
            return Ast.Call(Ast.Read(CurrentRuntimeFlowControlVariable), RuntimeFlowControl.GetMethod(methodName), args);
        }

        internal MSA.MethodCallExpression/*!*/ RfcCall(string/*!*/ methodName, MSA.Variable/*!*/ arg) {
            Utils.Assert.NotNull(methodName, arg);
            return Ast.Call(Ast.Read(CurrentRuntimeFlowControlVariable), RuntimeFlowControl.GetMethod(methodName), AstFactory.ToArgs(arg));
        }

        internal MSA.MethodCallExpression/*!*/ RfcCall(string/*!*/ methodName) {
            Utils.Assert.NotNull(methodName);
            return Ast.Call(Ast.Read(CurrentRuntimeFlowControlVariable), RuntimeFlowControl.GetMethod(methodName), AstFactory.EmptyExpressions);
        }

        internal MSA.MethodCallExpression/*!*/ RfcStaticCall(string/*!*/ methodName, MSA.Expression[]/*!*/ args) {
            Utils.Assert.NotNull(methodName, args);
            return Ast.Call(null, RuntimeFlowControl.GetMethod(methodName), args);
        }

        #endregion

        internal MSA.Expression/*!*/ MakeSelfRead() {
            if (_currentSelfVariable != null) {
                return MakeParameterRead(_currentSelfVariable);
            } else {
                return SelfReference.MakeContextSelf(this);
            }
        }

        internal MSA.BoundExpression MakeTempExpression(string name, Type type, SourceSpan span) {
            SymbolId symbol = SymbolTable.StringToId(name);
            MSA.Variable temp = _currentBlock.CreateTemporaryVariable(symbol, type);
            return Ast.Read(span, temp);
        }

        internal MSA.Expression/*!*/ MakeParameterRead(MSA.Variable/*!*/ parameter) {
            Utils.Assert.NotNull(parameter);
            return Ast.Read(parameter);
        }

        internal List<MSA.Expression>/*!*/ TranformExpressions(List<Expression>/*!*/ arguments) {
            Utils.Assert.NotNull(arguments);
            return TranformExpressions(arguments, new List<MSA.Expression>(arguments.Count));
        }

        internal List<MSA.Expression>/*!*/ TranformExpressions(List<Expression>/*!*/ arguments, List<MSA.Expression>/*!*/ result) {
            Utils.Assert.NotNullItems(arguments);
            Utils.Assert.NotNull(result);

            foreach (Expression arg in arguments) {
                result.Add(arg.TransformRead(this));
            }

            return result;
        }

        internal MSA.BlockStatement/*!*/ TransformBodyStatements(List<Statement>/*!*/ statements, IList<MSA.Statement> prologue, MSA.Variable resultVariable) {
            MSA.Statement[] result = Utils.Array.MakeArray(prologue, 0, statements.Count);

            int j = (prologue != null) ? prologue.Count : 0;
            for (int i = 0; i < statements.Count - 1; i++) {
                result[j++] = statements[i].Transform(this);
            }

            // if the last statement is an expression-statement do return <expr>:
            if (statements.Count > 0) {
                int i = statements.Count - 1;
                ExpressionStatement exprStmt = statements[i] as ExpressionStatement;
                if (exprStmt != null) {
                    MSA.Expression transformedExpr = exprStmt.Expression.TransformRead(this);
                    if (resultVariable != null) {
                        result[j++] = AstFactory.Assign(resultVariable, transformedExpr);
                    } else {
                        result[j++] = Ast.Return(transformedExpr);
                    }
                } else {
                    result[j++] = statements[i].Transform(this);
                }
            }

            return Ast.Block(result);
        }

        internal MSA.BlockStatement/*!*/ TransformStatements(List<Statement>/*!*/ statements) {
            Utils.Assert.NotNull(statements);

            MSA.Statement[] result = new MSA.Statement[statements.Count];
            for (int i = 0; i < result.Length; i++) {
                result[i] = statements[i].Transform(this);
            }

            return Ast.Block(result);
        }

        internal MSA.Expression/*!*/ TransformStatementsToExpression(List<Statement> statements) {
            if (statements != null && statements.Count > 0) {
                MSA.Expression[] result;

                ExpressionStatement lastExprStmt = statements[statements.Count - 1] as ExpressionStatement;
                if (lastExprStmt != null) {

                    if (statements.Count == 1) {
                        return lastExprStmt.Expression.TransformRead(this);
                    }

                    result = new MSA.Expression[statements.Count];

                    for (int i = 0; i < result.Length - 1; i++) {
                        result[i] = Ast.Void(statements[i].Transform(this));
                    }
                    result[result.Length - 1] = lastExprStmt.Expression.TransformRead(this);

                } else {
                    result = new MSA.Expression[statements.Count + 1];

                    for (int i = 0; i < result.Length - 1; i++) {
                        result[i] = Ast.Void(statements[i].Transform(this));
                    }
                    result[result.Length - 1] = Ast.Null();
                }
                return Ast.Comma(result.Length - 1, result);

            } else {
                return Ast.Null();
            }
        }

        internal List<MSA.Expression>/*!*/ TransformMapletsToExpressions(IList<Maplet>/*!*/ maplets) {
            Utils.Assert.NotNullItems(maplets);
            return TransformMapletsToExpressions(maplets, new List<MSA.Expression>(maplets.Count * 2));
        }
            
        internal List<MSA.Expression>/*!*/ TransformMapletsToExpressions(IList<Maplet>/*!*/ maplets, List<MSA.Expression>/*!*/ result) {
            Utils.Assert.NotNullItems(maplets);
            Utils.Assert.NotNull(result);

            foreach (Maplet entry in maplets) {
                result.Add(entry.Key.TransformRead(this));
                result.Add(entry.Value.TransformRead(this));
            }

            return result;
        }

        public MSA.Expression/*!*/ TransformToHashConstructor(IList<Maplet>/*!*/ maplets) {
            Utils.Assert.NotNull(maplets);

            return Ast.Call(
                null,
                RubyOps.GetMethod("MakeHash"),
                Ast.NewArray(typeof(object[]), TransformMapletsToExpressions(maplets))
            );
        }
    }
}
