/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System.Collections.Generic;
using System.Text;
using System.Diagnostics;

using Microsoft.Scripting;
using MSA = Microsoft.Scripting.Ast;

using Ruby.Compiler.AST;
using System;

namespace Ruby.Compiler {
    public enum ScopeKind {
        SourceUnit,
        Method,
        Module,
        Block,
        Initializer,
        RuntimeScope
    }

    public sealed class VariableInfo {
        private MSA.Variable _transformedVariable;
        private MSA.Variable.VariableKind _kind;

        public MSA.Variable.VariableKind Kind {
            get { return _kind; }
            set { _kind = value; }
        }
        
        public MSA.Variable TransformedVariable {
            get { return _transformedVariable; }
            set { _transformedVariable = value; }
        }

        public VariableInfo(MSA.Variable.VariableKind kind) {
            _kind = kind;
        }
    }

    [Serializable]
    public sealed class LexicalScope {
        private readonly Dictionary<SymbolId, VariableInfo>/*!*/ _locals;
        private readonly LexicalScope _parentScope;
        private readonly ScopeKind _kind;

        public ScopeKind Kind { get { return _kind; } }
        public LexicalScope ParentScope { get { return _parentScope; } }
        public Dictionary<SymbolId, VariableInfo>/*!*/ LocalVariables { get { return _locals; } }

        public LexicalScope(LexicalScope parentScope, ScopeKind kind)
            : this(parentScope, kind, new Dictionary<SymbolId, VariableInfo>()) {
        }

        public LexicalScope(LexicalScope parentScope, ScopeKind kind, Dictionary<SymbolId, VariableInfo>/*!*/ locals) {
            Debug.Assert(locals != null);

            _parentScope = parentScope;
            _kind = kind;
            _locals = locals;
        }

        public bool HasLocal(string id) {
            return HasLocal(SymbolTable.StringToId(id));
        }
        
        // returns true if id already exists here or in an outer scope
        public bool HasLocal(SymbolId id) {
            if (DefinedStatically(id)) // || DefinedDynamically(id))
                return true;
            else if (_kind == ScopeKind.Block)
                return _parentScope.HasLocal(id);
            else
                return false;
        }

        public LocalVariable AddParameter(SymbolId id, SourceSpan location) {
            Debug.Assert(!HasLocal(id));
            return CreateLocalHere(id, MSA.Variable.VariableKind.Parameter, location);
        }
        
        public Variable AddLocal(SymbolId id, SourceSpan location) {
            if (DefinedStatically(id))
                return new LocalVariable(this, id, location);
            //else if (DefinedDynamically(id))
            //    return new DynamicLocalVar(id, this, location);
            else if (HasLocal(id))
                return GetOuterLocal(id, this, 0, location);
            else
                return CreateLocalHere(id, MSA.Variable.VariableKind.Local, location);
        }

        private LocalVariable CreateLocalHere(SymbolId id, MSA.Variable.VariableKind kind, SourceSpan location) {
            _locals.Add(id, new VariableInfo(kind));
            return new LocalVariable(this, id, location);
        }

        private Variable GetOuterLocal(SymbolId id, LexicalScope blockScope, int depth, SourceSpan location) {
            if (DefinedStatically(id))
                return new LocalVariable(blockScope, id, /*depth, */location);
            //else if (DefinedDynamically(id))
            //    return new DynamicLocalVar(id, blockScope, /*depth, */location);
            else {
                Debug.Assert(_kind == ScopeKind.Block);
                return _parentScope.GetOuterLocal(id, blockScope, depth + 1, location);
            }
        }

        private bool DefinedStatically(SymbolId id) {
            return _locals.ContainsKey(id);
        }

        //public Variable add_locally(string id, SourceSpan location) {
        //    if (defined_statically(id))
        //        return new StaticLocalVar(this, id, location);
        //    else
        //        return create_local_here(id, location);
        //}

        //public bool DefinedDynamically(SymbolId id) {
        //    return false;
        //}

        internal void TransformLocals(MSA.CodeBlock/*!*/ methodBlock, AstGenerator/*!*/ gen) {
            Utils.Assert.NotNull(methodBlock, gen);

            foreach (KeyValuePair<SymbolId, VariableInfo> entry in _locals) {
                if (entry.Value.TransformedVariable == null) {
                    entry.Value.TransformedVariable = methodBlock.CreateVariable(entry.Key, entry.Value.Kind, typeof(object));
                } else {
                    Debug.Assert(entry.Value.Kind == MSA.Variable.VariableKind.Parameter);
                }
            }
        }

        /// <summary>
        /// Updates local variable table on this scope with transformed parameters.
        /// </summary>
        internal MSA.Variable[]/*!*/ TransformParameters(Parameters parameters, MSA.CodeBlock/*!*/ methodBlock, 
            int hiddenParameterCount, AstGenerator/*!*/ gen) {

            int paramCount = hiddenParameterCount;
            
            if (parameters == null) {
                return new MSA.Variable[0];
            }

            if (parameters.Mandatory != null) { 
                paramCount += parameters.Mandatory.Count;
            }

            if (parameters.Optional != null) {
                paramCount += parameters.Optional.Count;
            }

            if (parameters.Array != null) {
                paramCount += 1;
            }

            MSA.Variable[] result = new MSA.Variable[paramCount];

            int dlrParamIndex = hiddenParameterCount;

            if (parameters.Mandatory != null) {
                for (int i = 0; i < parameters.Mandatory.Count; i++) {
                    SymbolId paramName = parameters.Mandatory[i].Name;
                    MSA.Variable paramVar = MSA.Variable.Parameter(methodBlock, paramName, typeof(object));

                    SetTransformedParameter(paramName, paramVar);

                    result[dlrParamIndex++] = paramVar;
                }
            }

            if (parameters.Optional != null) {
                for (int i = 0; i < parameters.Optional.Count; i++) {
                    SimpleAssignmentExpression p = parameters.Optional[i];

                    SymbolId paramName = ((LocalVariable)p.Left).Name;
                    MSA.Expression defaultExpr = p.Right.TransformRead(gen);
                    MSA.Variable paramVar = MSA.Variable.Parameter(methodBlock, paramName, typeof(object), defaultExpr);

                    SetTransformedParameter(paramName, paramVar);

                    result[dlrParamIndex++] = paramVar;
                }
            }

            if (parameters.Array != null) {
                SymbolId paramName = parameters.Array.Name;
                MSA.Variable paramVar = MSA.Variable.Parameter(methodBlock, paramName, typeof(object));

                SetTransformedParameter(paramName, paramVar);

                result[dlrParamIndex++] = paramVar;
            }
            
            Debug.Assert(dlrParamIndex == result.Length);

            return result;
        }

        internal void SetTransformedParameter(SymbolId name, MSA.Variable/*!*/ variable) {
            Debug.Assert(_locals[name].Kind == MSA.Variable.VariableKind.Parameter && variable != null);
            _locals[name].TransformedVariable = variable;
        }

        internal VariableInfo/*!*/ ResolveVariable(SymbolId name) {
            
            LexicalScope scope = this;
            do {
                VariableInfo var;
                if (scope.LocalVariables.TryGetValue(name, out var)) {
                    return var;
                }
                scope = scope.ParentScope; // TODO: does parenting go thru scope boundaries?
            } while (scope != null);

            // TODO: soft error
            throw new NotImplementedException("undefined variable");
        }
    }
}

