/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System;
using System.IO;
using System.Collections.Generic;
using System.Text;

using Microsoft.Scripting;

using gppg;
using Ruby.Compiler.AST;
using Microsoft.Scripting.Hosting;
using System.Diagnostics;

namespace Ruby.Compiler {

    class SyntaxError : Exception {
        public SyntaxError(string message)
            : base(message) {

        }
    }

    // The non-autogenerated part of the Parser class.
    public partial class Parser {
        private int in_single = 0;
        private int in_def = 0;

        public SourceUnitTree _ast;
        public List<Initializer> _initializers; // lazy

        private Stack<LexicalScope>/*!*/ _lexicalScopes = new Stack<LexicalScope>();
        private SourceUnit _sourceUnit;
        private Tokenizer _tokenizer;
        private ErrorSink _errorSink;
        private ParserSink _parserSink;
        private int _generatedNameId;

        /// <summary>
        /// Used in "FOR block-var IN expression DO statement-list END" production 
        /// to carry the value from block-var to statement-list.
        /// </summary>
        //private LexicalScope _forLoopBodyScope;

        public Tokenizer Tokenizer {
            get { return _tokenizer; }
        }

        public ErrorSink ErrorSink {
            get { return _errorSink; }
        }

        protected override SourceSpan TokenSpan {
            get {
                return _tokenizer.TokenSpan;
            }
        }

        protected override TokenValue TokenValue {
            get {
                return _tokenizer.TokenValue;
            }
        }

        protected override int GetNextToken() {
            return (int)_tokenizer.GetNextToken();
        }

        protected override void ReportSyntaxError(string message) {
            // TODO:
            ErrorSink.Add(_sourceUnit, message, TokenSpan, -1, Severity.FatalError);
            throw new SyntaxError(message);
        }

        private SymbolId GenerateErrorLocalName() {
            return SymbolTable.StringToId("error#" + _generatedNameId++);
        }

        private SymbolId GenerateErrorConstantName() {
            return SymbolTable.StringToId("Error#" + _generatedNameId++);
        }

        public SourceUnitTree Parse(CompilerContext/*!*/ cc) {
            Utils.Assert.NotNull(cc);
            Debug.Assert(cc.Options is RubyCompilerOptions);

            RubyCompilerOptions options = (RubyCompilerOptions)cc.Options;

            _errorSink = cc.Errors;
            _parserSink = cc.ParserSink;

            _lexicalScopes.Clear();
            
            EnterScope(options.LexicalScope ?? new LexicalScope(options.LexicalScope, ScopeKind.SourceUnit));

            using (SourceUnitReader reader = cc.SourceUnit.GetReader()) {
                _sourceUnit = cc.SourceUnit;
                _tokenizer = new Tokenizer(this);
                _tokenizer.Initialize(null, reader, SourceLocation.MinValue);

                try {
                    Parse();
                } catch (SyntaxError) {
                    return null;
                } finally {
                    LeaveScope();
                }

                return _ast;
            }
        }

        public SourceUnitTree ParseInteractiveCode(CompilerContext/*!*/ cc, out InteractiveCodeProperties props) {
            Utils.Assert.NotNull(cc);

            SourceUnitTree tree = Parse(cc);
            
            if (_tokenizer.Errors.AnyError) {
                props = InteractiveCodeProperties.IsInvalid;
                if (_tokenizer.UnterminatedToken) {
                    props |= InteractiveCodeProperties.IsIncompleteToken;
                }
            } else {
                props = InteractiveCodeProperties.None;
            }

            return tree;
        }


        // ------------------------------------------------------------------------

        private LocalVariable DefineParameter(SymbolId id, SourceSpan location) {
            if (ID.GetNamespaceType(id) != NamespaceType.LOCAL) {
				_tokenizer.SoftError("formal argument must be a local variable");
	        } else if (CurrentScope.HasLocal(id)) {
				_tokenizer.SoftError("duplicate argument name");
			}
            return CurrentScope.AddParameter(id, location);	
        }

        private Initializer/*!*/ AddInitializer(Initializer/*!*/ initializer) {
            if (_initializers == null) {
                _initializers = new List<Initializer>();
            }

            _initializers.Add(initializer);

            return initializer;
        }
        
        protected override SourceSpan MergeLocations(SourceSpan start, SourceSpan end) {
            if (end == SourceSpan.None)
                return start;

            SourceSpan merged = start;
            // TODO:
            //SourceSpan merged = new SourceSpan(
            //    new SourceLocation(start.Start.Index, start.Start.Column, start.Start.Line),
            //    new SourceLocation(end.End.Index, end.End.Column, end.End.Line)
            //);
            
            return merged;
        }

        private LexicalScope/*!*/ EnterScope(LexicalScope/*!*/ scope) {
            Utils.Assert.NotNull(scope);
            _lexicalScopes.Push(scope);
            return scope;
        }
        
        private LexicalScope EnterNewScope(ScopeKind kind) {
            LexicalScope result = new LexicalScope(CurrentScope, kind);
            _lexicalScopes.Push(result);
            return result;
        }

        private LexicalScope LeaveScope() {
            return _lexicalScopes.Pop();
        }

        public LexicalScope CurrentScope {
            get {
                Debug.Assert(_lexicalScopes.Count > 0);
                return _lexicalScopes.Peek();
            }
        }

        private Expression gettable(string id, SourceSpan location) {
            return gettable(id, location);
        }

        private Expression gettable(SymbolId id, SourceSpan location) {
            if (id == Symbols.Self)
                return new SelfReference(location);
            else if (id == Symbols.Nil)
                return Literal.Nil(location);
            else if (id == Symbols.True)
                return Literal.True(location);
            else if (id == Symbols.False)
                return Literal.False(location);
            else if (id == Symbols.File) {
                SourceFileUnit sfu = _sourceUnit as SourceFileUnit;
                return new Literal(sfu != null ? sfu.Path : "", location);
            } else if (id == Symbols.Line)
                return new Literal(Tokenizer.TokenStart.Line, location);
            else if (ID.GetNamespaceType(id) == NamespaceType.LOCAL) {
                if (CurrentScope.HasLocal(id))
                    return CurrentScope.AddLocal(id, location);
                else
                    return new MethodCall(null, id, new Arguments(location), location);
            } else if (ID.GetNamespaceType(id) == NamespaceType.GLOBAL) {
                if (id == Symbols.UnderscoreGlobalVariable)
                    return CurrentScope.AddLocal(Symbols.UnderscoreGlobalVariable, location);
                else
                    return new GlobalVariable(id, location);
            } else if (ID.GetNamespaceType(id) == NamespaceType.INSTANCE)
                return new InstanceVariable(id, location);
            else if (ID.GetNamespaceType(id) == NamespaceType.CONST)
                return new ConstantVariable(CurrentScope, id, location);
            else if (ID.GetNamespaceType(id) == NamespaceType.CLASS)
                return new ClassVariable(CurrentScope, id, location);
            else {
                Tokenizer.SoftError("identifier {0} is not valid", id.ToString());
                return Literal.Nil(location);
            }
        }

        private LeftValue assignable(SymbolId id, SourceSpan location) {
            if (id == Symbols.Self) {
                Tokenizer.SoftError("Can't change the value of self");
                return CurrentScope.AddLocal(GenerateErrorLocalName(), location);
            } else if (id == Symbols.Nil) {
                Tokenizer.SoftError("Can't assign to nil");
                return CurrentScope.AddLocal(GenerateErrorLocalName(), location);
            } else if (id == Symbols.True) {
                Tokenizer.SoftError("Can't assign to true");
                return CurrentScope.AddLocal(GenerateErrorLocalName(), location);
            } else if (id == Symbols.False) {
                Tokenizer.SoftError("Can't assign to false");
                return CurrentScope.AddLocal(GenerateErrorLocalName(), location);
            } else if (id == Symbols.File) {
                Tokenizer.SoftError("Can't assign to __FILE__");
                return CurrentScope.AddLocal(GenerateErrorLocalName(), location);
            } else if (id == Symbols.Line) {
                Tokenizer.SoftError("Can't assign to __LINE__");
                return CurrentScope.AddLocal(GenerateErrorLocalName(), location);
            } else if (ID.GetNamespaceType(id) == NamespaceType.LOCAL) {
                return CurrentScope.AddLocal(id, location);
            } else if (ID.GetNamespaceType(id) == NamespaceType.GLOBAL) {
                if (id == Symbols.UnderscoreGlobalVariable)
                    return CurrentScope.AddLocal(Symbols.UnderscoreGlobalVariable, location);
                else
                    return new GlobalVariable(id, location);
            } else if (ID.GetNamespaceType(id) == NamespaceType.INSTANCE) {
                return new InstanceVariable(id, location);
            } else if (ID.GetNamespaceType(id) == NamespaceType.CONST) {
                return new ConstantVariable(CurrentScope, id, location);
            } else if (ID.GetNamespaceType(id) == NamespaceType.CLASS) {
                return new ClassVariable(CurrentScope, id, location);
            } else {
                Tokenizer.SoftError("identifier {0} is not valid", id.ToString());
            }

            return null;
        }

        private void backref_error(Node node) {
            if (node is NTH_REF)
                Tokenizer.SoftError("Cant set variable ${0}", ((NTH_REF)node).number);
            else if (node is BACK_REF)
                Tokenizer.SoftError("Cant set variable ${0}", ((BACK_REF)node).ch);
        }

        private List<T> MakeListOpt<T>(T item) where T : class {
            if (item != null) {
                List<T> result = new List<T>();
                result.Add(item);
                return result;
            } else {
                return null;
            }    
        }

        private List<Expression>/*!*/ CheckHashExpressions(List<Expression>/*!*/ expressions, SourceSpan location) {
            Utils.Assert.NotNull(expressions);

            if (expressions.Count % 2 != 0) {
                _errorSink.Add(_sourceUnit, "odd number list for Hash", location, -1, Severity.Error);
                expressions.Add(Literal.Nil(SourceSpan.None));
            }

            return expressions;
        }

        private Arguments/*!*/ RequireNoBlockArg(Arguments/*!*/ arguments) {
            Utils.Assert.NotNull(arguments);

            if (arguments.Block != null) {
                _errorSink.Add(_sourceUnit, "block argument should not be given", arguments.Block.Location, -1, Severity.Error);
                arguments.Block = null;
            }

            return arguments;
        }

        private Statement MakeLoopStatement(Statement/*!*/ statement, Expression/*!*/ condition, bool isWhileLoop, SourceSpan location) {
            ExpressionStatement exprStmt;
            BlockExpression blockExpr;

            bool isPostTest = ((exprStmt = statement as ExpressionStatement) != null && (blockExpr = exprStmt.Expression as BlockExpression) != null);
            return new WhileLoopStatement(CurrentScope, condition, isPostTest, isWhileLoop, statement, location);
        }
    }
}


