/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;

using Microsoft.Scripting.Hosting;
using Microsoft.Scripting;

namespace Ruby.Compiler {

    public sealed class RubyTokenCategorizer : TokenCategorizer {

        private Tokenizer _tokenizer;

        public override SourceLocation CurrentPosition {
            get { return _tokenizer.TokenStart; }
        }

        public override object CurrentState {
            get { return _tokenizer.CurrentState; }
        }

        public override bool IsRestartable {
            get { return false; }
        }

        public override ErrorSink ErrorSink {
            get {
                return _tokenizer.Errors;
            }
            set {
                _tokenizer.Errors = value;
            }
        }

        public RubyTokenCategorizer() : base() {
            _tokenizer = new Tokenizer(new ErrorSink(), true);
        }

        public override void Initialize(object state, SourceUnitReader sourceReader, SourceLocation initialLocation) {
            _tokenizer.Initialize(state, sourceReader, initialLocation);
        }

        public override bool SkipToken() {
            return (Tokens)_tokenizer.GetNextToken() != Tokens.EOF;
        }

        public override TokenInfo ReadToken() {

            TokenInfo result = new TokenInfo();

            Tokens token = (Tokens)_tokenizer.GetNextToken();
            result.SourceSpan = _tokenizer.TokenSpan;

            // TODO:
            switch (token) {
                case Tokens.kUNDEF:
                case Tokens.kRESCUE:
                case Tokens.kENSURE:
                case Tokens.kIF:
                case Tokens.kUNLESS:
                case Tokens.kTHEN:
                case Tokens.kELSIF:
                case Tokens.kELSE:
                case Tokens.kCASE:
                case Tokens.kWHEN:
                case Tokens.kWHILE:
                case Tokens.kUNTIL:
                case Tokens.kFOR:
                case Tokens.kBREAK:
                case Tokens.kNEXT:
                case Tokens.kREDO:
                case Tokens.kRETRY:
                case Tokens.kIN:
                case Tokens.kRETURN:
                case Tokens.kYIELD:
                case Tokens.kSUPER:
                case Tokens.kSELF:
                case Tokens.kNIL:
                case Tokens.kTRUE:
                case Tokens.kFALSE:
                case Tokens.kAND:
                case Tokens.kOR:
                case Tokens.kNOT:
                case Tokens.kIF_MOD:
                case Tokens.kUNLESS_MOD:
                case Tokens.kWHILE_MOD:
                case Tokens.kUNTIL_MOD:
                case Tokens.kRESCUE_MOD:
                case Tokens.kALIAS:
                case Tokens.kDEFINED:
                case Tokens.k__LINE__:
                case Tokens.k__FILE__:
                    result.Category = TokenCategory.Keyword;
                    break;

                case Tokens.kDEF:
                case Tokens.kCLASS:
                case Tokens.kMODULE:
                case Tokens.kEND:
                case Tokens.kBEGIN:
                case Tokens.klBEGIN:
                case Tokens.klEND:
                case Tokens.kDO:
                case Tokens.kDO_COND:
                case Tokens.kDO_BLOCK:
                    result.Category = TokenCategory.Keyword;
                    result.Trigger = TokenTriggers.MatchBraces;
                    break;

                case Tokens.tUPLUS:
                case Tokens.tUMINUS:
                case Tokens.tUMINUS_NUM:
                case Tokens.tPOW:
                case Tokens.tCMP:
                case Tokens.tEQ:
                case Tokens.tEQQ:
                case Tokens.tNEQ:
                case Tokens.tGEQ:
                case Tokens.tLEQ:
                case Tokens.tANDOP:
                case Tokens.tOROP:
                case Tokens.tMATCH:     // =~
                case Tokens.tNMATCH:    // !~
                case Tokens.tDOT2:      // ..
                case Tokens.tDOT3:      // ...
                case Tokens.tAREF:      // []
                case Tokens.tASET:      // []=
                case Tokens.tLSHFT:
                case Tokens.tRSHFT:
                case Tokens.tASSOC:
                case Tokens.tSTAR:
                case Tokens.tAMPER:
                case Tokens.tOP_ASGN:
                    result.Category = TokenCategory.Operator;
                    break;

                case Tokens.tSEPARATING_DOUBLE_COLON:
                case Tokens.tLEADING_DOUBLE_COLON:
                    result.Category = TokenCategory.Delimiter;
                    result.Trigger = TokenTriggers.MemberSelect;
                    break;
                
                case Tokens.tLBRACK:      // [
                case Tokens.tLBRACE:      // {
                case (Tokens)'{':
                case Tokens.tLBRACE_ARG:  // <whitespacce>{
                case (Tokens)'}':
                case (Tokens)']':
                case (Tokens)'|':
                    result.Category = TokenCategory.Grouping;
                    result.Trigger = TokenTriggers.MatchBraces;
                    break;

                case Tokens.tLPAREN:      // (
                case Tokens.tLPAREN_ARG:  // <whitespace>(
                    result.Category = TokenCategory.Grouping;
                    result.Trigger = TokenTriggers.MatchBraces | TokenTriggers.ParameterStart;
                    break;
                
                case (Tokens)')':
                    result.Category = TokenCategory.Grouping;
                    result.Trigger = TokenTriggers.MatchBraces | TokenTriggers.ParameterEnd;
                    break;    

                case (Tokens)',':
                    result.Category = TokenCategory.Delimiter;
                    result.Trigger = TokenTriggers.ParameterNext;
                    break;

                case (Tokens)'.':
                    result.Category = TokenCategory.Delimiter;
                    result.Trigger = TokenTriggers.MemberSelect;
                    break;
                
                case Tokens.tSTRING_END:
                    result.Category = TokenCategory.StringLiteral;
                    break;

                case Tokens.tSTRING_EMBEDDED_VARIABLE_BEGIN: // # in string followed by @ or $
                case Tokens.tSTRING_EMBEDDED_CODE_BEGIN: // # in string followed by {
                    result.Category = TokenCategory.Delimiter;
                    break;

                case Tokens.tIDENTIFIER:
                case Tokens.tFID:
                case Tokens.tGVAR:
                case Tokens.tIVAR:
                case Tokens.tCONSTANT:
                case Tokens.tCVAR:
                case Tokens.tNTH_REF:
                case Tokens.tBACK_REF:
                    result.Category = TokenCategory.Identifier;
                    break;
                
                case Tokens.tINTEGER:
                case Tokens.tFLOAT:
                    result.Category = TokenCategory.NumericLiteral;
                    break;

                case Tokens.tSTRING_CONTENT:
                case Tokens.tSTRING_BEG:
                case Tokens.tSHELL_STRING_BEGIN:
                case Tokens.tWORDS_BEG:
                case Tokens.tVERBATIM_WORDS_BEGIN:
                    result.Category = TokenCategory.StringLiteral;
                    break;

                case Tokens.tREGEXP_END:
                case Tokens.tREGEXP_BEG:
                    result.Category = TokenCategory.RegularExpressionLiteral;
                    break;

                case Tokens.tSYMBEG:
                    result.Category = TokenCategory.Delimiter;
                    break;

                case (Tokens)'#':
                    result.Category = TokenCategory.LineComment;
                    break;

                case (Tokens)'\n':
                    result.Category = TokenCategory.WhiteSpace;
                    break;
                
                case Tokens.EOF:
                    result.Category = TokenCategory.EndOfStream;
                    break;

                default:
                    result.Category = TokenCategory.None;
                    break;
            }

            return result;
        }
    }

}
