/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

#if NEW_PARSER

#define DUMP_TOKENS

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Scripting;
using Microsoft.Scripting.Hosting;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace Ruby.Compiler {

    //public enum Lex_State {
    //    EXPR_BEG,
    //    EXPR_END,
    //    EXPR_ARG,
    //    EXPR_CMDARG,
    //    EXPR_ENDARG,
    //    EXPR_MID,
    //    EXPR_FNAME,
    //    EXPR_DOT,
    //    EXPR_CLASS
    //}

    /* eolns:  \n                        
     * BEG     skip                
     * END     -> BEG, Token.EOLN
     * ARG     -> BEG, Token.EOLN
     * CMDARG  -> BEG, Token.EOLN
     * ENDARG  -> BEG, Token.EOLN
     * MID     -> BEG, Token.EOLN
     * FNAME   skip
     * DOT     skip
     * CLASS   skip
     * 
     * -> END:
     * end of here-doc
     * 
    */

    //public struct TokenSemanticValue {
    //    public Node node;
    //    public LVALUE lval;
    //    public string id;
    //    public int num;
    //    public Terminator terminator;
    //}
    
    public enum LexicalState {
        Initial,
        MultilineComment,
        FunctionName,
    }

    public sealed class Tokenizer {
        private const int EOF = -1;
        private const int DefaultBufferCapacity = 1024;

        private struct State {
            public LexicalState LexicalState;
            
            public State(State state) {
                LexicalState = state.LexicalState;
            }

            public State(object dummy) {
                LexicalState = LexicalState.Initial;
            }
        }

        private State _state;

        // lex_strterm - string terminator "blah #x foo" -> "blah " + #x + " foo"
        // lex_state

        // CMDARG_PUSH(int number)
        // CMDARG_POP
        // COND_PUSH
        // COND_POP

        // tokenizer properties:
        private SourceUnit _sourceUnit;
        private TokenizerBuffer _buffer;
        private ErrorSink _errors;

        public object CurrentState {
            get {
                return _state;
            }
        }

        public SourceUnit SourceUnit {
            get {
                return _sourceUnit;
            }
        }

        public ErrorSink Errors {
            get { return _errors; }
            set {
                if (value == null) throw new ArgumentNullException("value");
                _errors = value;
            }
        }

        internal bool IsEndOfFile {
            get {
                return _buffer.Peek() == EOF;
            }
        }

        internal SourceSpan TokenStart {
            get {
                return _buffer.TokenStart;
            }
        }

        internal SourceSpan TokenEnd {
            get {
                return _buffer.TokenEnd;
            }
        }

        internal SourceSpan TokenSpan {
            get {
                return _buffer.TokenSpan;
            }
        }

        public Tokenizer(ErrorSink errorSink) {
            if (errorSink == null) throw new ArgumentNullException("errorSink");

            _errors = errorSink;
            _state = new State(null);
            _sourceUnit = null;
            _buffer = null;
        }

        public void Initialize(SourceUnit sourceUnit) {
            if (sourceUnit == null) throw new ArgumentNullException("sourceUnit");

            Initialize(null, sourceUnit.GetReader(), SourceSpan.MinValue, DefaultBufferCapacity);
        }

        public void Initialize(object state, SourceUnitReader sourceReader, SourceSpan initialLocation) {
            Initialize(state, sourceReader, initialLocation, DefaultBufferCapacity);
        }

        public void Initialize(object state, SourceUnitReader sourceReader, SourceSpan initialLocation, int bufferCapacity) {
            if (sourceReader == null) throw new ArgumentNullException("sourceReader");

            if (state != null) {
                if (!(state is State)) throw new ArgumentException();
                _state = new State((State)state);
            } else {
                _state = new State(null);
            }

            _sourceUnit = sourceReader.SourceUnit;

            // TODO: we can reuse the buffer if there is enough free space:
            _buffer = new TokenizerBuffer(sourceReader, initialLocation, bufferCapacity, true);

            DumpBeginningOfUnit();
        }

        public int ReadToken() {
            if (_buffer == null) throw new InvalidOperationException("Uninitialized");
            return (int)GetNextToken();
        }

        internal Tokens GetNextToken() {

            _buffer.DiscardToken();

            switch (_state.LexicalState) {
                case LexicalState.Initial: return ReadInitialToken();
                case LexicalState.FunctionName: return ReadFunctionName();

                default:
                    Debug.Assert(false, "");
                    return Tokens.None;
            }
        }

        private Tokens ReadInitialToken() {
           Debug.Assert(_state.LexicalState == LexicalState.Initial);

           int ch = _buffer.Read();

            switch (ch) {
                case -1: 
                    _buffer.MarkSingleLineTokenEnd();
                    return Tokens.EndOfFile;

                case '#':
                    _buffer.ReadLine();
                    _buffer.MarkSingleLineTokenEnd();
                    return Tokens.SingleLineComment;

                // =, =>, ==, ===, =begin, =end
                case '=':
                    return ReadInitialEqualToken();
                
                case ' ':
                case '\t':
                case '\r':
                    break;

                case '\n':
                    break;

                case '"':
                    break;

                case '\'':
                    break;

                case '%':
                    break;

                case '`':
                    break;

                case '0':
                case '1':
                case '2':
                case '3':
                case '4':
                case '5': 
                case '6': 
                case '7': 
                case '8': 
                case '9':
                    return ReadNumericLiteral(ch);

                case '?':
                    break;

                case '!':
                case '&':
                case '*':
                case '+':
                case '-':
                case '/':
                case '^':
                case '|':
                case '>':
                case '<':
                case '~':

                    break;

                case ':':
                    // TODO:
                    return Tokens.tSYMBEG;

                case '$':
                    return ReadGlobalName();

                case '@':
                    if (_buffer.Read('@')) {
                        if (SkipAlphanumeric()) {
                            // class variable:
                            _buffer.MarkSingleLineTokenEnd();
                            return Tokens.tCVAR;
                        } 
                    } else if (SkipAlphanumeric()) {
                        // instance variable:
                        _buffer.MarkSingleLineTokenEnd();
                        return Tokens.tIVAR;
                    }

                    _buffer.MarkSingleLineTokenEnd();
                    return Tokens.UnexpectedCharacter;

                default:
                    Tokens result;
                    
                    if ((result = SkipKeyword(ch)) != Tokens.None) {
                        _buffer.MarkSingleLineTokenEnd();
                        return result;
                    }

                    if (ch >= 'A' && ch <= 'Z') {
                        SkipAlphanumeric();
                        _buffer.MarkSingleLineTokenEnd();
                        return Tokens.tCONSTANT;
                    } else if (ch >= 'a' && ch <= 'z' || ch == '_') {
                        SkipAlphanumeric();
                        _buffer.MarkSingleLineTokenEnd();
                        return Tokens.tIDENTIFIER;
                    }
                    break;
            }

            _buffer.MarkSingleLineTokenEnd();
            return Tokens.UnexpectedCharacter;
        }

        private Tokens ReadFunctionName() {
            Debug.Assert(_state.LexicalState == LexicalState.FunctionName);
            return Tokens.Error;
        }

        /// <summary>
        /// Skips [_a-zA-Z0-9]*
        /// Returns <c>true</c> if some characters were read.
        /// </summary>
        private bool SkipAlphanumeric() {
            int ch = TokenizerBuffer.InvalidCharacter;

            do { ch = _buffer.Read(); } 
            while (ch >= 'A' && ch <= 'Z' || ch >= 'a' && ch <= 'z' && ch >= '0' && ch <= '9' || ch == '_');

            _buffer.Back();
            return ch != TokenizerBuffer.InvalidCharacter;
        }

        private Tokens ReadInitialEqualToken() {
            Debug.Assert(_buffer.GetCharRelative(-1) == '=');
            
            int ch = _buffer.Peek();

            // ==, ===
            if (ch == '=') {
                _buffer.SeekRelative(+1);
                if (_buffer.Read('=')) {
                    _buffer.MarkSingleLineTokenEnd();
                    return Tokens.tEQQ;
                }
                _buffer.MarkSingleLineTokenEnd();
                return Tokens.tEQ;
            }

            // =>
            if (ch == '>') {
                _buffer.MarkSingleLineTokenEnd();
                return Tokens.tASSOC;
            }

            // =begin
            if (_buffer.Read("begin")) {
                _buffer.ReadLine();
                _state.LexicalState = LexicalState.MultilineComment; // TODO: PUSH?
                return Tokens.MultiLineComment;
            }

            // =end
            if (_buffer.Read("end")) {
                _buffer.ReadLine();
                _buffer.MarkMultiLineTokenEnd();
                _state.LexicalState = LexicalState.Initial; // TODO: POP?
                return Tokens.MultiLineComment;
            }

            _buffer.MarkSingleLineTokenEnd();
            return Tokens.tOP_ASGN;
        }

        private Tokens ReadGlobalName() {
            Debug.Assert(_buffer.GetCharRelative(-1) == '$');

            // GLOBAL-VARIABLE:
            // [$][_a-zA-Z][_a-zA-Z0-9]* 
            // [$][~`_\\>=<;:/.+,-*'&$"!]
            // [$]-[0-9_a-zA-Z]
            
            // tNTH_REF,                // [$][0-9]+
            // tBACK_REF,               // [$][`'+&]

            // TODO:

            _buffer.MarkSingleLineTokenEnd();
            return Tokens.tGVAR;
        }

        private Tokens ReadNumericLiteral(int ch) {
            // INTEGER:
            // [1-9]([0-9_]*[1-9])?
            // 0([0-7_]*[0-7])?
            // 0[xX][0-9a-fA-F]([0-9a-fA-F_]*[0-9a-fA-F])?
            // 0[dD][0-9]([0-9_]*[0-9])?
            // 0[bB][01]([01_]*[01])?
            // 0[oO][0-7]([0-7_]*[0-7])?
            // [?]{char}
            
            // {char} is
            //   (\\(c|C-|M-))*([^\\]|\\[abefnrstv]|\\[0-7]{1,3}|\\x[0-9a-fA-F]{1-2})

            // FLOAT:
            // (0|[1-9]([0-9_]*[0-9])?)[.][0-9_]*[0-9]([eE][+-]?[0-9]([0-9_]*[0-9])?)
        
            
            // TODO:
            return Tokens.tINTEGER;
        }

        private Tokens SkipKeyword(int ch) {
            // TODO: generate automaton for keywords

            if (_buffer.Read("do")) return Tokens.k__FILE__;
            if (_buffer.Read("if")) return Tokens.kIF;
            if (_buffer.Read("in")) return Tokens.kIN;
            if (_buffer.Read("or")) return Tokens.kOR;
            if (_buffer.Read("nil")) return Tokens.kNIL;
            if (_buffer.Read("not")) return Tokens.kNOT;
            if (_buffer.Read("and")) return Tokens.kAND;
            if (_buffer.Read("def")) return Tokens.kDEF;
            if (_buffer.Read("for")) return Tokens.kFOR;
            if (_buffer.Read("end")) return Tokens.kEND;
            if (_buffer.Read("END")) return Tokens.klEND;
            if (_buffer.Read("next")) return Tokens.kNEXT;
            if (_buffer.Read("redo")) return Tokens.kREDO;
            if (_buffer.Read("then")) return Tokens.kTHEN;
            if (_buffer.Read("true")) return Tokens.kTRUE;
            if (_buffer.Read("self")) return Tokens.kSELF;
            if (_buffer.Read("case")) return Tokens.kCASE;
            if (_buffer.Read("when")) return Tokens.kWHEN;
            if (_buffer.Read("retry")) return Tokens.kRETRY;
            if (_buffer.Read("super")) return Tokens.kSUPER;
            if (_buffer.Read("undef")) return Tokens.kUNDEF;
            if (_buffer.Read("until")) return Tokens.kUNTIL;
            if (_buffer.Read("while")) return Tokens.kWHILE;
            if (_buffer.Read("yield")) return Tokens.kYIELD;
            if (_buffer.Read("BEGIN")) return Tokens.kBEGIN;
            if (_buffer.Read("alias")) return Tokens.kALIAS;
            if (_buffer.Read("begin")) return Tokens.kBEGIN;
            if (_buffer.Read("break")) return Tokens.kBREAK;
            if (_buffer.Read("class")) return Tokens.kCLASS;
            if (_buffer.Read("elsif")) return Tokens.kELSIF;
            if (_buffer.Read("false")) return Tokens.kFALSE;
            if (_buffer.Read("ensure")) return Tokens.kENSURE;
            if (_buffer.Read("module")) return Tokens.kMODULE;
            if (_buffer.Read("rescue")) return Tokens.kRESCUE;
            if (_buffer.Read("return")) return Tokens.kRETURN;
            if (_buffer.Read("unless")) return Tokens.kUNLESS;
            if (_buffer.Read("defined?")) return Tokens.kDEFINED;
            if (_buffer.Read("__FILE__")) return Tokens.k__FILE__;
            if (_buffer.Read("__LINE__")) return Tokens.k__LINE__;
            return Tokens.None;
            
            //"__FILE__", "__LINE__", "BEGIN", "END", "alias", "and", "begin", "break", "case", "class", "def",
            //"defined?", "do", "else", "elsif", "end", "ensure", "false", "for", "if",
            //"in", "module", "next", "nil", "not", "or", "redo", "rescue", "retry", "return",
            //"self", "super", "then", "true", "undef", "unless", "until", "when", "while", "yield",
        }

        [Conditional("DUMP_TOKENS")]
        private void DumpBeginningOfUnit() {
            Console.WriteLine("--- Source unit: '{0}' ---", _sourceUnit.Name);
        }

        [Conditional("DUMP_TOKENS")]
        private void DumpToken(Tokens token) {
            Console.WriteLine("{0} `{1}`", token, GetTokenString().Replace("\r", "\\r").Replace("\n", "\\n").Replace("\t", "\\t"));
        }

        private string GetTokenString() {
            return "";
        }
    }
}
#endif