/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

#if NEW_PARSER
using System;
using System.Collections.Generic;
using System.Text;

// true operators: .. ... ! not && and || or != !~ += *= /= %= ^= etc.

namespace Ruby.Compiler {
    public enum Tokens {
        None,
        Error,
        WhiteSpace,
        EndOfFile,
        UnexpectedCharacter,

        SingleLineComment,
        MultiLineComment,

        // keywords:
        kCLASS,kMODULE,kDEF,kUNDEF,kBEGIN,kRESCUE,kENSURE,kEND,kIF,kUNLESS,kTHEN,kELSIF,kELSE,
        kCASE,kWHEN,kWHILE,kUNTIL,kFOR,kBREAK,kNEXT,kREDO,kRETRY,kIN,kDO,kDO_COND,kDO_BLOCK,
        kRETURN,kYIELD,kSUPER,kSELF,kNIL,kTRUE,kFALSE,kAND,kOR,kNOT,kIF_MOD,kUNLESS_MOD,
        kWHILE_MOD,kUNTIL_MOD,kRESCUE_MOD,kALIAS,kDEFINED,klBEGIN,klEND,k__LINE__,k__FILE__,

        // % token<id:string>
        tCONSTANT,   // CONSTANT:            [A-Z][_a-zA-Z0-9]*                         
        tIDENTIFIER, // IDENTIFIER:          [_a-z][_a-zA-Z0-9]*                        
        tFID,        // FUNCTION-IDENTIFIER: [_a-zA-Z][_a-zA-Z0-9]*[!?=]?               
        
        tGVAR,       
        // GLOBAL-VARIABLE:
        // [$][_a-zA-Z][_a-zA-Z0-9]* 
        // [$][~`_\\>=<;:/.+,-*'&$"!]
        // [$]-[0-9_a-zA-Z]
        
        tIVAR,       // INSTANCE-VARIABLE:   @[_a-zA-Z][_a-zA-Z0-9]*                  
        tCVAR,       // CLASS-VARIABLE:      @@[_a-zA-Z][_a-zA-Z0-9]*                 
        
        // % token<id:string>
        tOP_ASGN,    // =, !=, %=, &=, *=, +=, -=, /=, ^=, |=, ||=, &&=, <<=, >>=, **=

        tINTEGER,              
        // [1-9]([0-9_]*[1-9])?
        // 0([0-7_]*[0-7])?
        // 0[xX][0-9a-fA-F]([0-9a-fA-F_]*[0-9a-fA-F])?
        // 0[dD][0-9]([0-9_]*[0-9])?
        // 0[bB][01]([01_]*[01])?
        // 0[oO][0-7]([0-7_]*[0-7])?
        // [?]{char}
        
        // {char} is
        //   (\\(c|C-|M-))*([^\\]|\\[abefnrstv]|\\[0-7]{1,3}|\\x[0-9a-fA-F]{1-2})
        //
        //   ... \c, \C- mean & 0x9f, \M- means | 0x80
        //   ... [abefnrstv] are escaped special characters
        //   ... \000 octal
        //   ... \xFF hex
        //
        // [-2**30, 2*30-1] -> Fixnum -- define CLR as 32bit platform?
        //        otherwise -> Bignum

        tFLOAT,
        // (0|[1-9]([0-9_]*[0-9])?)[.][0-9_]*[0-9]([eE][+-]?[0-9]([0-9_]*[0-9])?)
        
        tNTH_REF,                // [$][0-9]+
        tBACK_REF,               // [$][`'+&]

        tSTRING_BEG,             // ["']|%[qQ]?[:open:]
        tSTRING_END,             // ["']|[:close:]

        tSTRING_EMBEDDED_CODE_BEGIN,            // #          in string
        tSTRING_EMBEDDED_VARIABLE_BEGIN,            // #[{]       in string 
        
        tSTRING_CONTENT,         // string content w/o #
        
        tSHELL_STRING_BEGIN,            // `|%x[:open:]
        
        tREGEXP_BEG,             // (%r[:open:]|[/])
        tREGEXP_END,             // ([:close:]|[/])[eimnosux]* 

        tSYMBEG,                 // [:]["]?

        tWORDS_BEG,              // %W[:open:]
        tVERBATIM_WORDS_BEGIN,             // %w[:open:]

        tAREF,                   // []  (?)
        tASET,                   // []= (?)
        tASSOC,                  // =>
        
        tAMPER,                  // &
        tCMP,                    // <=>
        tEQ,                     // ==  (?)
        tEQQ,                    // === (?)
        tMATCH,                  // ~=
		GreaterThen,             // >
		LessThen,                // <
		tLEQ,                    // <=
		tLSHFT,                  // <<
		tRSHFT,                  // >>

		tUPLUS,                  // +@ (unary plus method on Numeric class)

        tUMINUS,                 // -@ (unary minus method on Numeric class)
        tUMINUS_NUM,             // - (unary minus ???)

        tPOW,                    // **
        tNEQ,                    // !=
        tGEQ,                    // >=
        tANDOP,                  // &&
        tOROP,                   // ||
        tNMATCH,                 // !~
		Mul,                     // *
		tSTAR,		             // * in arguments
		Div,                     // /
		Mod,                     // %
		Twiddle,                 // ~
		BackQuote,               // `
        Comma,                   // ,
        tDOT2,                   // ..
        tDOT3,                   // ...
        tSEPARATING_DOUBLE_COLON,                 // ::
        tLEADING_DOUBLE_COLON,                 // :::

        tLBRACK,                 // [

        tLBRACE,                 // {
        tLBRACE_ARG,             // ' {'
        
        tLPAREN,                 // (
        tLPAREN_ARG,             // ' (' 
    }
}
#endif