/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using System.Reflection;
using System.IO;
using System.Diagnostics;
using System.Threading;

using Microsoft.Scripting;
using Microsoft.Scripting.Generation;
using Microsoft.Scripting.Hosting;
using Microsoft.Scripting.Actions;
using MSA = Microsoft.Scripting.Ast;

using Ruby.Compiler;
using Ruby.Runtime;
using Ruby.Builtins;
   
namespace Ruby.Hosting {

    using SingletonFactory = SingletonEngineFactory<RubyEngine, RubyEngineOptions, RubyLanguageProvider>;
    
    public sealed class RubyEngine : ScriptEngine {

        private static readonly Guid RubyLanguageGuid = new Guid("F03C4640-DABA-473f-96F1-391400714DAB");
        private static readonly object ContextRegistrationKey = new object();

        private readonly ContextId _contextId;
        private readonly CodeContext/*!*/ _defaultContext;
        private readonly RubyActionBinder/*!*/ _actionBinder;
        private readonly Loader/*!*/ _loader;
        private RubyExecutionContext/*!*/ _executionContext;
        
        // HACK: casing switch
        private bool _rubyCaseMemberLists;

        internal bool RubyCaseMemberLists {
            get { return _rubyCaseMemberLists; }
            set { _rubyCaseMemberLists = value; }
        }

        public CodeContext/*!*/ DefaultContext {
            get {
                return _defaultContext;
            }
        }

        internal Loader/*!*/ Loader {
            get {
                return _loader;
            }
        }

        internal ContextId ContextId {
            get {
                return _contextId;
            }
        }

        internal RubyActionBinder/*!*/ ActionBinder {
            get { return _actionBinder; }
        }

        internal RubyExecutionContext/*!*/ ExecutionContext {
            get {
                return _executionContext;
            }
        }

        internal static readonly SingletonFactory/*!*/ Factory = new SingletonFactory(
            delegate(RubyLanguageProvider provider, RubyEngineOptions options) { return new RubyEngine(provider, options); },
            GetSetupInformation);

        public override string/*!*/ VersionString {
            get {
                return String.Format("IronRuby Pre-Alpha ({0}) on .NET {1}", RubyVersionInfo.Version, Environment.Version);
            }
        }

        public override Guid LanguageGuid {
            get {
                return RubyLanguageGuid;
            }
        }

        public new RubyEngineOptions/*!*/ Options {
            get { return (RubyEngineOptions)base.Options; }
        }

        public override ScriptCompiler/*!*/ Compiler {
            get {
                return new RubyScriptCompiler(this);
            }
        }

        // TODO: remove
        public override ActionBinder DefaultBinder {
            get {
                return _actionBinder;
            }
        }

        #region Construction

        /// <summary>
        /// Provides the current Ruby engine. Asks the current <see cref="ScriptEnvironment"/> for a one of it doesn't exist yet.
        /// </summary>
        public static RubyEngine/*!*/ CurrentEngine {
            get {
                return Factory.GetInstance();
            }
        }

        private static RubyEngineOptions/*!*/ GetSetupInformation() {
            // TODO: look-up appdomain data, config, etc. - maybe unified with the SE.GetSetupInfo
            return new RubyEngineOptions();
        }

        /// <summary>
        /// To be called by <see cref="GetLocalService"/> only.
        /// </summary>
        private RubyEngine(RubyLanguageProvider/*!*/ provider, RubyEngineOptions/*!*/ engineOptions)
            : base(provider, engineOptions) {

            _contextId = ContextId.RegisterContext(ContextRegistrationKey);
            _defaultContext = new CodeContext(new Scope(), new RubyContext(this, new RubyCompilerOptions()), new RubyModuleContext(null));
            _actionBinder = new RubyActionBinder(_defaultContext);

            // TODO: top-package should be per isolation context, engine may be per isolation context as well
            _executionContext = new RubyExecutionContext(this, DynamicHelpers.TopPackage); 
            _loader = new Loader(this);
        }

        #endregion

        internal void ClearExecutionContext() {
            _executionContext = new RubyExecutionContext(this, _executionContext.GlobalNamespaceDictionary);
        }

        public override CompilerOptions/*!*/ GetDefaultCompilerOptions() {
            return new RubyCompilerOptions();
        }

        protected override LanguageContext GetLanguageContext(ScriptModule/*!*/ module) {
            Debug.Assert(module != null);
            return new RubyContext(this, (RubyCompilerOptions)GetModuleCompilerOptions(module));
        }

        public override CompilerOptions GetModuleCompilerOptions(ScriptModule module) {
            RubyCompilerOptions result = new RubyCompilerOptions();

            // TODO: only if OpenVariablesModule() <==> module.Type == "System.Console" 

            Dictionary<SymbolId, VariableInfo> locals = new Dictionary<SymbolId, VariableInfo>();
            foreach (object name in module.Scope.GetAllKeys(DefaultContext.LanguageContext)) {
                string str = name as string;
                if (str != null) {
                    locals.Add(SymbolTable.StringToId(str), new VariableInfo(MSA.Variable.VariableKind.Global));
                }
            }

            result.LexicalScope = new LexicalScope(null, ScopeKind.RuntimeScope, locals);

            return result;
        }

        protected override LanguageContext/*!*/ GetLanguageContext(CompilerOptions/*!*/ compilerOptions) {
            return new RubyContext(this, (RubyCompilerOptions)compilerOptions);
        }

        protected override string[] FormatObjectMemberNames(IList<object> names) {
            if (names == null) throw new ArgumentNullException("value");

            //Dictionary<string,bool> result = new Dictionary<string, bool>(names.Count);
            List<string> result = new List<string>(names.Count);

            foreach (object name in names) {
                string strName = name as string;
                
                // TODO: remove constants
                if (strName != null && !strName.StartsWith("C:") && !strName.EndsWith("=") && strName.IndexOf('_') == -1) {
                    //if (strName.StartsWith("set_") || strName.StartsWith("get_") || strName.StartsWith("add_")) {
                    //    strName = strName.Substring("set_".Length);
                    //} else if (strName.StartsWith("remove_")) {
                    //    strName = strName.Substring("remove_".Length);
                    //}

                    if (_rubyCaseMemberLists) {
                        strName = RubyOps.MangleName(strName);
                    }

                    result.Add(strName);
                }
            }
            return result.ToArray();
        }

        protected override void PrintInteractiveCodeResult(object obj) {
            Kernel.PutString(DefaultContext, null, "=> " + RubySites.Inspect(obj).ToString());
        }

        #region // TODO: workarounds

        protected override IList<object> Ops_GetAttrNames(CodeContext context, object obj) {
            // obsolete:
            //// HACK: making operators invisible in intelli-sense (shoud just call GetAttrNames, not CLS flipping):
            //try {
            //    context.ModuleContext.ShowCls = false;
            //    ICustomMembers ids = obj as ICustomMembers;

            //    if (ids != null) {
            //        return ids.GetCustomMemberNames(context);
            //    }

            //    List<object> res = new List<object>();
            //    foreach (SymbolId x in DynamicHelpers.GetDynamicType(obj).GetMemberNames(context, obj))
            //        res.Add(SymbolTable.IdToString(x));

            //    //!!! ugly, we need to check for non-SymbolID keys
            //    ISuperDynamicObject dyno = obj as ISuperDynamicObject;
            //    if (dyno != null) {
            //        IAttributesCollection iac = dyno.Dict;
            //        if (iac != null) {
            //            foreach (object id in iac.Keys) {
            //                if (!res.Contains(id)) res.Add(id);
            //            }
            //        }
            //    }
            //    return res;
            //} finally {
            //    context.ModuleContext.ShowCls = true;
            //}
            throw new NotImplementedException();
        }

        protected override bool Ops_TryGetAttr(CodeContext context, object obj, SymbolId id, out object value) {
            // obsolete:
            //if (_rubyCaseMemberLists) {
            //    id = SymbolTable.StringToId(RubyOps.UnmangleName(SymbolTable.IdToString(id)));
            //    return RubyOps.TryGetAttr(context, obj, id, out value);
            //}
            //return RubyOps.TryGetAttr(context, obj, id, out value);
            throw new NotImplementedException();
        }

        protected override bool Ops_IsCallable(CodeContext context, object obj) {
            // obsolete:
            //if (obj is ICallableWithCodeContext || obj is ICallableWithThis)
            //    return true;

            //return false;
            throw new NotImplementedException();
        }

        protected override object Ops_Call(CodeContext context, object obj, object[] args) {
            // obsolete:
            //ICallableWithCodeContext icc = obj as ICallableWithCodeContext;
            //if (icc != null) return icc.Call(context, args);

            //object res;
            //if (!DynamicHelpers.GetDynamicType(obj).TryInvokeBinaryOperator(context, Operators.Call, obj, args, out res))
            //    throw new ArgumentTypeException(String.Format("{0} is not callable", DynamicHelpers.GetDynamicType(obj)));

            //return res;

            throw new NotImplementedException();

        }

        #endregion
    }
}
