/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System;
using System.IO;
using System.Text;
using System.Diagnostics;
using System.Collections;
using System.Collections.Generic;

using Assembly = System.Reflection.Assembly;
using Microsoft.Scripting;
using Microsoft.Scripting.Hosting;

using Ruby.Runtime;
using Ruby.Hosting;
using Ruby.Compiler;
using Ruby.Builtins;

namespace Ruby.Runtime {
    [Flags]
    internal enum LoadFlags {
        None = 0,
        LoadOnce = 1,
        LoadIsolated = 2,
    }

    internal sealed class Loader {

        private enum FileKind {
            RubySourceFile,
            NonRubySourceFile,
            Assembly
        }

        private RubyEngine/*!*/ _engine;

        public Loader(RubyEngine/*!*/ engine) {
            Utils.Assert.NotNull(engine);
            _engine = engine;
        }

        /// <summary>
        /// Returns <b>true</b> if a Ruby file is successfully loaded, <b>false</b> if it is already loaded.
        /// </summary>
        internal bool LoadFile(CodeContext/*!*/ context, object self, string/*!*/ path, LoadFlags flags) {
            Utils.Assert.NotNull(context, path);

            // relative => RubyOps.GetGlobalVariable(":");
            // ext .rb => source file
            // ext .so, .o, .dll => lib
            // other exts => add .rb, .so, .o, .dll
            // add name of the loaded in RubyOps.GetGlobalVariable($")
            // put source content to => SCRIPT_LINES__

            SourceFileUnit sfu;
            Assembly assembly;
            FileKind kind;
            bool found;
            try {
                found = ResolveFile(path, flags, out kind, out sfu, out assembly);
            } catch (Exception e) {
                throw new LoadError(e.Message, e);
            }

            if (!found) {
                throw new LoadError("File not found");
            }

            switch (kind) {
                case FileKind.RubySourceFile:
                    // Ruby code:
                    CompiledCode code = sfu.Compile();
                    ScriptCode scriptCode = ScriptCode.FromCompiledCode(code);


                    // TODO: we need an instance here (why a module object doesn't work as a module self?)
                    //if ((flags & LoadFlags.LoadIsolated) == LoadFlags.LoadIsolated) {
                    //    ((RubyContext)scriptCode.LanguageContext).SelfObject = RubyOps.MakeModule();
                    //}

                    scriptCode.Run(context.Scope, context.ModuleContext);

                    // TODO: caching
                    return true;

                // TODO:
                //case FileKind.NonRubySourceFile:
                //    // other langauge's code:
                //    ScriptModule module = sfu.CompileToModule();
                //    module.Execute();

                //    // TODO: include

                //    return module;

                case FileKind.Assembly:
                    try {
                        // TODO:
                        bool result = DynamicHelpers.TopPackage.LoadAssembly(assembly);
                        DynamicHelpers.TopPackage.GetCustomMemberDictionary(context);
                        return result;
                    } catch (Exception e) {
                        throw new LoadError(e.Message, e);
                    }
            }

            throw Utils.Assert.Unreachable;
        }

        private bool ResolveFile(string/*!*/ path, LoadFlags loadFlags, out FileKind kind, out SourceFileUnit sourceUnit, 
            out Assembly assembly) {

            sourceUnit = null;
            assembly = null;
            string extension = Path.GetExtension(path).ToLower();
            string resolvedPath;
            string listPath;

            List<object> loadedFileList = null;
            if ((loadFlags & LoadFlags.LoadOnce) == LoadFlags.LoadOnce) {
                loadedFileList = GetLoadedFilesList();
            }

            if (extension == ".rb") {
                kind = FileKind.RubySourceFile;
                resolvedPath = FindFile(listPath = path, true);
                if (loadedFileList != null && loadedFileList.Contains(listPath)) return false;
                sourceUnit = GetSourceUnit(resolvedPath);
            } else if (extension == ".dll") {
                kind = FileKind.Assembly;
                resolvedPath = FindFile(listPath = path, true);
                if (loadedFileList != null && loadedFileList.Contains(listPath)) return false;
                assembly = ScriptDomainManager.CurrentManager.PAL.LoadAssemblyFromPath(resolvedPath);
            } else {

                resolvedPath = FindFile(listPath = path + ".rb", false);
                if (resolvedPath != null) {
                    kind = FileKind.RubySourceFile;
                    if (loadedFileList != null && loadedFileList.Contains(listPath)) return false;
                    sourceUnit = GetSourceUnit(resolvedPath);
                } else {
                    resolvedPath = FindFile(listPath = path + ".dll", false);
                    if (resolvedPath != null) {
                        kind = FileKind.Assembly;
                        if (loadedFileList != null && loadedFileList.Contains(listPath)) return false;
                        assembly = ScriptDomainManager.CurrentManager.PAL.LoadAssemblyFromPath(resolvedPath);
                    } else {
#if SILVERLIGHT
                        LanguageProvider provider;
                        if (ScriptDomainManager.CurrentManager.TryGetLanguageProviderByFileExtension(extension, out provider)) {
                            sourceUnit = ScriptDomainManager.CurrentManager.Host.GetSourceFileUnit(provider.GetEngine(),
                                listPath = path, Path.GetFileNameWithoutExtension(path));
                        }
#else
                        sourceUnit = ScriptDomainManager.CurrentManager.Host.ResolveSourceFileUnit(listPath = path);
#endif
                        if (sourceUnit != null) {
                            if (sourceUnit.Engine is RubyEngine) {
                                kind = FileKind.RubySourceFile;
                            } else {
                                kind = FileKind.NonRubySourceFile;
                            }
                        } else {
                            assembly = ScriptDomainManager.CurrentManager.PAL.LoadAssembly(listPath = path);
                            kind = FileKind.Assembly;
                        }
                    }
                } 
            }

            if (sourceUnit != null || assembly != null) {
                GetLoadedFilesList().Add(listPath);
                return true;
            }

            return false;
        }

        private List<object>/*!*/ GetLoadedFilesList() {
            object value;
            if (!_engine.ExecutionContext.GlobalVariables.TryGetValue(Symbols.DoubleQuote, out value)) {
                Debug.Assert(false, "Global variable $\" should be always defined");
            }

            Debug.Assert(value is List<object>);
            return (List<object>)value;
        }

        private SourceFileUnit/*!*/ GetSourceUnit(string/*!*/ path) {
            return ScriptDomainManager.CurrentManager.Host.GetSourceFileUnit(RubyEngine.CurrentEngine, path, Path.GetFileName(path));
        }

        private string FindFile(string/*!*/ filePath, bool mustExist) {
#if !SILVERLIGHT 
            if (ScriptDomainManager.CurrentManager.Host.SourceFileExists(filePath)) {
                return filePath;
            }

            object pathsObj;
            if (RubyEngine.CurrentEngine.ExecutionContext.GlobalVariables.TryGetValue(Symbols.Colon, out pathsObj)) {
                // TODO: string conversion
                foreach (string path in ((string)pathsObj).Split(Path.PathSeparator)) {
                    string fullPath;

                    try {
                        fullPath = Path.Combine(path, filePath);

                        if (ScriptDomainManager.CurrentManager.Host.SourceFileExists(fullPath)) {
                            return fullPath;
                        }

                    } catch (ArgumentException) {
                        // invalid characters in path
                    }
                }
            }

            // TODO:
            if (mustExist) {
                throw new Exception("LoadError: " + filePath);
            }
#endif
            return null;
        }
    }
}
