/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;
using Ruby.Hosting;

using Microsoft.Scripting;
using Microsoft.Scripting.Actions;
using MSA = Microsoft.Scripting.Ast;
using MSG = Microsoft.Scripting.Generation;
using Ast = Microsoft.Scripting.Ast.Ast;
using System.Diagnostics;
using Ruby.Compiler.AST;
using Microsoft.Scripting.Generation;
using Ruby.Compiler;
using Ruby.Builtins;
using System.Reflection;
using System.Reflection.Emit;
using Microsoft.Scripting.Math;
using System.Collections;

namespace Ruby.Runtime {
    public sealed class RubyActionBinder : ActionBinder {
        internal RubyActionBinder(CodeContext/*!*/ defaultContext)
            : base(defaultContext) {
        }

        protected override StandardRule<T> MakeRule<T>(CodeContext callerContext, Action/*!*/ action, object[]/*!*/ args) {
            switch (action.Kind) {
                case ActionKind.InvokeMember:
                    return MakeRuleForInvokeMember<T>((InvokeMemberAction)action, callerContext, args);

                default:
                    return null;
            }
        }

        internal static StandardRule<T> GetUserClassRule<T>(Action/*!*/ action, CodeContext/*!*/ callerContext, object[]/*!*/ args) {
            Utils.Assert.NotNull(action, callerContext, args);

            switch (action.Kind) {
                case ActionKind.InvokeMember:
                    return MakeRuleForInvokeMember<T>((InvokeMemberAction)action, callerContext, args);

                default:
                    return null;
            }
        }

        internal static StandardRule<T> MakeRuleForInvokeMember<T>(InvokeMemberAction/*!*/ action, CodeContext/*!*/ callerContext, object[]/*!*/ args) {
            Utils.Assert.NotNull(callerContext, args, action);
            Debug.Assert(args.Length >= 1);
            Debug.Assert(action.IndexOfArgument(MSA.ArgumentKind.Instance) == 0, "InvokeMemberAction must have an instance argument");

            ActionBinder binder = callerContext.LanguageContext.Binder;
            StandardRule<T> result = new StandardRule<T>();
            RubyExecutionContext ec = RubyOps.GetExecutionContext(callerContext);
            object target = args[0];
            MSA.Expression targetParameter = result.Parameters[0];

            // tests type of the target:
            RubyModule module;
            if (target == null) {
                result.AddTest(Ast.Equal(targetParameter, Ast.Constant(target)));
                module = ec.NilClass;
            } else if (target is bool) {
                result.AddTest(Ast.AndAlso(
                    Ast.TypeIs(targetParameter, typeof(bool)),
                    Ast.Equal(Ast.Cast(targetParameter, typeof(bool)), Ast.Constant(target))
                ));

                module = (bool)target ? ec.TrueClass : ec.FalseClass;

            } else if ((module = target as RubyModule) != null) {
                result.AddTest(Ast.Equal(targetParameter, Ast.RuntimeConstant(module)));
            } else {
                AddNonNullTypeTest(result, target, targetParameter);
                module = ec.GetClassOf(target);
            }

            // tests version of the target:
            module.AddVersionTest(result);

            // TODO: CLR types shouldn't be represented by DynamicTypes:
            DynamicType dynamicType = target as DynamicType;
            if (dynamicType != null) {
                targetParameter = Ast.ReadProperty(Ast.Cast(result.Parameters[0], typeof(DynamicType)), typeof(DynamicType).GetProperty("UnderlyingSystemType"));
                result.AddTest(Ast.Equal(targetParameter, Ast.Constant(dynamicType.UnderlyingSystemType)));
                target = ec.GetOrCreateClass(dynamicType.UnderlyingSystemType);
            }
            
            SymbolId? methodNameArgument = null;

            RubyMethodBase method = ec.ResolveMethod(target, action.MemberName);
            if (method == null) {

                string name = SymbolTable.IdToString(action.MemberName);

                // TODO:
                if (dynamicType != null) {
                    if (name.EndsWith("=")) {
                        name = name.Substring(0, name.Length - 1);
                        string unmangled = RubyOps.UnmangleName(name);
                        if (GetClrStaticMemberRule(result, dynamicType, SymbolTable.StringToId("set_" + name), callerContext, action, target, args)) return result;
                        if (GetClrStaticMemberRule(result, dynamicType, SymbolTable.StringToId("set_" + unmangled), callerContext, action, target, args)) return result;
                    } else {

                        string unmangled = RubyOps.UnmangleName(name);
                        
                        if (args.Length == 1) {
                            if (GetClrStaticMemberRule(result, dynamicType, SymbolTable.StringToId("get_" + name), callerContext, action, target, args)) return result;
                            if (GetClrStaticMemberRule(result, dynamicType, SymbolTable.StringToId("get_" + unmangled), callerContext, action, target, args)) return result;
                        }

                        if (GetClrStaticMemberRule(result, dynamicType, action.MemberName, callerContext, action, target, args)) return result;
                        if (GetClrStaticMemberRule(result, dynamicType, SymbolTable.StringToId(unmangled), callerContext, action, target, args)) return result;
                    }
                }

                RubyClass targetClass = ec.GetClassOf(target);
                if (targetClass.ShowUnderlyingTypeMembers) {
                    dynamicType = DynamicHelpers.GetDynamicTypeFromType(targetClass.UnderlyingSystemType);

                    if (name.EndsWith("=")) {
                        name = name.Substring(0, name.Length - 1);
                        string unmangled = RubyOps.UnmangleName(name);
                        if (GetClrInstanceMemberRule(result, dynamicType, SymbolTable.StringToId("set_" + name), callerContext, action, target, args)) return result;
                        if (GetClrInstanceMemberRule(result, dynamicType, SymbolTable.StringToId("set_" + unmangled), callerContext, action, target, args)) return result;
                    } else {
                        string unmangled = RubyOps.UnmangleName(name);

                        if (args.Length == 1) {
                            if (GetClrInstanceMemberRule(result, dynamicType, SymbolTable.StringToId("get_" + name), callerContext, action, target, args)) return result;
                            if (GetClrInstanceMemberRule(result, dynamicType, SymbolTable.StringToId("get_" + unmangled), callerContext, action, target, args)) return result;
                        }

                        if (GetClrInstanceMemberRule(result, dynamicType, action.MemberName, callerContext, action, target, args)) return result;
                        if (GetClrInstanceMemberRule(result, dynamicType, SymbolTable.StringToId(unmangled), callerContext, action, target, args)) return result;
                    }
                }

                method = ec.ResolveMethod(target, Symbols.MethodMissing);
                if (method == null) {
                    SetMissingMethodError(binder, result, targetParameter, action.MemberName);
                    Debug.Assert(result.Target != null && result.Test != null);
                    return result;
                }

                methodNameArgument = action.MemberName;
            }

            switch (method.Kind) {
                case MethodKind.InstanceAllocator:
                case MethodKind.InstanceConstructor:
                    Debug.Assert(!methodNameArgument.HasValue);
                    Debug.Assert(target is RubyClass); // TODO: 
                    ((RubyClass)target).SetRuleForCreateInstance<T>(result, binder, action.ArgumentInfos, args, 
                        method.Kind == MethodKind.InstanceConstructor, false);
                    break;

                case MethodKind.Regular:
                    method.SetInvocationRule(binder, result, action.ArgumentInfos, args, methodNameArgument);
                    break;

                case MethodKind.Block:
                    throw Utils.Assert.Unreachable;
            }

            Debug.Assert(result.Target != null && result.Test != null);
            return result;
        }

        private static bool GetClrInstanceMemberRule<T>(StandardRule<T>/*!*/ result, DynamicType/*!*/ dynamicType, SymbolId name, 
            CodeContext/*!*/ callerContext, InvokeMemberAction/*!*/ action, object target, object[] args) {
            object member;
            if (dynamicType.TryGetBoundMember(callerContext, target, name, out member)) {
                ReflectedEvent.BoundEvent evnt = member as ReflectedEvent.BoundEvent;
                if (evnt != null) {
                    HookEvent(callerContext.LanguageContext.Binder, evnt.Event.Info, result, action, args);
                    return true;
                }

                BoundBuiltinFunction f = member as BoundBuiltinFunction;
                if (f != null) {
                    RubyMethodGroupInfo.SetInvocationRuleInternal<T>(f.Name, f.Target.Targets, callerContext.LanguageContext.Binder, result,
                        action.ArgumentInfos,
                        args,
                        result.Parameters, null);

                    Debug.Assert(result.Target != null && result.Test != null);
                    return true;
                }
            }
            return false;
        }

        private static bool GetClrStaticMemberRule<T>(StandardRule<T>/*!*/ result, DynamicType/*!*/ dynamicType, SymbolId/*!*/ name,
            CodeContext/*!*/ callerContext, InvokeMemberAction/*!*/ action, object target, object[] args) {
            object member;
            if (dynamicType.TryGetBoundMember(callerContext, target, name, out member)) {
                BuiltinFunction f = member as BuiltinFunction;
                if (f != null) {
                    RubyMethodGroupInfo.SetInvocationRuleInternal<T>(f.Name, f.Targets, callerContext.LanguageContext.Binder, result,
                        Utils.Array.RemoveFirst(action.ArgumentInfos),
                        Utils.Array.RemoveFirst(args),
                        Utils.Array.RemoveFirst(result.Parameters), null);

                    Debug.Assert(result.Target != null && result.Test != null);
                    return true;
                }
            }
            return false;
        }

        internal static void AddNonNullTypeTest<T>(StandardRule<T>/*!*/ result, object/*!*/ obj, MSA.Expression/*!*/ parameter) {
            AddTypeTest(result, obj.GetType(), parameter);
        }

        internal static void SetMissingMethodError<T>(ActionBinder/*!*/ binder, StandardRule<T>/*!*/ rule, MSA.Expression/*!*/ target, SymbolId missingMethodName) {
            rule.SetTarget(rule.MakeError(binder, Ast.Call(null, RubyOps.GetMethod("MakeMissingMethodException"), target, Ast.Constant(missingMethodName))));
        }

        internal static void AddTypeTest<T>(StandardRule<T>/*!*/ rule, Type/*!*/ type, MSA.Expression/*!*/ expr) {
            if (expr.ExpressionType == type && type.IsSealed) {
                // TODO: allow null tests
                rule.AddTest(Ast.True());
                return;
            }

            if (!expr.ExpressionType.IsValueType) {
                rule.AddTest(Ast.NotEqual(expr, Ast.Null()));
            }

            rule.AddTest(Ast.Equal(Ast.Call(expr, typeof(object).GetMethod("GetType")), Ast.Constant(type)));
        }

        #region Helpers

        internal static MSA.Expression/*!*/ MakeCallExpression(Delegate/*!*/ method, MSA.Expression[]/*!*/ actualArgs) {

            bool inlineDelegateCall = method.Method.GetType() == typeof(MethodInfo);
            
            ParameterInfo[] ps = method.Method.GetParameters();
            if (ps.Length > 0) {
                List<MSA.Expression> realArgs = new List<MSA.Expression>(actualArgs.Length);

                int paramIndex = 0;
                if (method.Target != null) {
                    Debug.Assert(ps[paramIndex].ParameterType.IsAssignableFrom(method.Target.GetType()));

                    // delegate closed over constant pool:
                    if (inlineDelegateCall) {
                        realArgs.Add(Ast.RuntimeConstant(method.Target));
                    }
                    paramIndex++;
                }

                if (ps[paramIndex].ParameterType == typeof(CodeContext)) {
                    realArgs.Add(Ast.CodeContext());
                    paramIndex++;
                }

                realArgs.AddRange(actualArgs);
                actualArgs = realArgs.ToArray();
            }

            if (inlineDelegateCall) {
                return Ast.Call(null, method.Method, actualArgs);
            } else {
                return Ast.Call(Ast.RuntimeConstant(method), method.GetType().GetMethod("Invoke"), actualArgs);
            }
        }

        private static void HookEvent<T>(ActionBinder/*!*/ binder, EventInfo/*!*/ eventInfo, StandardRule<T>/*!*/ result, InvokeMemberAction/*!*/ action, 
            object[]/*!*/ args) {

            int blockIndex = action.IndexOfArgument(MSA.ArgumentKind.Block);
            if (blockIndex != -1) {
                Proc proc = (Proc)args[blockIndex];

                MethodInfo addMethod = eventInfo.GetAddMethod();
                if (addMethod == null) {
                    throw new NotImplementedError("non-public event");
                }

                MSA.CodeBlock handler = Ast.EventHandlerBlock("#handler", eventInfo);
                handler.ExplicitCodeContextExpression = Ast.ReadProperty(Ast.RuntimeConstant(proc), Proc.CodeContextProperty);

                MSA.Expression[] procArgs = new MSA.Expression[handler.Parameters.Count + 2];
                procArgs[0] = Ast.RuntimeConstant(proc);
                procArgs[1] = Ast.New(BlockReturnReasonHolder.Constructor);
                ReadAll(handler.Parameters, procArgs, 2);

                if (handler.ReturnType == typeof(void)) {
                    handler.Body = Ast.Statement(Ast.Action.Call(CallAction.Simple, typeof(object), procArgs));
                } else {
                    handler.Body = Ast.Return(Ast.Action.Call(CallAction.Simple, handler.ReturnType, procArgs));
                }

                handler.BindClosures();

                // return <instance>.<event> += new Delegate(handler)
                result.SetTarget(result.MakeReturn(binder, 
                    Ast.Call(result.Parameters[0], addMethod, Ast.CodeBlockReference(SourceSpan.None, handler, eventInfo.EventHandlerType))));
            } else {
                // TODO: make error
                throw new NotImplementedError("no block given");
            }
        }

        private static void ReadAll(IList<MSA.Variable> variables, IList<MSA.Expression> expressions, int start) {
            for (int i = 0, j = start; i < variables.Count; i++, j++) {
                expressions[j] = Ast.Read(variables[i]);
            }
        }

        #endregion

        #region Conversions

        // TODO:
        public override void EmitConvertFromObject(MSG.CodeGen/*!*/ cg, Type/*!*/ toType) {
            Utils.Assert.NotNull(cg, toType);
            if (toType == typeof(object)) return;

            MethodInfo fastConvertMethod = GetFastConvertMethod(toType);
            if (fastConvertMethod != null) {
                cg.EmitCall(fastConvertMethod);
                return;
            }

            if (toType == typeof(void)) {
                cg.Emit(OpCodes.Pop);
            } else if (typeof(Delegate).IsAssignableFrom(toType)) {
                cg.EmitType(toType);
                cg.EmitCall(typeof(Converter), "ConvertToDelegate");
                cg.Emit(OpCodes.Castclass, toType);
            } else {
                Label end = cg.DefineLabel();
                cg.Emit(OpCodes.Dup);
                cg.Emit(OpCodes.Isinst, toType);

                cg.Emit(OpCodes.Brtrue_S, end);
                cg.Emit(OpCodes.Ldtoken, toType);
                cg.EmitCall(GetGenericConvertMethod(toType));
                cg.MarkLabel(end);

                cg.Emit(OpCodes.Unbox_Any, toType); //??? this check may be redundant
            }
        }

        public override object Convert(object obj, Type/*!*/ toType) {
            Utils.Assert.NotNull(toType);
            return Converter.Convert(obj, toType);
        }

        public override MSA.Expression ConvertExpression(MSA.Expression expr, Type toType) {
            Type exprType = expr.ExpressionType;

            if (toType == typeof(object)) {
                if (exprType.IsValueType) {
                    return Ast.Cast(expr, toType);
                } else {
                    return expr;
                }
            }

            if (toType.IsAssignableFrom(exprType)) {
                return expr;
            }

            MSA.BoundExpression be = expr as MSA.BoundExpression;
            if (be != null && be.Variable.KnownType != null) {
                if (toType.IsAssignableFrom(be.Variable.KnownType)) {
                    return Ast.Cast(expr, toType);
                }
            }

            // We used to have a special case for int -> double...
            if (exprType != typeof(object)) {
                expr = Ast.Cast(expr, typeof(object));
            }

            if (toType == typeof(object)) return expr;

            MethodInfo fastConvertMethod = GetFastConvertMethod(toType);
            if (fastConvertMethod != null) {
                return Ast.Call(null, fastConvertMethod, expr);
            }

            if (typeof(Delegate).IsAssignableFrom(toType)) {
                return Ast.Cast(
                    Ast.Call(
                        null,
                        typeof(Converter).GetMethod("ConvertToDelegate"),
                        expr,
                        Ast.Constant(toType)
                    ),
                    toType
                );
            }

            return Ast.Condition(
                Ast.TypeIs(
                    expr,
                    toType),
                Ast.Cast(
                    expr,
                    toType),
                Ast.Cast(
                    Ast.Call(
                        null, GetGenericConvertMethod(toType),
                        expr, Ast.Constant(toType.TypeHandle)
                    ),
                    toType
                )
            );
        }

        private static MethodInfo GetGenericConvertMethod(Type toType) {
            if (toType.IsValueType) {
                if (toType.IsGenericType && toType.GetGenericTypeDefinition() == typeof(Nullable<>)) {
                    return typeof(Converter).GetMethod("ConvertToNullableType");
                } else {
                    return typeof(Converter).GetMethod("ConvertToValueType");
                }
            } else {
                return typeof(Converter).GetMethod("ConvertToReferenceType");
            }
        }

        private static MethodInfo GetFastConvertMethod(Type toType) {
            if (toType == typeof(char)) {
                return typeof(Converter).GetMethod("ConvertToChar");
            } else if (toType == typeof(int)) {
                return typeof(Converter).GetMethod("ConvertToInt32");
            } else if (toType == typeof(string)) {
                return typeof(Converter).GetMethod("ConvertToString");
            } else if (toType == typeof(long)) {
                return typeof(Converter).GetMethod("ConvertToInt64");
            } else if (toType == typeof(double)) {
                return typeof(Converter).GetMethod("ConvertToDouble");
            } else if (toType == typeof(bool)) {
                return typeof(Converter).GetMethod("ConvertToBoolean");
            } else if (toType == typeof(BigInteger)) {
                return typeof(Converter).GetMethod("ConvertToBigInteger");
            } else if (toType == typeof(Complex64)) {
                return typeof(Converter).GetMethod("ConvertToComplex64");
            } else if (toType == typeof(IEnumerable)) {
                return typeof(Converter).GetMethod("ConvertToIEnumerable");
            } else if (toType == typeof(float)) {
                return typeof(Converter).GetMethod("ConvertToSingle");
            } else if (toType == typeof(byte)) {
                return typeof(Converter).GetMethod("ConvertToByte");
            } else if (toType == typeof(sbyte)) {
                return typeof(Converter).GetMethod("ConvertToSByte");
            } else if (toType == typeof(short)) {
                return typeof(Converter).GetMethod("ConvertToInt16");
            } else if (toType == typeof(uint)) {
                return typeof(Converter).GetMethod("ConvertToUInt32");
            } else if (toType == typeof(ulong)) {
                return typeof(Converter).GetMethod("ConvertToUInt64");
            } else if (toType == typeof(ushort)) {
                return typeof(Converter).GetMethod("ConvertToUInt16");
            } else if (toType == typeof(Type)) {
                return typeof(Converter).GetMethod("ConvertToType");
            } else {
                return null;
            }
        }
        
        public override bool CanConvertFrom(Type fromType, Type toType, NarrowingLevel level) {
            return Converter.CanConvertFrom(fromType, toType, level);
//            return fromType == toType || fromType.IsSubclassOf(toType);
        }

        public override bool PreferConvert(Type t1, Type t2) {
            return Converter.PreferConvert(t1, t2);
//            return t1 == t2;
        }

        #endregion
    }
}
