/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;

using Microsoft.Scripting;
using Microsoft.Scripting.Hosting;

using Ruby.Compiler;
using System.Diagnostics;
using Ruby.Builtins;
using Microsoft.Scripting.Actions;
using Ruby.Hosting;
using System.IO;

namespace Ruby.Runtime {
    public sealed class RubyContext : LanguageContext {
        private readonly ContextId _id;
        private object/*!*/ _selfObject;
        private LexicalScope/*!*/ _lexicalScope;
        private RubyEngine _engine;
        
        public LexicalScope LexicalScope {
            get { return _lexicalScope; }
            internal set { Debug.Assert(value != null); _lexicalScope = value; } 
        }

        public object/*!*/ SelfObject { 
            get { return _selfObject; }
            internal set { Debug.Assert(value != null); _selfObject = value; } 
        }

        private RubyMethodVisibility _currentVisibility;

        public RubyMethodVisibility CurrentVisibility {
            get { return _currentVisibility; }
            set { _currentVisibility = value; }
        }

        public override ContextId ContextId {
            get {
                return _id;
            }
        }

        public override ActionBinder Binder {
            get {
                return base.Binder;
            }
        }

        public override ScriptEngine Engine {
            get { return _engine; }
        }

        public RubyContext(RubyEngine/*!*/ engine, RubyCompilerOptions/*!*/ options) {
            Utils.Assert.NotNull(options);

            _engine = engine;
            _id = engine.ContextId;
            // TODO: should be new object()
            _selfObject = new object();
            _lexicalScope = new LexicalScope(null, ScopeKind.RuntimeScope);

            // default visibility - a LC is created for class/module declaration scope:
            _currentVisibility = RubyMethodVisibility.Public;
        }

        public override CompilerOptions/*!*/ GetCompilerOptions() {
            RubyCompilerOptions result = new RubyCompilerOptions();
            result.LexicalScope = _lexicalScope;
            return result;
        }

        public override ModuleContext CreateModuleContext(ScriptModule/*!*/ module) {
            Utils.Assert.NotNull(module);

            return new RubyModuleContext(module, CreateModuleSelf(module));
        }

        public override void ModuleContextEntering(ModuleContext/*!*/ newContext) {
            Utils.Assert.NotNull(newContext);
            RubyModuleContext newRubyContext = (RubyModuleContext)newContext;

            // sets self to the encapsulating class if there is any:
            if (newRubyContext.ModuleSelf != null) {
                _selfObject = newRubyContext.ModuleSelf;
            }
        }

        private object CreateModuleSelf(ScriptModule/*!*/ module) {
            // TODO: module.ModuleKind == Guid(...)
            if (module.Kind == ScriptModuleKind.XamlPage) {
                throw new NotImplementedException();
                // return RubyClass.Allocate(XamlPage.InitializeScriptModule(module));
            } else {
                return null;
            }
        }

        // TODO: (DLR) virtual dispatch on each AND/OR expression, could we just feed code gen with static method infos?
        public override bool IsTrue(object obj) {
            return RubyOps.IsTrue(obj);
        }

    }
}
