/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;
using Ruby.Hosting;
using Microsoft.Scripting;
using Ruby.Builtins;
using Ruby.Compiler;
using System.Diagnostics;
using Microsoft.Scripting.Math;
using Microsoft.Scripting.Generation;
using System.Threading;
using System.Reflection;
using System.Reflection.Emit;

namespace Ruby.Runtime {

    /// <summary>
    /// An isolation context: could be per thread/request or shared accross certain threads.
    /// </summary>
    public sealed class RubyExecutionContext {

        /// <summary>
        /// Always thread static.
        /// </summary>
        [ThreadStatic]
        private static Exception _currentException;

        private readonly RubyEngine/*!*/ _engine;
        private readonly RuntimeErrorSink/*!*/ _runtimeErrorSink;
        private readonly Dictionary<SymbolId, object>/*!*/ _globalVariables;
        private readonly Dictionary<Type, RubyClass>/*!*/ _classCache;
        private readonly Dictionary<Type, RubyModule>/*!*/ _moduleCache;
        private ICustomMembers/*!*/ _globalNamespaceDictionary;
        private static int _classId = 0;

        // classes used by runtime (we need to update initialization generator if any of these are added):
        private RubyModule/*!*/ _kernelModule;
        private RubyClass/*!*/ _objectClass;
        private RubyClass/*!*/ _classClass;
        private RubyClass/*!*/ _nilClass;
        private RubyClass/*!*/ _trueClass;
        private RubyClass/*!*/ _falseClass;
        private RubyClass/*!*/ _exceptionClass;
        private RubyClass/*!*/ _runtimeErrorClass;
        private RubyClass/*!*/ _standardErrorClass;

        // internally set by Initializer:
        internal RubyModule/*!*/ KernelModule { get { return _kernelModule; } }
        internal RubyClass/*!*/ ObjectClass { get { return _objectClass; } }
        internal RubyClass/*!*/ ClassClass { get { return _classClass; } set { _classClass = value; } }
        internal RubyClass/*!*/ NilClass { get { return _nilClass; } set { _nilClass = value; } }
        internal RubyClass/*!*/ TrueClass { get { return _trueClass; } set { _trueClass = value; } }
        internal RubyClass/*!*/ FalseClass { get { return _falseClass; } set { _falseClass = value; } }
        internal RubyClass/*!*/ ExceptionClass { get { return _exceptionClass; } set { _exceptionClass = value; } }
        internal RubyClass/*!*/ RuntimeErrorClass { get { return _runtimeErrorClass; } set { _runtimeErrorClass = value; } }
        internal RubyClass/*!*/ StandardErrorClass { get { return _standardErrorClass; } set { _standardErrorClass = value; } }

        // TODO: can be optimized if the EC is not shared, but in EH, we may not care
        public Exception CurrentException {
            get { return _currentException; } 
            set { _currentException = value; } 
        }

        public RuntimeErrorSink/*!*/ RuntimeErrorSink {
            get { return _runtimeErrorSink; }
        }

        public Dictionary<SymbolId, object>/*!*/ GlobalVariables {
            get { return _globalVariables; }
        }

        public ICustomMembers/*!*/ GlobalNamespaceDictionary {
            get { return _globalNamespaceDictionary; }
        }

        internal RubyEngine/*!*/ Engine {
            get { return _engine; }
        }

        public bool ShowCls {
            get { return false; }
        }	

        public RubyExecutionContext(RubyEngine/*!*/ engine, ICustomMembers/*!*/ globalNamespaceDictionary) {
            Utils.Assert.NotNull(engine, globalNamespaceDictionary);
            _engine = engine;

            _runtimeErrorSink = new RuntimeErrorSink();
            _globalVariables = new Dictionary<SymbolId, object>();
            _classCache = new Dictionary<Type, RubyClass>();
            _moduleCache = new Dictionary<Type, RubyModule>();
            _globalNamespaceDictionary = globalNamespaceDictionary;

            InitializeCaches();
            InitializeGlobalConstants();
            InitializeGlobalVariables();
        }

        /// <summary>
        /// Clears thread static variables.
        /// </summary>
        internal static void ClearThreadStatics() {
            _currentException = null;
        }

        internal void InitializeGlobalVariables() {
            // TODO: globals init: "." here?
            _globalVariables.Add(Symbols.DoubleQuote, new List<object>());
            _globalVariables.Add(Symbols.Comma, String.Empty);
            _globalVariables.Add(Symbols.ForwardSlash, "\n");
        }

        private void InitializeGlobalConstants() {
            Debug.Assert(_objectClass != null);

            string version = RubyVersionInfo.Version;
            string platform = "i386-mswin32";   // TODO: make this the correct string for MAC OS X in Silverlight
            string releaseDate = RubyVersionInfo.ReleaseDate;

            _objectClass.SetConstant("RUBY_VERSION", version);
            _objectClass.SetConstant("RUBY_PLATFORM", platform);
            _objectClass.SetConstant("RUBY_RELEASE_DATE", releaseDate);

            _objectClass.SetConstant("VERSION", version);
            _objectClass.SetConstant("PLATFORM", platform);
            _objectClass.SetConstant("RELEASE_DATE", releaseDate);
        }

        internal void InitializeCaches() {
            
            // TODO: library loader - for 3rd party libs, dynamic lib loading, [RubyLibrary(initializer)], etc...
            new BuiltinsInitializer().LoadModules(this);

            Debug.Assert(_classClass != null && _nilClass != null && _exceptionClass != null && 
                _runtimeErrorClass != null && _standardErrorClass != null);
        }

        internal void RegisterObjectAndKernel(Action<RubyModule>/*!*/ objectInitializer, Action<RubyModule>/*!*/ kernelInitializer) {
            Utils.Assert.NotNull(objectInitializer, kernelInitializer);
            Debug.Assert(_kernelModule == null && _objectClass == null);
            
            // we need to create object class before any other since constants are added in its dictionary:
            _kernelModule = new RubyModule(this, Symbols.Kernel, kernelInitializer);
            _objectClass = new RubyClass(this, Symbols.Object, typeof(object), objectInitializer, null, true, ShowCls);
            _objectClass.SetMixins(new RubyModule[] { _kernelModule });

            _classCache.Add(_objectClass.UnderlyingSystemType, _objectClass);

            _objectClass.SetConstant(_objectClass.Name, _objectClass);
            _objectClass.SetConstant(_kernelModule.Name, _kernelModule);
        }

        internal void AddClassToCache(RubyClass/*!*/ classObject) {
            Utils.Assert.NotNull(classObject);
            _classCache.Add(classObject.UnderlyingSystemType, classObject);
        }

        internal void AddModuleToCache(RubyModule/*!*/ module, Type/*!*/ trait) {
            Utils.Assert.NotNull(module, trait);
            _moduleCache.Add(trait, module);
        }

        internal RubyClass/*!*/ GetClass(Type/*!*/ type) {
            Utils.Assert.NotNull(type);
            return _classCache[type];
        }

        internal RubyModule/*!*/ GetModule(Type/*!*/ type) {
            Utils.Assert.NotNull(type);
            return _moduleCache[type];
        }

        internal RubyClass/*!*/ GetOrCreateClass(Type/*!*/ type) {
            Utils.Assert.NotNull(type);

            // TODO: optimize (if this context isn't shared, we don't need to use locks)
            lock (_classCache) {
                return GetOrCreateClassNoLock(type);
            }
        }

        private RubyClass/*!*/ GetOrCreateClassNoLock(Type/*!*/ type) {
            Utils.Assert.NotNull(type);

            RubyClass result;
            if (_classCache.TryGetValue(type, out result)) {
                return result;
            }

            return _classCache[type] = 
                new RubyClass(this, SymbolTable.StringToId(type.Name), type, null, GetOrCreateClassNoLock(type.BaseType), true, true);
        }

        public bool IsInstanceOf(object value, object classObject) {
            RubyClass c = classObject as RubyClass;
            if (c != null) {
                return GetClassOf(value).IsSubclassOf(c);
            }

            return false;
        }

        public RubyClass/*!*/ GetClassOf(object obj) {
            if (obj == null) return _nilClass;
            if (obj is bool) {
                return (bool)obj ? _trueClass : _falseClass;
            }
            return GetOrCreateClass(obj.GetType());
        }

        public RubyClass AsClass(object obj) {
            RubyClass c = obj as RubyClass;
            if (c != null) {
                return c;
            }

            Type type = obj as Type;
            if (type != null) {
                return GetOrCreateClass(type);
            }

            return null;
        }

        internal RubyModule/*!*/ GetConstantOwner(object target) {

            RubyModule module = target as RubyModule;
            if (module == null) {
                return (target != null) ? GetClassOf(target) : _objectClass;
            }

            return module;
        }

        internal RubyClass/*!*/ MakeClass(SymbolId name, RubyClass/*!*/ superClass) {
            Utils.Assert.NotNull(superClass);

            AssemblyGen ag = ScriptDomainManager.CurrentManager.Snippets.Assembly;
            TypeGen typeGen = ag.DefinePublicType("Ruby.Classes." + SymbolTable.IdToString(name) + Interlocked.Increment(ref _classId), superClass.UnderlyingSystemType);

            // parameterless constructor:
            CodeGen ctorGen = typeGen.DefineConstructor(Type.EmptyTypes);
            ctorGen.Emit(OpCodes.Ldarg_0);
            ctorGen.Emit(OpCodes.Call, superClass.UnderlyingSystemType.GetConstructor(Type.EmptyTypes));
            ctorGen.Emit(OpCodes.Ret);

            EmitDynamicObjectImplementation(typeGen);

            Type clrType = typeGen.FinishType();

            return _classCache[clrType] = new RubyClass(this, name, clrType, null, superClass, false, false);
        }

        private void EmitDynamicObjectImplementation(TypeGen/*!*/ typeGen) {
            typeGen.TypeBuilder.AddInterfaceImplementation(typeof(IDynamicObject));

            CodeGen getRuleMethod = typeGen.DefineMethodOverride(typeof(IDynamicObject).GetMethod("GetRule"));
            MethodInfo mi = typeof(RubyOps).GetMethod("GetClassRule");
            GenericTypeParameterBuilder[] types = ((MethodBuilder)getRuleMethod.MethodInfo).DefineGenericParameters("T");

            for (int i = 0; i < 3; i++) getRuleMethod.EmitArgGet(i);

            getRuleMethod.EmitCall(mi.MakeGenericMethod(types));
            getRuleMethod.EmitReturn();
            getRuleMethod.Finish();

            CodeGen getContextMethod = typeGen.DefineMethodOverride(typeof(IDynamicObject).GetMethod("get_LanguageContext"));
            getContextMethod.EmitCall(typeof(RubyOps), "GetLanguageContext");
            getContextMethod.EmitReturn();
            getContextMethod.Finish();
        }

        internal RubyModule/*!*/ MakeModule(SymbolId name) {
            return new RubyModule(this, name, null);
        }

        internal RubyModule/*!*/ DefineModule(RubyModule/*!*/ declaringModule, SymbolId name, RubyClass superClass) {
            Utils.Assert.NotNull(declaringModule);

            bool isClass = superClass != null;
            object moduleObject;
            if (declaringModule.TryResolveConstant(name, out moduleObject)) {
                RubyModule module = moduleObject as RubyModule;
                if (module != null && module.IsClass == isClass) {
                    return module;
                }
            } else {
                // create class/module object:
                RubyModule module = isClass ? MakeClass(name, superClass) : MakeModule(name);
                declaringModule.SetConstant(name, module);

                return module;
            }

            throw RubyExceptions.CreateTypeError(String.Format(isClass ? "{0} is not a class" : "{0} is not a module", name));
        }

        internal RubyMethodBase ResolveMethod(object target, SymbolId name) {
            RubyModule module = target as RubyModule;

            if (module != null) {
                return module.ResolveMethod(name);
            } else {
                return GetClassOf(target).ResolveMethod(name, true);
            }
        }
    }
}
