/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Scripting;
using System.Reflection;
using Microsoft.Scripting.Actions;
using Microsoft.Scripting.Generation;
using Ast = Microsoft.Scripting.Ast.Ast;
using MSA = Microsoft.Scripting.Ast;
using Ruby.Hosting;
using System.Diagnostics;

namespace Ruby.Runtime {
    public sealed class RubyMethodGroupInfo : RubyMethodBase {
        private readonly Delegate[]/*!*/ _overloads;

        internal RubyMethodGroupInfo(SymbolId name, Delegate[]/*!*/ overloads, RubyMethodVisibility visibility, MethodKind kind)
            : base(name, visibility, kind) {
            Utils.Assert.NotNullItems(overloads);
            _overloads = overloads;
        }

        #region Dynamic Sites

        public override void SetInvocationRule<T>(ActionBinder/*!*/ binder, StandardRule<T>/*!*/ result, ArgumentInfo[] argInfos,
            object[]/*!*/ args, SymbolId? methodNameArgument) {

            Utils.Assert.NotNull(binder, result, args);
            Debug.Assert(Kind == MethodKind.Regular);

            SetInvocationRuleInternal<T>(NameString, Utils.Reflection.GetMethodInfos(_overloads), binder, result, argInfos, args, 
                result.Parameters, methodNameArgument);
        }

        internal static void SetInvocationRuleInternal<T>(string/*!*/ name, MethodBase[]/*!*/ overloads, ActionBinder/*!*/ binder, 
            StandardRule<T>/*!*/ rule, ArgumentInfo[] argInfos, object[]/*!*/ args, MSA.Expression[]/*!*/ parameters, SymbolId? methodNameArgument) {
            
            // TODO: parameters - improve DLR
            // TODO: optional Procs

            // TODO: 
            Utils.Assert.NotNull(binder, rule, args);

            MethodBinder methodBinder = MethodBinder.MakeBinder(binder, name, overloads, BinderType.Normal);

            Type[] argTypes = GetArgTypes(argInfos, args, methodNameArgument);

            MethodCandidate candidate = methodBinder.MakeBindingTarget(CallType.None, argTypes);

            if (candidate != null) {
                MSA.Expression[] actualArgs = MakeRule(rule, argInfos, args, parameters, methodNameArgument);
                rule.SetTarget(rule.MakeReturn(binder, candidate.Target.MakeExpression(binder, actualArgs)));
            } else {
                rule.SetTarget(rule.MakeError(binder, Ast.Call(null, RubyOps.GetMethod("MakeInvalidArgumentTypesError"), Ast.Constant(name))));
            }
        }

        private static MSA.Expression[]/*!*/ MakeRule<T>(StandardRule<T>/*!*/ rule, ArgumentInfo[] argInfos, object[]/*!*/ args, MSA.Expression[]/*!*/ parameters, SymbolId? methodNameArgument) {
            // TODO: the test should be generated by MakeBindingTarget

            MSA.Expression[] actualArgs = new MSA.Expression[parameters.Length + (methodNameArgument.HasValue ? 1 : 0)];

            int i = 0;
            int j = 0;
            if (i < argInfos.Length && GetArgumentKind(argInfos, i) == MSA.ArgumentKind.Instance) {
                // test already added by method resolver
                if (args[i] == null) {
                    actualArgs[j] = Ast.Null();
                } else {
                    actualArgs[j] = Ast.Cast(parameters[i], args[i].GetType());
                }
                i++;
                j++;
            }

            if (methodNameArgument.HasValue) {
                actualArgs[j] = Ast.Constant(methodNameArgument.Value);
                j++;
            }

            while (i < args.Length) {
                switch (GetArgumentKind(argInfos, i)) {
                    case MSA.ArgumentKind.Instance:
                        throw Utils.Assert.Unreachable;

                    case MSA.ArgumentKind.Simple:
                    case MSA.ArgumentKind.Block:
                        if (args[i] == null) {
                            rule.AddTest(Ast.Equal(parameters[i], Ast.Null()));
                            actualArgs[j] = Ast.Null();
                        } else {
                            RubyActionBinder.AddTypeTest(rule, args[i].GetType(), parameters[i]);
                            actualArgs[j] =  Ast.Cast(parameters[i], args[i].GetType());
                        }
                        break;

                    case MSA.ArgumentKind.List:
                        // expr ::= expr && (list = RubyOps.AsArrayOfLength(parameter[i], length) != null)
                        AddTestForListArg(rule, args[i], parameters[i]);
                        actualArgs[j] = Ast.Cast(parameters[i], typeof(List<object>));
                        break;

                    default:
                        throw new NotSupportedException();
                }

                i++;
                j++;
            }

            Debug.Assert(i == args.Length && j == actualArgs.Length);
            return actualArgs;
        }

        private static Type[]/*!*/ GetArgTypes(ArgumentInfo[] argInfos, object[]/*!*/ args, SymbolId? methodNameArgument) {

            Type[] result = new Type[args.Length + (methodNameArgument.HasValue ? 1 : 0)];

            int i = 0;
            int j = 0;
            if (i < argInfos.Length && GetArgumentKind(argInfos, i) == MSA.ArgumentKind.Instance) {
                result[j] = (args[i] != null) ? args[i].GetType() : typeof(object);
                i++;
                j++;
            }

            if (methodNameArgument.HasValue) {
                result[j] = typeof(SymbolId);
                j++;
            }

            while (i < args.Length) {
                result[j] = (args[i] != null) ? args[i].GetType() : typeof(object);
                i++;
                j++;
            }

            Debug.Assert(i == args.Length && j == result.Length);
            return result;
        }

        private static MSA.ArgumentKind GetArgumentKind(ArgumentInfo[] argInfos, int index) {
            return argInfos != null ? argInfos[index].Kind : MSA.ArgumentKind.Simple;
        }

        #endregion
    }
}

