/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Scripting;
using System.Diagnostics;
using System.Reflection;
using Microsoft.Scripting.Actions;
using Ruby.Hosting;
using MSA = Microsoft.Scripting.Ast;
using Ast = Microsoft.Scripting.Ast.Ast;

namespace Ruby.Runtime {
    public sealed class RubyMethodInfo : RubyMethodBase {
        private readonly Delegate/*!*/ _method;
        private readonly int _mandatoryParamCount;
        private readonly int _optionalParamCount;
        private readonly bool _hasParamsArray;

        public Delegate/*!*/ Method { get { return _method; } }
        public int MandatoryParamCount { get { return _mandatoryParamCount; } }
        public int OptionalParamCount { get { return _optionalParamCount; } }
        public bool HasParamsArray { get { return _hasParamsArray; } }

        public int ClrParameterCount { 
            get {
                return (IsBlock) ?
                    _mandatoryParamCount + // always (this block, BlockReturnReason)
                    _optionalParamCount +
                    (_hasParamsArray ? 1 : 0)
                :
                    1 + // self
                    1 + // block
                    _mandatoryParamCount +
                    _optionalParamCount +
                    (_hasParamsArray ? 1 : 0);
                }
        }

        // block:
        internal RubyMethodInfo(Delegate/*!*/ method, int optional, bool hasParamsArray) 
            : base(SymbolId.Empty, RubyMethodVisibility.Public, MethodKind.Block) {
            Debug.Assert(method != null);

            _method = method;
            _optionalParamCount = optional;
            _mandatoryParamCount = 2;
            _hasParamsArray = hasParamsArray;
        }
        
        // method:
        internal RubyMethodInfo(Delegate/*!*/ method, SymbolId name, int mandatory, int optional, bool hasParamsArray, RubyMethodVisibility visibility) 
            : base(name, visibility, MethodKind.Regular) {
            Debug.Assert(method != null);

            _method = method;
            _mandatoryParamCount = mandatory;
            _optionalParamCount = optional;
            _hasParamsArray = hasParamsArray;
        }

        #region Dynamic Sites

        public override void SetInvocationRule<T>(ActionBinder/*!*/ binder, StandardRule<T>/*!*/ result, ArgumentInfo[] argInfos,
            object[]/*!*/ args, SymbolId? methodNameArgument) {

            Utils.Assert.NotNull(binder, result, args);
            Debug.Assert(argInfos == null || argInfos.Length == args.Length);

            MSA.Expression[] actualArgs = new MSA.Expression[ClrParameterCount];
            List<MSA.Expression> actualArrayArgs = null;

            // method: self, block
            // block: no special args
            int specialParameterCount = IsBlock ? 0 : 2;

            int actualArgIndex = specialParameterCount;
            int argsPassed = 0;
            
            // first non-special argument:
            if (methodNameArgument.HasValue) {
                actualArgs[actualArgIndex++] = Ast.Constant(methodNameArgument.Value);
                argsPassed++;
            }
            
            for (int parameterIndex = 0; parameterIndex < args.Length; parameterIndex++) {
                switch ((argInfos != null) ? argInfos[parameterIndex].Kind : MSA.ArgumentKind.Simple) {
                    case MSA.ArgumentKind.Instance:
                        if (IsBlock) goto case MSA.ArgumentKind.Simple;
                        Debug.Assert(actualArgs[0] == null, "one self only");
                        actualArgs[0] = result.Parameters[parameterIndex];
                        break;

                    case MSA.ArgumentKind.Block:
                        if (IsBlock) throw new NotSupportedException();
                        Debug.Assert(actualArgs[1] == null, "one block only");
                        actualArgs[1] = result.Parameters[parameterIndex];
                        break;

                    case MSA.ArgumentKind.Simple:
                        SetActualArg(actualArgs, ref actualArrayArgs, ref actualArgIndex, ref argsPassed, result.Parameters[parameterIndex]);
                        break;

                    case MSA.ArgumentKind.List:
                        int listLength;
                        MSA.Variable listVariable;
                        AddTestForListArg(result, args[parameterIndex], result.Parameters[parameterIndex], out listLength, out listVariable);
                        
                        // splat list:
                        for (int i = 0; i < listLength; i++) {
                            SetActualArg(actualArgs, ref actualArrayArgs, ref actualArgIndex, ref argsPassed,
                                Ast.Call(Ast.Read(listVariable), typeof(IList<object>).GetMethod("get_Item"), Ast.Constant(i)));
                        }

                        break;

                    default:
                        throw new NotImplementedException();
                }
            }

            if (IsBlock) {
                if (argsPassed < _mandatoryParamCount)
                    throw new NotImplementedException("TODO");

                // fill missing arguments with nulls:
                while (argsPassed < _mandatoryParamCount + _optionalParamCount) {
                    actualArgs[actualArgIndex++] = Ast.Constant(null);
                    argsPassed++;
                }

                if (_hasParamsArray) {
                    Debug.Assert(actualArgIndex == actualArgs.Length - 1);

                    // TODO: optimize
                    actualArgs[actualArgIndex++] = Ast.Call(null, RubyOps.GetMethod("MakeArray"),
                        (actualArrayArgs != null) ? Ast.NewArray(typeof(object[]), actualArrayArgs) : Ast.NewArray(typeof(object[])));
                }

            } else {

                // no block given:
                if (actualArgs[1] == null) {
                    actualArgs[1] = Ast.Null();
                }

                if (argsPassed < _mandatoryParamCount) {
                    // error: too few args
                    result.SetTarget(result.MakeError(binder, MakeErrorExpression(argsPassed, _mandatoryParamCount)));
                    return;
                } else {

                    // TODO: optimize 
                    //MSA.Variable placeholder = result.GetTemporary(typeof(object), "#default");

                    // add optional placeholders:
                    while (argsPassed < _mandatoryParamCount + _optionalParamCount) {
                        actualArgs[actualArgIndex++] = Ast.ReadField(null, RubyOps.DefaultArgumentField);
                        argsPassed++;
                    }

                    if (_hasParamsArray) {
                        Debug.Assert(actualArgIndex == actualArgs.Length - 1);

                        // TODO: optimize
                        actualArgs[actualArgIndex++] = Ast.Call(null, RubyOps.GetMethod("MakeArray"),
                            (actualArrayArgs != null) ? Ast.NewArray(typeof(object[]), actualArrayArgs) : Ast.NewArray(typeof(object[])));

                    } else if (actualArrayArgs != null) {
                        // error: too many args
                        result.SetTarget(result.MakeError(binder, MakeErrorExpression(argsPassed, _mandatoryParamCount)));
                        return;
                    }
                }
            }

            Debug.Assert(actualArgIndex == actualArgs.Length);
            Utils.Assert.NotNullItems(actualArgs);

            result.SetTarget(result.MakeReturn(binder, RubyActionBinder.MakeCallExpression(_method, actualArgs)));
        }

        private static MSA.Expression/*!*/ MakeErrorExpression(int actualCount, int mandatoryCount) {
            return Ast.Call(null, RubyOps.GetMethod("MakeWrongNumberOfArgumentsError"),
                Ast.Constant(actualCount), Ast.Constant(mandatoryCount));
        }

        private void SetActualArg(MSA.Expression[]/*!*/ actualArgs, ref List<MSA.Expression> actualArrayArgs,
            ref int actualArgIndex, ref int argsPassed, MSA.Expression/*!*/ value) {

            if (argsPassed < _mandatoryParamCount + _optionalParamCount) {
                Debug.Assert(actualArgIndex < actualArgs.Length);
                actualArgs[actualArgIndex++] = value;
            } else {
                if (actualArrayArgs == null) {
                    actualArrayArgs = new List<MSA.Expression>();
                }
                actualArrayArgs.Add(value);
            }

            argsPassed++;
        }

        #endregion
    }
}
