/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System;
using System.Collections.Generic;
using System.Text;
using System.Text.RegularExpressions;
using System.Diagnostics;
using System.Reflection;
using System.Reflection.Emit;

using Microsoft.Scripting;
using Microsoft.Scripting.Math;
using Microsoft.Scripting.Actions;
using MSA = Microsoft.Scripting.Ast;

using Ruby.Hosting;
using Ruby.Builtins;
using Ruby.Compiler;
using Microsoft.Scripting.Generation;


namespace Ruby.Runtime {

    public static class RubyConsole {
        // TODO: wire this up to the appropriate warning console output
        // TODO: figure out how to wire up scanner line information to this message -- this signature will change
        public static void Warning(string message) {
            Console.WriteLine("warning: " + message);
        }
    }

    /// <summary>
    /// Emitted to the IL.
    /// </summary>
    public static class RubyOps {

        /// <summary>
        /// DefaultArgument placeholder.
        /// </summary>
        public static readonly object/*!*/ DefaultArgument = new object();
        internal static readonly object[]/*!*/ NullArray = new object[] { null };

        internal static FieldInfo/*!*/ DefaultArgumentField {
            get {
                return typeof(RubyOps).GetField("DefaultArgument");
            }
        }

        // obsolete: static bool _libraryInitialized = false;
        
        
#if OBSOLETE
       
       internal static bool TryGetAttr(CodeContext context, object o, SymbolId name, out object ret) {
            ICustomMembers ids = o as ICustomMembers;
            if (ids != null) {
                return ids.TryGetCustomMember(context, name, out ret);
            }

            return GetDynamicType(o).TryGetMember(context, o, name, out ret);
        }

        internal static RubyClass/*!*/ GetDynamicType(object obj) {
            if (obj is bool) {
                return ((bool)obj) ? _trueClass : _falseClass;
            } else {
                return DynamicHelpers.GetDynamicType(obj);
            }
        }
        
        internal static object CallClrMember(CodeContext/*!*/ context, object target, string/*!*/ name, object[] args, object block) {
            Utils.Assert.NotNull(context, name);
            RubyClass type = GetDynamicType(target);
            object member;

            // TODO:
            if (TryGetClrMember(context, target, name, out member)) {
                if (member is ICallableWithCodeContext) {
                    return ((ICallableWithCodeContext)member).Call(context, args);
                } else if(member is ICallableWithThis) { //  Ops.IsCallable(member) && !(member is RubyModule)
                    return ((ICallableWithThis)member).Call(context, null, args);
                } else if (TryHookEvent(member, context, null, target, args, block)) { // TODO: self?
                    return null;
                } else {
                    return member;
                }
            }

            throw MakeMissingMethodException(target, name);
        }
        
        private static bool TryGetClrMember(CodeContext/*!*/ context, object target, string/*!*/ name, out object member) {
            if (TryGetAttr(context, target, SymbolTable.StringToId(name), out member) ||
                TryGetAttr(context, null, SymbolTable.StringToId(name), out member)) {
                return true;
            }

            string unmangled = UnmangleName(name);
            SymbolId unmangledSymbol = SymbolTable.StringToId(unmangled); //SymbolTable.StringToCaseInsensitiveId(unmangled);

            if (TryGetAttr(context, target, unmangledSymbol, out member) ||
                TryGetAttr(context, null, unmangledSymbol, out member)) {
                
                return true;
            }
            return false;
        }
        // emitted:
        public static object InvokeMethod(string/*!*/ name, CodeContext context, bool hasExplicitTarget, object target, object block, object[]/*!*/ args) {
            object result;

            if (context == null) context = RubyEngine.CurrentEngine.DefaultContext;

            if (!TryInvokeMethod(context, hasExplicitTarget, target, name, args, block, out result)) {
                if (!TryInvokeMethod(context, hasExplicitTarget, target, "method_missing", Utils.Array.Insert<object>(name, args), block, out result)) {
                    throw MakeMissingMethodException(target, name);
                }
            }
            return result;
        }

        public static bool TryInvokeMethod(CodeContext/*!*/ context, bool hasExplicitTarget, object target, string/*!*/ name, object[]/*!*/ args, object block, out object result) {
            Debug.Assert(name != null && args != null);

            if (target == null) {
                target = WellKnownSingletons.NilObject;
            } else if (target is bool) {
                if ((bool)target) {
                    target = WellKnownSingletons.TrueObject;
                } else {
                    target = WellKnownSingletons.FalseObject;
                }
            }

            object method;
            if (!TryLookupMethod(context, target, name, out method)) {
                result = null;
                return false;
            }

            // Both instance and class methods have self argument as the first parameter.

            RubyMethodInfo rubyMethod = method as RubyMethodInfo;
            if (rubyMethod != null) {
                rubyMethod.Instance = target;
                result = rubyMethod.CallWithBlock(context, block, args);
                return true;
            }

            if (TryHookEvent(method, context, self, target, args, block)) {
                result = null;
                return true;
            }

            // TODO: block as the last argument?
            if (method is ICallableWithCodeContext) {
                result = ((ICallableWithCodeContext)method).Call(context, args);
            } else if (method is ICallableWithThis) { //  Ops.IsCallable(member) && !(member is RubyModule)
                result = ((ICallableWithThis)method).Call(context, null, args);
            } else {
                result = method;
            }
            return true;
        }
      
        internal static bool TryHookEvent(object member, CodeContext/*!*/ context, object self, object target, object[]/*!*/ args, object block) {
            ReflectedEvent.BoundEvent evnt = member as ReflectedEvent.BoundEvent;
            if (evnt != null) {
                if (block != null) {
                    Callable blockMethod = block as Callable;

                    if (blockMethod == null) {
                        Debug.Assert(self != null); // TODO: calling via method_missing
                        blockMethod = MakeBlock(context, self, (Delegate)block);
                    }

                    evnt.InPlaceAdd(context, blockMethod);
                    return true;
                } else {
                    object handler;

                    if (!TryLookupMethod(context, args[0], (string)args[1], out handler)) {
                        throw new Exception("Invalid handler");
                    }

                    Callable handlerMethod = new Callable(((Callable)handler).Delegate, false);
                    handlerMethod.Instance = args[0];
                    handlerMethod.Context = context;

                    evnt.InPlaceAdd(context, handlerMethod);
                    return true;
                }
            }
            return false;
        }

        internal static object CallMethod(CodeContext/*!*/ context, object instance, Delegate/*!*/ method, 
            object block, bool hasBlock, object[] args) {

            Debug.Assert(context.LanguageContext is RubyContext);

            if (hasBlock) {
                args = Utils.Array.Insert(block, args);
            }

            if (instance != null) {
                args = Utils.Array.Insert(instance, args);
            }

            if (method is CallTarget0) {
                return ((CallTarget0)method)();
            } else if (method is CallTarget1) {
                return ((CallTarget1)method)(args[0]);
            } else if (method is CallTarget2) {
                return ((CallTarget2)method)(args[0], args[1]);
            } else if (method is CallTarget3) {
                return ((CallTarget3)method)(args[0], args[1], args[2]);
            } else if (method is CallTarget4) {
                return ((CallTarget4)method)(args[0], args[1], args[2], args[3]);
            } else if (method is CallTarget5) {
                return ((CallTarget5)method)(args[0], args[1], args[2], args[3], args[4]);
            } else if (method is CallTargetN) {
                return ((CallTargetN)method)(args);
            } else if (method is CallTargetWithContext0) {
                return ((CallTargetWithContext0)method)(context);
            } else if (method is CallTargetWithContext1) {
                return ((CallTargetWithContext1)method)(context, args[0]);
            } else if (method is CallTargetWithContext2) {
                return ((CallTargetWithContext2)method)(context, args[0], args[1]);
            } else if (method is CallTargetWithContext3) {
                return ((CallTargetWithContext3)method)(context, args[0], args[1], args[2]);
            } else if (method is CallTargetWithContext4) {
                return ((CallTargetWithContext4)method)(context, args[0], args[1], args[2], args[3]);
            } else if (method is CallTargetWithContext5) {
                return ((CallTargetWithContext5)method)(context, args[0], args[1], args[2], args[3], args[4]);
            } else if (method is CallTargetWithContextN) {
                return ((CallTargetWithContextN)method)(context, args);
            } 

            throw new NotSupportedException("Delegate '" + method.ToString() + "' not supported");
        }
#endif

        #region Emitted

        // emitted:
        public static Proc/*!*/ DefineBlock(CodeContext/*!*/ context, object/*!*/ self, Delegate/*!*/ clrMethod,
            int parameterCount, bool hasParamsArray) {
            Utils.Assert.NotNull(context, self, clrMethod);

            // closes block over self and context
            return new Proc(ProcKind.Block, new RubyMethodInfo(clrMethod, parameterCount, hasParamsArray), context, self);
        }

        // emitted:
        public static object DefineMethod(CodeContext/*!*/ context, object/*!*/ self, object target, SymbolId name, Delegate/*!*/ clrMethod,
            int mandatory, int optional, bool hasParamsArray) {
            Utils.Assert.NotNull(context, self, clrMethod);

            RubyContext rubyContext = (RubyContext)context.LanguageContext;

            RubyMethodInfo method = new RubyMethodInfo(clrMethod, name, mandatory, optional, hasParamsArray, rubyContext.CurrentVisibility);

            if (target != null) {
                // class or singleton method:
                RubyModule module = target as RubyModule;
                if (module != null) {
                    // class method:
                    module.SetClassMethod(method);
                } else {
                    // singleton method:
                    throw new NotSupportedException("Singletons");
                }
            } else {
                // TODO: modularity, interop
                //if (ReferenceEquals(self, ((RubyModuleContext)context.ModuleContext).ModuleSelf)) {
                //    GetGlobalScope(context).SetName(symbolicName, method);
                //    method.Instance = self;
                //}

                // declaring module is self.class unless self is a module object:
                RubyModule declaringModule = self as RubyModule;
                if (declaringModule == null) {
                    declaringModule = GetExecutionContext(context).GetClassOf(self);
                }

                declaringModule.SetInstanceMethod(method);
            }

            return null;
        }
        
        // emitted
        public static RubyModule/*!*/ DefineModule(CodeContext/*!*/ context, object/*!*/ self, SymbolId name) {
            Utils.Assert.NotNull(context, self);
            RubyExecutionContext ec = GetExecutionContext(context);
            return ec.DefineModule(ec.GetConstantOwner(self), name, null);
        }

        // emitted 
        public static RubyModule/*!*/ DefineClass(CodeContext/*!*/ context, object self, SymbolId name, object superClassObject) {
            Utils.Assert.NotNull(context);
            RubyExecutionContext ec = GetExecutionContext(context);

            RubyClass superClass;

            if (superClassObject != null) {
                superClass = ec.AsClass(superClassObject);
                if (superClass == null) {
                    throw RubyExceptions.CreateTypeError(
                        String.Format("superclass must be a Class ({0} given)", ec.GetClassOf(superClassObject).Name)
                    );
                }
            } else {
                superClass = ec.ObjectClass;
            }

            return ec.DefineModule(ec.GetConstantOwner(self), name, superClass);
        }
        
        // emitted:
        public static object GetConstant(CodeContext/*!*/ context, object target, SymbolId name) {
            RubyExecutionContext ec = GetExecutionContext(context);
            object result;

            // TODO: hack, we need dynamic sites for constants first and DLR needs IDO on ScriptModules
            if (target is ScriptModule) {
                if (((ScriptModule)target).TryGetCustomMember(context, name, out result)) {
                    return result;
                }
            }

            RubyModule owner = ec.GetConstantOwner(target);
            
            if (!owner.TryResolveConstant(name, out result)) {
                throw RubyExceptions.CreateNameError(String.Format("uninitialized constant {0}",
                    ReferenceEquals(owner, ec.ObjectClass) ? SymbolTable.IdToString(name) : owner.Name + "::" + name));
            }
            return result;
        }

        // emitted:
        public static object SetConstant(CodeContext/*!*/ context, object target, SymbolId name, object value) {
            GetExecutionContext(context).GetConstantOwner(target).SetConstant(name, value);
            return value;
        }
        
        // emitted:
        public static object GetSelf(CodeContext/*!*/ context) {
            Debug.Assert(context != null);
            // TODO: new module model
            return ((RubyContext)context.LanguageContext).SelfObject;
        }

        // emitted:
        public static List<object>/*!*/ MakeArray(object[]/*!*/ items) {
            Debug.Assert(items != null);
            return new List<object>(items);
        }

        // emitted:
        public static Dictionary<object, object>/*!*/ MakeHash(object[]/*!*/ items) {
            Debug.Assert(items != null && items.Length % 2 == 0);

            Dictionary<object, object> result = new Dictionary<object, object>(items.Length / 2);
            for (int i = 0; i < items.Length; i += 2) {
                Debug.Assert(i + 1 < items.Length);
                result.Add(items[i], items[i + 1]);
            }

            return result;
        }

        // emitted:
        public static object[]/*!*/ AddSplattedArguments(object[]/*!*/ args, object splattee) {
            List<object> list = splattee as List<object>;
            if (list != null) {
                args = Utils.Array.AppendRange(args, list, 0);
            }
            return args;
        }

        // emitted:
        public static object[]/*!*/ AddSplattedAndAdditionalArguments(object[]/*!*/ args, object splattee, object additionalArgument) {
            List<object> list = splattee as List<object>;
            if (list != null) {
                args = Utils.Array.AppendRange(args, list, 1);
                args[args.Length - 1] = additionalArgument;
            }
            return args;
        }

        // emitted:
        public static object GetGlobalVariable(CodeContext/*!*/ context, SymbolId name) {
            object value;
            GetExecutionContext(context).GlobalVariables.TryGetValue(name, out value);
            return value;
        }

        // emitted:
        public static object SetGlobalVariable(CodeContext/*!*/ context, SymbolId name, object value) {
            return GetExecutionContext(context).GlobalVariables[name] = value;
        }

        // emitted:
        public static object Concat(object[] parts) {
            Utils.Assert.NotNull(parts);

            // TODO:
            StringBuilder result = new StringBuilder();
            foreach (object part in parts) {
                result.Append(part);
            }

            return result.ToString();
        }

        // emitted:
        public static object ConcatMutableString(object[] parts) {
            Utils.Assert.NotNull(parts);

            MutableString str = new MutableString();
            foreach (object part in parts) {
                str.Append(RubySites.ToS(part));
            }
            return str;
        }

        // emitted:
        public static object CreateRegex(object pattern) {
            return new System.Text.RegularExpressions.Regex(pattern.ToString());
        }

        // emitted:
        public static object CreateRegex2(object[] patterns) {
            return new System.Text.RegularExpressions.Regex(Concat(patterns).ToString());
        }

        // emitted:
        public static bool IsTrue(object obj) {
            return (obj is bool) ? (bool)obj == true : obj != null;
        }

        // emitted:
        public static bool IsFalse(object obj) {
            return (obj is bool) ? (bool)obj == false : obj == null;
        }

        // emitted:
        public static Exception GetCurrentException(CodeContext/*!*/ context) {
            return GetExecutionContext(context).CurrentException;
        }

        // emitted:
        public static Exception SetCurrentException(CodeContext/*!*/ context, object value) {
            Exception e = value as Exception;
            if (e == null) {
                throw RubyExceptions.CreateTypeError("assigning non-exception to $!");
            }
            return GetExecutionContext(context).CurrentException = e;
        }
        
        // emitted:
        public static Exception SaveException(CodeContext/*!*/ context, Exception/*!*/ currentException) {
            Utils.Assert.NotNull(context, currentException);

            RubyExecutionContext ec = GetExecutionContext(context);
            Exception previous = ec.CurrentException;
            ec.CurrentException = currentException;
            return previous;
        }

        // emitted:
        public static void RestoreException(CodeContext/*!*/ context, Exception previousException) {
            GetExecutionContext(context).CurrentException = previousException;
        }

        private static DynamicSite<object, object, bool> _compareExceptionSite = DynamicSite<object, object, bool>.Create(
            // TODO: visibility (hasExplicitTarget)
            new InvokeMemberAction(Symbols.StrictEqual, InvokeMemberActionFlags.IsCallWithThis, new ArgumentInfo[] {
                new ArgumentInfo(MSA.ArgumentKind.Instance),
                new ArgumentInfo(MSA.ArgumentKind.Simple) 
            })
        );

        // emitted:
        public static bool CompareException(CodeContext/*!*/ context, object/*!*/ classObject) {
            Utils.Assert.NotNull(context, classObject);
            RubyExecutionContext ec = GetExecutionContext(context);
            return _compareExceptionSite.Invoke(context, classObject, ec.CurrentException);
        }

        // emitted:
        public static bool CompareDefaultException(CodeContext/*!*/ context, object/*!*/ self) {
            RubyExecutionContext ec = GetExecutionContext(context);
            
            // MRI doesn't call === here
            return ec.IsInstanceOf(ec.CurrentException, ec.StandardErrorClass);
        }

        // emitted:
        public static IList<object> AsArrayOfLength(object obj, int requiredLength) {
             IList<object> result = obj as IList<object>;
             return (result != null && result.Count == requiredLength) ? result : null;
        }

        // emitted:
        public static ArgumentException/*!*/ MakeWrongNumberOfArgumentsError(int actual, int expected) {
            return new ArgumentException(String.Format("wrong number of arguments ({0} for {1})", actual, expected));
        }

        // emitted
        public static Exception/*!*/ MakeMissingMethodException(object target, string name) {
            // TODO:
            return new MissingMethodException(((target != null) ? target.ToString() : "NilClass") + "::" + name);
        }

        // emitted
        public static Exception/*!*/ MakeInvalidArgumentTypesError(string methodName) {
            // TODO:
            return new MissingMethodException(String.Format("No overload of method `{0}' matches the specified argument types", methodName));
        }

        #endregion

        internal static MethodInfo/*!*/ GetMethod(string/*!*/ name) {
            MethodInfo result = typeof(RubyOps).GetMethod(name, BindingFlags.Public | BindingFlags.Static);
            Utils.Assert.NotNull(result);
            return result;
        }
        
        // unit testing helper
        public static void ClearGlobalSpace() {
            DynamicHelpers.TopPackage.Clear();
            RubyExecutionContext.ClearThreadStatics();
            RubyEngine.CurrentEngine.ClearExecutionContext();
            RubyEngine.CurrentEngine.ActionBinder.ClearRuleCache();
        }

        internal static RubyExecutionContext/*!*/ GetExecutionContext(CodeContext/*!*/ codeContext) {
            // TODO: can we get EC from CC (via module)?
            return RubyEngine.CurrentEngine.ExecutionContext;
        }

        // emitted:
        public static Range CreateRange(object begin, object end) {
            return new Range(begin, end, false);
        }

        // emitted:
        public static Range CreateExclusiveRange(object begin, object end) {
            return new Range(begin, end, true);
        }

        // emitted:
        public static MutableString CreateMutableString(object str) {
            if (str is MutableString)
                return (MutableString)str;
            else
                return new MutableString(str.ToString());
        }

        // emitted: (IDO implementation for Ruby classes)
        public static StandardRule<T> GetClassRule<T>(Action/*!*/ action, CodeContext/*!*/ callerContext, object[]/*!*/ args) {
            return RubyActionBinder.GetUserClassRule<T>(action, callerContext, args);
        }

        // emitted: (IDO implementation for Ruby classes)
        public static LanguageContext/*!*/ GetLanguageContext() {
            return RubyEngine.CurrentEngine.DefaultContext.LanguageContext;
        }

#if TODO // CLR interop (IEnumerable.Each)
        public static object ForEach(CodeContext/*!*/ context, object/*!*/ instance, Proc/*!*/ block) {
            Utils.Assert.NotNull(context, instance, block);

            object[] args = new object[1];

            System.Collections.IEnumerable einst = instance as System.Collections.IEnumerable;
            if (einst != null) {
                foreach (object item in einst) {
                    args[0] = item;

                    // TODO:
                    BlockReturnReason returnReason;
                    RuntimeFlowControl.ClrMethod.YieldBlock(context, block, args, out returnReason);
                    if (returnReason == BlockReturnReason.BreakOrRetry) break;
                }
            } else {
                foreach (object item in (IEnumerable<object>)instance) {
                    args[0] = item;

                    // TODO:
                    BlockReturnReason returnReason;
                    RuntimeFlowControl.ClrMethod.YieldBlock(context, block, args, out returnReason);
                    if (returnReason == BlockReturnReason.BreakOrRetry) break;
                }
            }

            return null;
        }
#endif

        private static Scope/*!*/ GetGlobalScope(CodeContext/*!*/ context) {
            Scope result = context.Scope;
            while (result.Parent != null) result = result.Parent;
            return result;
        }

        public static string/*!*/ UnmangleName(string/*!*/ name) {
            StringBuilder sb = new StringBuilder(name.Length);
            bool upcase = true;
            for (int i = 0; i < name.Length; i++) {
                if (name[i] == '_') {
                    upcase = true;
                } else {
                    if (upcase) {
                        sb.Append(Char.ToUpper(name[i]));
                        upcase = false;
                    } else {
                        sb.Append(name[i]);
                    }
                }

            }
            return sb.ToString();
        }

        // TODO: internal (unit test)
        public static string/*!*/ MangleName(string/*!*/ name) {
            Utils.Assert.NotNull(name);

            StringBuilder result = new StringBuilder(name.Length);

            for (int i = 0; i < name.Length; i++) {
                if (Char.IsUpper(name[i])) {
                    if (!(i == 0 || i + 1 < name.Length && Char.IsUpper(name[i + 1]) || i + 1 == name.Length && Char.IsUpper(name[i - 1]))) {
                        result.Append('_');
                    }
                    result.Append(Char.ToLower(name[i]));
                } else {
                    result.Append(name[i]);
                }
            }

            return result.ToString();
        }
        
        
    }
}
