/* ****************************************************************************
 *
 * Copyright (c) Microsoft Corporation. 
 *
 * This source code is subject to terms and conditions of the Microsoft Permissive License. A 
 * copy of the license can be found in the License.html file at the root of this distribution. If 
 * you cannot locate the  Microsoft Permissive License, please send an email to 
 * ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
 * by the terms of the Microsoft Permissive License.
 *
 * You must not remove this notice, or any other, from this software.
 *
 *
 * ***************************************************************************/

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using Ruby.Runtime;
using System.Collections;
using Microsoft.Scripting;
using Microsoft.Scripting.Actions;
using Ruby.Compiler;
using Ruby.Hosting;
using MSA = Microsoft.Scripting.Ast;

namespace Ruby.Builtins {
    public static class RubySites {
        private static FastDynamicSite<object, MutableString> InspectSharedSite = FastDynamicSite<object, MutableString>.Create(
            RubyEngine.CurrentEngine.DefaultContext,
            new InvokeMemberAction(SymbolTable.StringToId("inspect"), InvokeMemberActionFlags.IsCallWithThis | InvokeMemberActionFlags.ReturnNonCallable,
                new ArgumentInfo[] { new ArgumentInfo(MSA.ArgumentKind.Instance) }));

        public static MutableString Inspect(object obj) {
            return InspectSharedSite.Invoke(obj);
        }

        private static FastDynamicSite<RubyClass, object> AllocateSharedSite = FastDynamicSite<RubyClass, object>.Create(
            RubyEngine.CurrentEngine.DefaultContext,
            new InvokeMemberAction(SymbolTable.StringToId("allocate"), InvokeMemberActionFlags.None,
            new ArgumentInfo[] { new ArgumentInfo(MSA.ArgumentKind.Instance) }));

        public static object Allocate(RubyClass klass) {
            return AllocateSharedSite.Invoke(klass);
        }

        private static FastDynamicSite<object, object, int> CompareSharedSite = FastDynamicSite<object, object, int>.Create(
            RubyEngine.CurrentEngine.DefaultContext,
            new InvokeMemberAction(SymbolTable.StringToId("<=>"), InvokeMemberActionFlags.IsCallWithThis,
            new ArgumentInfo[] { new ArgumentInfo(MSA.ArgumentKind.Instance), new ArgumentInfo(MSA.ArgumentKind.Simple) }));

        public static int Compare(object lhs, object rhs) {
            return CompareSharedSite.Invoke(lhs, rhs);
        }

        private static FastDynamicSite<object, object, bool> EqualSharedSite = FastDynamicSite<object, object, bool>.Create(
            RubyEngine.CurrentEngine.DefaultContext,
            new InvokeMemberAction(SymbolTable.StringToId("=="), InvokeMemberActionFlags.IsCallWithThis,
            new ArgumentInfo[] { new ArgumentInfo(MSA.ArgumentKind.Instance), new ArgumentInfo(MSA.ArgumentKind.Simple) }));

        public static bool Equal(object lhs, object rhs) {
            return EqualSharedSite.Invoke(lhs, rhs);
        }

        private static FastDynamicSite<object, SymbolId, bool> RespondToSharedSite = FastDynamicSite<object, SymbolId, bool>.Create(
            RubyEngine.CurrentEngine.DefaultContext,
            new InvokeMemberAction(SymbolTable.StringToId("respond_to?"), InvokeMemberActionFlags.IsCallWithThis,
            new ArgumentInfo[] { new ArgumentInfo(MSA.ArgumentKind.Instance), new ArgumentInfo(MSA.ArgumentKind.Simple) }));

        public static bool RespondTo(object obj, SymbolId name) {
            return RespondToSharedSite.Invoke(obj, name);
        }

        public static bool RespondTo(object obj, string name) {
            return RespondToSharedSite.Invoke(obj, SymbolTable.StringToId(name));
        }

        private static FastDynamicSite<object, int> ToISharedSite = FastDynamicSite<object, int>.Create(
            RubyEngine.CurrentEngine.DefaultContext,
            new InvokeMemberAction(SymbolTable.StringToId("to_i"), InvokeMemberActionFlags.IsCallWithThis,
            new ArgumentInfo[] { new ArgumentInfo(MSA.ArgumentKind.Instance) }));

        public static int ToI(object value) {
            return ToISharedSite.Invoke(value);
        }

        private static FastDynamicSite<object, int> ToIntSharedSite = FastDynamicSite<object, int>.Create(
            RubyEngine.CurrentEngine.DefaultContext,
            new InvokeMemberAction(SymbolTable.StringToId("to_int"), InvokeMemberActionFlags.IsCallWithThis,
            new ArgumentInfo[] { new ArgumentInfo(MSA.ArgumentKind.Instance) }));

        public static int ToInt(object value) {
            if (value == null)
                throw RubyExceptions.CreateTypeError("no implicit conversion from nil to integer");

            if (RubySites.RespondTo(value, "to_int"))
                return ToIntSharedSite.Invoke(value);
            else
                throw RubyExceptions.CannotConvertTypeToTargetType(value, "Integer");
        }

        private static FastDynamicSite<object, MutableString> ToStrSharedSite = FastDynamicSite<object, MutableString>.Create(
            RubyEngine.CurrentEngine.DefaultContext,
            new InvokeMemberAction(SymbolTable.StringToId("to_str"), InvokeMemberActionFlags.IsCallWithThis,
            new ArgumentInfo[] { new ArgumentInfo(MSA.ArgumentKind.Instance) }));

        public static MutableString ToStr(object value) {
            return ToStrSharedSite.Invoke(value);
        }

        private static FastDynamicSite<object, Range> ToRangeSharedSite = FastDynamicSite<object, Range>.Create(
            RubyEngine.CurrentEngine.DefaultContext,
            new InvokeMemberAction(SymbolTable.StringToId("to_range"), InvokeMemberActionFlags.IsCallWithThis,
            new ArgumentInfo[] { new ArgumentInfo(MSA.ArgumentKind.Instance) }));

        public static Range ToRange(object value) {
            return ToRangeSharedSite.Invoke(value);
        }

        private static FastDynamicSite<object, List<object>> ToArySharedSite = FastDynamicSite<object, List<object>>.Create(
            RubyEngine.CurrentEngine.DefaultContext,
            new InvokeMemberAction(SymbolTable.StringToId("to_ary"), InvokeMemberActionFlags.IsCallWithThis,
            new ArgumentInfo[] { new ArgumentInfo(MSA.ArgumentKind.Instance) }));

        public static List<object> ToAry(object obj) {
            return ToArySharedSite.Invoke(obj);
        }

        private static FastDynamicSite<object, MutableString> ToSSharedSite = FastDynamicSite<object, MutableString>.Create(
            RubyEngine.CurrentEngine.DefaultContext,
            new InvokeMemberAction(SymbolTable.StringToId("to_s"), InvokeMemberActionFlags.IsCallWithThis,
            new ArgumentInfo[] { new ArgumentInfo(MSA.ArgumentKind.Instance) }));

        public static MutableString ToS(object value) {
            return ToSSharedSite.Invoke(value);
        }

        private static FastDynamicSite<Range, object> RangeBeginSharedSite = FastDynamicSite<Range, object>.Create(
            RubyEngine.CurrentEngine.DefaultContext,
            new InvokeMemberAction(SymbolTable.StringToId("begin"), InvokeMemberActionFlags.IsCallWithThis,
            new ArgumentInfo[] { new ArgumentInfo(MSA.ArgumentKind.Instance) }));

        public static object RangeBegin(Range range) {
            return RangeBeginSharedSite.Invoke(range);
        }

        private static FastDynamicSite<Range, object> RangeEndSharedSite = FastDynamicSite<Range, object>.Create(
            RubyEngine.CurrentEngine.DefaultContext,
            new InvokeMemberAction(SymbolTable.StringToId("end"), InvokeMemberActionFlags.IsCallWithThis,
            new ArgumentInfo[] { new ArgumentInfo(MSA.ArgumentKind.Instance) }));

        public static object RangeEnd(Range range) {
            return RangeEndSharedSite.Invoke(range);
        }

        private static FastDynamicSite<Range, bool> RangeExcludeEndSharedSite = FastDynamicSite<Range, bool>.Create(
            RubyEngine.CurrentEngine.DefaultContext,
            new InvokeMemberAction(SymbolTable.StringToId("exclude_end?"), InvokeMemberActionFlags.IsCallWithThis,
            new ArgumentInfo[] { new ArgumentInfo(MSA.ArgumentKind.Instance) }));

        public static bool RangeExcludeEnd(Range range) {
            return RangeExcludeEndSharedSite.Invoke(range);
        }

        private static FastDynamicSite<RubyModule, RubyModule, object> ModuleAppendFeaturesSharedSite = FastDynamicSite<RubyModule, RubyModule, object>.Create(
            RubyEngine.CurrentEngine.DefaultContext,
            new InvokeMemberAction(SymbolTable.StringToId("append_features"), InvokeMemberActionFlags.None,
            new ArgumentInfo[] { new ArgumentInfo(MSA.ArgumentKind.Instance), new ArgumentInfo(MSA.ArgumentKind.Simple) }));

        public static object ModuleAppendFeatures(RubyModule self, RubyModule module) {
            return ModuleAppendFeaturesSharedSite.Invoke(self, module);
        }

        private static FastDynamicSite<RubyModule, RubyModule, object> ModuleIncludedSharedSite = FastDynamicSite<RubyModule, RubyModule, object>.Create(
            RubyEngine.CurrentEngine.DefaultContext,
            new InvokeMemberAction(SymbolTable.StringToId("included"), InvokeMemberActionFlags.None,
            new ArgumentInfo[] { new ArgumentInfo(MSA.ArgumentKind.Instance), new ArgumentInfo(MSA.ArgumentKind.Simple) }));

        public static object ModuleIncluded(RubyModule self, RubyModule module) {
            return ModuleIncludedSharedSite.Invoke(self, module);
        }

        private static FastDynamicSite<object, Exception> ExceptionSharedSite = FastDynamicSite<object, Exception>.Create(
            RubyEngine.CurrentEngine.DefaultContext,
            new InvokeMemberAction(SymbolTable.StringToId("exception"), InvokeMemberActionFlags.None,
            new ArgumentInfo[] { new ArgumentInfo(MSA.ArgumentKind.Instance) }));

        public static Exception Exception(object self) {
            return ExceptionSharedSite.Invoke(self);
        }
    }
}
