#####################################################################################
#
#  Copyright (c) Microsoft Corporation. All rights reserved.
#
# This source code is subject to terms and conditions of the Microsoft Permissive License. A 
# copy of the license can be found in the License.html file at the root of this distribution. If 
# you cannot locate the  Microsoft Permissive License, please send an email to 
# ironruby@microsoft.com. By using this source code in any fashion, you are agreeing to be bound 
# by the terms of the Microsoft Permissive License.
#
# You must not remove this notice, or any other, from this software.
#
#
#####################################################################################

THIS_DIRECTORY = File.expand_path(File.dirname(__FILE__)).downcase
CURR_DIRECTORY = Dir.pwd.downcase

require File.join(THIS_DIRECTORY, 'common')

$skip_list = '''
Compat/*
Syntax/*
Builtin/Array/*
Builtin/Class/*
Builtin/Range/*
Runtime/Block/*
Runtime/Class/*
Runtime/Exception/test_raise_string.rb
Runtime/Exception/test_rescue_modifier.rb
Runtime/Exception/test_retry.rb
Runtime/Exception/test_ensure.rb
Runtime/Exception/test_match.rb
Runtime/Exception/test_return.rb
Runtime/Module/*
Runtime/Expression/*
Runtime/Method/test_defaultvalue.rb
Runtime/Method/test_normal_arg.rb
Runtime/Method/test_recursive.rb
Util/*
XLang/*
Tools/*
    '''.downcase.split

$skip_list.collect! { |t| File.join(THIS_DIRECTORY, t) }

class TestDirectory
    attr_reader :directory

    def initialize(dir)
        @directory = dir.downcase
        @local_skip_list = $skip_list.find_all { |p| p.include? @directory }
    end 
    
    def get_test_list()
        Dir.glob("test_*.rb").reject { |p| @local_skip_list.include? File.expand_path(p).downcase }
    end 
    
    def get_directory_list()
        temp = []
        Dir.foreach('.') do |d|
            next if d == '.' or d == '..'
            next unless File.directory? d
            temp << File.expand_path(d).downcase
        end
        temp.reject { |p| @local_skip_list.include? File.expand_path(p) + "/*" } 
    end 
end 

def runTests(driver, td)
    if td.get_test_list.size > 0 
        total = failure = 0
        printf "running test*.rb under #{td.directory} \n"

        td.get_test_list.each do |f|
            total += 1
            if driver.run(f) == 0 
                printf "."
            else
                printf "X(%s)", f
                failure += 1
            end 
        end 
        
        printf "  [%d/%d]\n", failure, total
        failure
    else 
        0
    end 
end 

$failure = 0
def walk_directory_tree(driver, td)
    saved = Dir.pwd
    begin
        Dir.chdir(td.directory)

        $failure += runTests(driver, td)

        td.get_directory_list.each do |sub_dir|
            walk_directory_tree driver, TestDirectory.new(sub_dir)
        end 

    ensure
        Dir.chdir(saved)
    end 
end 

#
# ruby.exe -dir:a,b,c
#
test_directories = []
ARGV.each do |arg|
    if arg =~ /-dir:(.+)/i
        test_directories = $1.split(',')
    end 
end
test_directories.collect! { |d| File.join(CURR_DIRECTORY, d).downcase }

applicable_drivers = [ 
    Test::CRubyDriver,
    Test::IronRubyDriver_M1,
]

applicable_drivers.each do |driver|
    puts "==============================================================\n"
    puts "============   Running under #{driver}  \n"
    puts "==============================================================\n"

    if test_directories.size > 0
        test_directories.each do |subdir|  
            walk_directory_tree driver, TestDirectory.new(subdir)
        end 
    else
        walk_directory_tree driver, TestDirectory.new(CURR_DIRECTORY)
    end 
end

if THIS_DIRECTORY == CURR_DIRECTORY
    $special_list = '''
Syntax
    '''.downcase.split
    $special_list.collect! { |d| File.join(THIS_DIRECTORY, d).downcase }

    if test_directories.size > 0
        $special_list = $special_list.find_all { |d| test_directories.include? d}
    end 
    
    $special_list.each do |d|
        saved = Dir.pwd
        begin 
            Dir.chdir(d)
            run_directory_rb = "run_#{File.basename(d)}.rb"
            puts "==============================================================\n"
            puts "running #{run_directory_rb} under #{d}\n"
            result = Test::CRubyDriver.run(run_directory_rb, nil)
            if result == 0
                puts "pass\n"
            else
                $failure += result
                puts "fail (#{result})\n"
            end 
        ensure
            Dir.chdir(saved)
        end 
    end 
end 

puts "Summary: #{$failure}\n"
exit($failure)